#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <errno.h>

#ifdef HAVE_SYSTEMD_SD_DAEMON_H
#include <systemd/sd-daemon.h>
#endif

#include "arsockserver.hh"

SocketServer::Client::Client(SocketServer *server, const Glib::RefPtr<Gio::SocketConnection>& connection)
	: m_server(server), m_connection(connection), m_refcount(1), m_closeprotect(0), m_firstreply(true)
{
	if (true)
		std::cout << "client create " << (unsigned long)this << " pid " << getpid() << std::endl;
	if (!m_connection || !m_server) {
		m_server = 0;
		m_connection.reset();
		return;
	}
	m_connection->get_socket()->property_blocking() = false;
	m_connin = Glib::signal_io().connect(sigc::mem_fun(*this, &Client::on_input), m_connection->get_socket()->get_fd(), Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	m_server->add_client(get_ptr());
	if (true)
		std::cout << "client connect " << (unsigned long)this << " pid " << getpid() << " fd " << m_connection->get_socket()->get_fd() << std::endl;
}

SocketServer::Client::~Client()
{
	close();
	if (true)
		std::cout << "client destroy " << (unsigned long)this << " pid " << getpid() << std::endl;
}

void SocketServer::Client::reference(void) const
{
        g_atomic_int_inc(&m_refcount);
	if (false)
		std::cout << "client " << (unsigned long)this << " pid " << getpid() << " reference " << m_refcount << std::endl;
}

void SocketServer::Client::unreference(void) const
{
	bool nz(!g_atomic_int_dec_and_test(&m_refcount));
	if (false)
		std::cout << "client " << (unsigned long)this << " pid " << getpid() << " unreference " << m_refcount << std::endl;
        if (nz)
                return;
        delete this;
}

void SocketServer::Client::close(void)
{
	m_connin.disconnect();
	m_connout.disconnect();
	m_sendqueue.clear();
	if (m_longpoll)
		m_longpoll->unset_longpoll(get_ptr());
	m_longpoll.reset();
	m_logfilter.clear();
	m_logdiscard.clear();
	if (!m_connection)
		return;
	if (true)
		std::cout << "client " << (unsigned long)this << " pid " << getpid() << " close fd " << m_connection->get_socket()->get_fd() << std::endl;
	m_connection->get_socket()->close();
	m_connection.reset();
	if (m_closeprotect == 1)
		m_closeprotect = 2;
	else if (m_server)
		m_server->remove_client(get_ptr());
}

void SocketServer::Client::send(const Json::Value& v)
{
	if (!m_connection)
		return;
	m_sendqueue.push_back(v);
	m_connout.disconnect();
	m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Client::on_output), m_connection->get_socket()->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
	m_firstreply = false;
}

bool SocketServer::Client::on_output(Glib::IOCondition iocond)
{
	m_connout.disconnect();
	if (m_sendqueue.empty())
		return true;
	Json::FastWriter writer;
	std::string msg(writer.write(*m_sendqueue.begin()));
	try {
		gssize r(m_connection->get_socket()->send(msg.c_str(), msg.size()));
		if (r == -1) {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket output return value: " << r << std::endl;
			close();
			return true;
		}
		m_sendqueue.erase(m_sendqueue.begin());
		if (true)
			std::cout << "client " << (unsigned long)this << " pid " << getpid() << " tx " << msg << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code() == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket output would block" << std::endl;
		} else {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket output error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (!m_sendqueue.empty())
		m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Client::on_output), m_connection->get_socket()->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
	return true;
}

bool SocketServer::Client::on_input(Glib::IOCondition iocond)
{
	if (!m_connection) {
		m_connin.disconnect();
		return true;
	}
	if (!m_server)
		return false;
	if (m_server->is_sockclient()) {
		if (true)
			std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " receive in socket client" << std::endl;
		close();
		return false;
	}
	Json::Value root;
	Json::Value reply;
	if (m_firstreply) {
		reply["version"] = VERSION;
		reply["provider"] = "cfmu";
	}
	try {
		char buf[16384];
 		gssize r(m_connection->get_socket()->receive(buf, sizeof(buf)));
		if (r <= 0) {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			close();
			return true;
		}
		if (r >= sizeof(buf)) {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			reply["error"] = std::string("invalid message size");
			send(reply);
			return true;
		}
		{
			Json::Reader reader;
			if (!reader.parse(buf, buf + r, root, false)) {
				if (true)
					std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " cannot parse JSON: " << std::string(buf, buf + r);
				reply["error"] = std::string("cannot parse JSON");
				send(reply);
				return true;
			}
		}
		if (true)
			std::cout << "client " << (unsigned long)this  << " pid " << getpid()<< " rx " << std::string(buf, buf + r) << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code() == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket input would block" << std::endl;
			return true;
		} else {
			if (true)
				std::cerr << "client " << (unsigned long)this << " pid " << getpid() << " socket input error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (m_longpoll)
		m_longpoll->unset_longpoll(get_ptr());
	m_longpoll.reset();
	m_logfilter.clear();
	m_logdiscard.clear();
	{
		std::string sess;
		if (root.isMember("session"))
                        sess = root.get("session", "").asString();
		if (sess.empty()) {
			reply["error"] = std::string("no session");
			send(reply);
			return true;
		}
		reply["session"] = sess;
		if (!m_server) {
			reply["error"] = std::string("no server");
			send(reply);
			return true;
		}
		if (root.isMember("quit")) {
			const Json::Value& q(root["quit"]);
			if (q.isBool() && q.asBool()) {
				reply["quit"] = m_server->stop_router(sess);
				send(reply);
				return true;
			}
		}
		m_closeprotect = 1;
		m_longpoll = m_server->find_router(sess);
		if (m_closeprotect == 2) {
			m_closeprotect = 0;
			if (m_server)
				m_server->remove_client(get_ptr());
			return false;
		}
		m_closeprotect = 0;
	}
	if (!m_longpoll) {
		reply["error"] = std::string("server busy");
		send(reply);
		return true;
	}
	if (root.isMember("cmds")) {
		const Json::Value& cmds(root["cmds"]);
		if (cmds.isObject()) {
			m_longpoll->send(cmds);
		} else if (cmds.isArray()) {
			for (Json::ArrayIndex i = 0; i < cmds.size(); ++i) {
				const Json::Value& cmd(cmds[i]);
				if (!cmd.isObject())
					continue;
				m_longpoll->send(cmd);
			}
		}
	}
	if (root.isMember("noreply")) {
		const Json::Value& noreply(root["noreply"]);
		if (noreply.isBool() && noreply.asBool()) {
			m_longpoll.reset();
			send(reply);
			return true;
		}
	}
	if (root.isMember("logfilter")) {
		const Json::Value& filters(root["logfilter"]);
		if (filters.isArray()) {
			for (Json::ArrayIndex i = 0; i < filters.size(); ++i) {
				const Json::Value& filter(filters[i]);
				if (!filter.isString())
					continue;
				std::string s(filter.asString());
				if (s.empty())
					continue;
				m_logfilter.insert(s);
			}
		}
	}
	if (root.isMember("logdiscard")) {
		const Json::Value& filters(root["logdiscard"]);
		if (filters.isArray()) {
			for (Json::ArrayIndex i = 0; i < filters.size(); ++i) {
				const Json::Value& filter(filters[i]);
				if (!filter.isString())
					continue;
				std::string s(filter.asString());
				if (s.empty())
					continue;
				m_logdiscard.insert(s);
			}
		}
	}
	Json::Value& replycmds(reply["cmds"] = m_longpoll->receive(m_logfilter, m_logdiscard));
	if (replycmds.size())
		send(reply);
	else
		m_longpoll->set_longpoll(get_ptr());
	return true;
}

void SocketServer::Client::msgavailable(void)
{
	if (!m_longpoll)
		return;
	Json::Value reply;
	if (m_firstreply) {
		reply["version"] = VERSION;
		reply["provider"] = "cfmu";
	}
	m_longpoll->unset_longpoll(get_ptr());
	reply["session"] = m_longpoll->get_session();
	reply["cmds"] = m_longpoll->receive(m_logfilter, m_logdiscard);
	m_longpoll.reset();
	send(reply);
}

SocketServer::Router::Router(SocketServer *server, pid_t pid, const Glib::RefPtr<Gio::Socket>& sock, const std::string& sess)
	: m_server(server), m_socket(sock), m_session(sess), m_refcount(1), m_pid(pid)
{
	m_accesstime.assign_current_time();
	m_starttime = m_accesstime;
	if (true)
		std::cout << "router create " << (unsigned long)this << " pid " << getpid() << " session " << m_session << std::endl;
	if (!m_server || !m_socket || m_pid == (pid_t)-1) {
		m_server = 0;
		m_socket.reset();
		m_pid = -1;
		return;
	}
	m_connin = Glib::signal_io().connect(sigc::mem_fun(*this, &Router::on_input), m_socket->get_fd(), Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	m_connchildwatch = Glib::signal_child_watch().connect(sigc::mem_fun(*this, &Router::child_watch), m_pid);
	if (true)
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " started pid " << m_pid
			  << " fd " << m_socket->get_fd() << std::endl;
}

SocketServer::Router::~Router()
{
	if (true)
		std::cout << "router destroy " << (unsigned long)this << " session " << m_session << std::endl;
	close();
	m_connchildwatch.disconnect();
	if (m_pid != (pid_t)-1) {
		Glib::signal_child_watch().connect(sigc::ptr_fun(&Router::child_reaper), m_pid);
		m_pid = -1;
	}
}

void SocketServer::Router::reference(void) const
{
        g_atomic_int_inc(&m_refcount);
	if (false)
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " reference " << m_refcount << std::endl;
}

void SocketServer::Router::unreference(void) const
{
        bool nz(!g_atomic_int_dec_and_test(&m_refcount));
	if (false)
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " unreference " << m_refcount << std::endl;
	if (nz)
                return;
        delete this;
}

bool SocketServer::Router::on_output(Glib::IOCondition iocond)
{
	m_connout.disconnect();
	if (m_sendqueue.empty())
		return true;
	Json::FastWriter writer;
	std::string msg(writer.write(*m_sendqueue.begin()));
	try {
		gssize r(m_socket->send(msg.c_str(), msg.size()));
		if (r == -1) {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket output return value: " << r << std::endl;
			close();
			return true;
		}
		m_sendqueue.erase(m_sendqueue.begin());
		if (true)
			std::cout << "router " << (unsigned long)this << " pid " << getpid() << " tx " << msg << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code () == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket output would block" << std::endl;
		} else {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket output error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	if (!m_sendqueue.empty())
		m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Router::on_output), m_socket->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
	return true;
}

bool SocketServer::Router::on_input(Glib::IOCondition iocond)
{
	if (!m_socket) {
		m_connin.disconnect();
		return true;
	}
	if (!m_server)
		return false;
	if (m_server->is_sockclient()) {
		if (true)
			std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " receive in socket client" << std::endl;
		close();
		return false;
	}
	Json::Value root;
	try {
		char buf[65536];
 		gssize r(m_socket->receive(buf, sizeof(buf)));
		if (r <= 0) {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			close();
			return true;
		}
		if (r >= sizeof(buf)) {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket input return value: " << r << std::endl;
			close();
			return true;
		}
		{
			Json::Reader reader;
			if (!reader.parse(buf, buf + r, root, false)) {
				if (true)
					std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " cannot parse JSON: " << std::string(buf, buf + r);
				close();
				return true;
			}
		}
		if (true)
			std::cout << "router " << (unsigned long)this << " pid " << getpid() << " rx " << std::string(buf, buf + r) << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code () == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket input would block" << std::endl;
			return true;
		} else {
			if (true)
				std::cerr << "router " << (unsigned long)this << " pid " << getpid() << " socket input error: " << error.what() << std::endl;
			close();
			return true;
		}
	}
	std::string cmdname(root.get("cmdname", "").asString());
	if (cmdname == "log")
		m_logqueue.push_back(root);
	else
		m_recvqueue.push_back(root);
	notify_longpoll();
	return true;
}

void SocketServer::Router::send(const Json::Value& v)
{
	if (!m_socket)
		return;
	m_accesstime.assign_current_time();
	m_sendqueue.push_back(v);
	if (false) {
		Json::FastWriter writer;
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " txqueue " << writer.write(*m_sendqueue.end()) << std::endl;
	}
	m_connout.disconnect();
	m_connout = Glib::signal_io().connect(sigc::mem_fun(*this, &Router::on_output), m_socket->get_fd(), Glib::IO_OUT | Glib::IO_ERR);
}

Json::Value SocketServer::Router::receive(const stringset_t& logfilter, const stringset_t& logdiscard)
{
	Json::Value v;
	unsigned int cnt(0);
	for (std::list<Json::Value>::iterator i(m_recvqueue.begin()), e(m_recvqueue.end()); i != e; ) {
		std::list<Json::Value>::iterator ii(i);
		++i;
		v.append(*ii);
		m_recvqueue.erase(ii);
		++cnt;
		if (cnt >= 128)
			break;
	}
	for (std::list<Json::Value>::iterator i(m_logqueue.begin()), e(m_logqueue.end()); i != e; ) {
		std::list<Json::Value>::iterator ii(i);
		++i;
		std::string item(ii->get("item", "").asString());
		if (!item.empty()) {
			if (logdiscard.find(item) != logdiscard.end()) {
				m_recvqueue.erase(ii);
				continue;
			}
			if (logfilter.find(item) != logfilter.end())
				continue;
		}
		if (cnt < 128) {
			v.append(*ii);
			m_recvqueue.erase(ii);
		} else if (logdiscard.empty()) {
			break;
		}
	}
	return v;
}

void SocketServer::Router::close(void)
{
	if (true) {
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " close pid " << m_pid;
		if (m_socket)
			std::cout << " fd " << m_socket->get_fd();
		std::cout << std::endl;
	}
	m_connin.disconnect();
	m_connout.disconnect();
	m_sendqueue.clear();
	m_recvqueue.clear();
	m_logqueue.clear();
	m_socket.reset();
	if (m_pid != (pid_t)-1)
		kill(m_pid, SIGHUP);
	notify_longpoll();
}

void SocketServer::Router::zap(void)
{
	if (true) {
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " zap pid " << m_pid;
		if (m_socket)
			std::cout << " fd " << m_socket->get_fd();
		std::cout << std::endl;
	}
	m_connchildwatch.disconnect();
	m_connin.disconnect();
	m_connout.disconnect();
	m_sendqueue.clear();
	m_recvqueue.clear();
	m_logqueue.clear();
	m_pid = -1;
	m_socket.reset();
	notify_longpoll();
}

void SocketServer::Router::child_watch(GPid pid, int child_status)
{
	if (pid != m_pid || pid == (pid_t)-1) {
		Glib::spawn_close_pid(pid);
		return;
	}
	m_connchildwatch.disconnect();
	Glib::spawn_close_pid(pid);
	if (true)
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " reaped pid " << m_pid
			  << " status " << child_status << std::endl;
        m_pid = -1;
	notify_longpoll();
	if (m_server)
		m_server->reap_died_routers();
}

void SocketServer::Router::child_reaper(GPid pid, int child_status)
{
	Glib::spawn_close_pid(pid);
}

bool SocketServer::Router::handle_timeout(const Glib::TimeVal& tvacc, const Glib::TimeVal& tvstart)
{
	if (!m_socket || (tvacc < m_accesstime && tvstart < m_starttime))
		return false;
	close();
	return true;
}

void SocketServer::Router::set_longpoll(const Client::ptr_t& clnt)
{
	if (m_longpoll == clnt)
		return;
	notify_longpoll();
	m_longpoll = clnt;
	if (true)
		std::cout << "router " << (unsigned long)this << " pid " << getpid() << " set longpoll client "
			  << (unsigned long)m_longpoll.operator->() << std::endl;
}

void SocketServer::Router::unset_longpoll(const Client::ptr_t& clnt)
{
	if (m_longpoll == clnt) {
		if (true)
			std::cout << "router " << (unsigned long)this << " pid " << getpid() << " unset longpoll client "
				  << (unsigned long)m_longpoll.operator->() << std::endl;
		m_longpoll.reset();
	}
}

void SocketServer::Router::notify_longpoll(void)
{
	if (m_longpoll) {
		if (true)
			std::cout << "router " << (unsigned long)this << " pid " << getpid() << " notify client "
				  << (unsigned long)m_longpoll.operator->() << std::endl;
		m_longpoll->msgavailable();
	}
	m_longpoll.reset();
}

SocketServer::SocketServer(CFMUAutoroute *autoroute, const Glib::RefPtr<Glib::MainLoop>& mainloop,
			   unsigned int connlimit, unsigned int timeout, unsigned int maxruntime, int xdisplay)
	: m_autoroute(autoroute), m_mainloop(mainloop), m_connectlimit(connlimit),
	  m_timeout(timeout), m_maxruntime(maxruntime), m_xdisplay(xdisplay), m_x11srvrun(false)
{
	m_cmdlist["nop"] = &SocketServer::cmd_nop;
	m_cmdlist["quit"] = &SocketServer::cmd_quit;
	m_cmdlist["departure"] = &SocketServer::cmd_departure;
	m_cmdlist["destination"] = &SocketServer::cmd_destination;
	m_cmdlist["crossing"] = &SocketServer::cmd_crossing;
	m_cmdlist["enroute"] = &SocketServer::cmd_enroute;
	m_cmdlist["levels"] = &SocketServer::cmd_levels;
	m_cmdlist["exclude"] = &SocketServer::cmd_exclude;
	m_cmdlist["tfr"] = &SocketServer::cmd_tfr;
	m_cmdlist["atmosphere"] = &SocketServer::cmd_atmosphere;
	m_cmdlist["cruise"] = &SocketServer::cmd_cruise;
	m_cmdlist["optimization"] = &SocketServer::cmd_optimization;
	m_cmdlist["preferred"] = &SocketServer::cmd_preferred;
	m_cmdlist["aircraft"] = &SocketServer::cmd_aircraft;
	m_cmdlist["start"] = &SocketServer::cmd_start;
	m_cmdlist["stop"] = &SocketServer::cmd_stop;
	m_cmdlist["continue"] = &SocketServer::cmd_continue;
	m_cmdlist["preload"] = &SocketServer::cmd_preload;
	m_cmdlist["clear"] = &SocketServer::cmd_clear;
}

SocketServer::~SocketServer()
{
	m_rtrtimeoutconn.disconnect();
	m_clients.clear();
	m_routers.clear();
	remove_listensockaddr();
	stop_x11();
}

void SocketServer::add_client(const Client::ptr_t& p)
{
	if (p)
		m_clients.insert(p);
}

void SocketServer::remove_client(const Client::ptr_t& p)
{
	clients_t::iterator i(m_clients.find(p));
	if (i != m_clients.end())
		m_clients.erase(i);
}

void SocketServer::remove_listensockaddr(void)
{
	if (!m_listensockaddr)
		return;
	if (m_listensockaddr->property_address_type() == Gio::UNIX_SOCKET_ADDRESS_PATH) {
		Glib::RefPtr<Gio::File> f(Gio::File::create_for_path(m_listensockaddr->property_path()));
		try {
			f->remove();
		} catch (...) {
		}
	}
	m_listensockaddr.reset();
}

void SocketServer::listen(const std::string& path, uid_t socketuid, gid_t socketgid, mode_t socketmode)
{
	m_listenservice.reset();
	m_listensock.reset();
	remove_listensockaddr();
#ifdef HAVE_SYSTEMDDAEMON
	{
		int n = sd_listen_fds(0);
		if (n > 1)
			throw std::runtime_error("SystemD: too many file descriptors received");
		if (n == 1) {
			int fd = SD_LISTEN_FDS_START + 0;
			m_listensock = Gio::Socket::create_from_fd(fd);
		}
	}
#endif
	if (!m_listensock) {
		m_listensock = Gio::Socket::create(Gio::SOCKET_FAMILY_UNIX, Gio::SOCKET_TYPE_SEQPACKET, Gio::SOCKET_PROTOCOL_DEFAULT);
	        m_listensockaddr = Gio::UnixSocketAddress::create(path, Gio::UNIX_SOCKET_ADDRESS_PATH);
		if (true) {
			Glib::RefPtr<Gio::File> f(Gio::File::create_for_path(path));
			try {
				f->remove();
			} catch (...) {
			}
		}
		m_listensock->bind(m_listensockaddr, true);
		m_listensock->listen();
		//if (fchown(m_listensock->get_fd(), socketuid, socketgid))
		if (chown(path.c_str(), socketuid, socketgid))
			std::cerr << "Cannot set socket " << path << " uid/gid to " << socketuid << '/' << socketgid
				  << ": " << strerror(errno) << " (" << errno << ')' << std::endl;
		//if (fchmod(m_listensock->get_fd(), socketmode))
		if (chmod(path.c_str(), socketmode))
			std::cerr << "Cannot set socket " << path << " mode to " << std::oct << '0'
				  << (unsigned int)socketmode << std::dec << ": " << strerror(errno) << " (" << errno << ')' << std::endl;
	}
	m_listenservice = Gio::SocketService::create();
	m_listenservice->signal_incoming().connect(sigc::mem_fun(*this, &SocketServer::on_connect));
	if (!m_listenservice->add_socket(m_listensock))
		throw std::runtime_error("Cannot listen to socket");
	m_listenservice->start();
}

bool SocketServer::on_connect(const Glib::RefPtr<Gio::SocketConnection>& connection, const Glib::RefPtr<Glib::Object>& source_object)
{
	Client::ptr_t p(new Client(this, connection));
	return true;
}

SocketServer::Router::ptr_t SocketServer::find_router(const std::string& session)
{
	if (session.empty())
		return Router::ptr_t();
	if (true) {
		std::cout << "find_router: pid " << getpid() << std::endl;
		for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e; ++i)
			std::cout << "find_router " << (unsigned long)i->second.operator->() << " session " << i->first << std::endl;
	}
	routers_t::iterator i(m_routers.find(session));
	if (i != m_routers.end()) {
		if (!i->second->is_dead())
			return i->second;
		if (true)
			std::cout << "router " << (unsigned long)i->second.operator->() << " session " << i->first << " remove" << std::endl;
		m_routers.erase(i);
	}
	if (m_routers.size() >= m_connectlimit)
		return Router::ptr_t();
	start_x11();
	int socks[2];
	if (socketpair(AF_LOCAL, SOCK_SEQPACKET, 0, socks))
		return Router::ptr_t();
	pid_t pid = fork();
	if (!pid) {
		// child
		close(socks[0]);
		m_listenservice->stop();
		m_listensockaddr.reset();
		m_listenservice.reset();
		m_listensock.reset();
		if (true)
			std::cout << "New session " << session << " kill clients" << std::endl;
		for (clients_t::iterator i(m_clients.begin()), e(m_clients.end()); i != e; ) {
			clients_t::iterator i2(i);
			++i;
			(*i2)->close();
		}
		if (true)
			std::cout << "New session " << session << " kill routers" << std::endl;
		while (!m_routers.empty()) {
			m_routers.begin()->second->zap();
			m_routers.erase(m_routers.begin());
		}
		if (true)
			std::cout << "New session " << session << " start sockclient" << std::endl;
		sockclient(socks[1]);
		return Router::ptr_t();
	}
	close(socks[1]);
	if (pid == -1) {
		close(socks[0]);
		return Router::ptr_t();
	}
	if (m_routers.empty())
		m_rtrtimeoutconn = Glib::signal_timeout().connect_seconds(sigc::mem_fun(*this, &SocketServer::router_timeout_handler), std::min(std::max(m_timeout / 10U, 1U), 10U));
	Router::ptr_t p(new Router(this, pid, Gio::Socket::create_from_fd(socks[0]), session));
	m_routers[session] = p;
	if (true) {
		std::cout << "find_router: after creation: pid " << getpid() << std::endl;
		for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e; ++i)
			std::cout << "find_router " << (unsigned long)i->second.operator->() << " session " << i->first << std::endl;
	}
	return p;
}

bool SocketServer::stop_router(const std::string& session)
{
	if (session.empty())
		return Router::ptr_t();
	routers_t::iterator i(m_routers.find(session));
	if (i == m_routers.end())
		return false;
	if (i->second->is_dead()) {
		if (true)
			std::cout << "router " << (unsigned long)i->second.operator->() << " session " << i->first << " remove" << std::endl;
		m_routers.erase(i);
		return false;
	}
	i->second->close();
	return true;
}

void SocketServer::reap_died_routers(void)
{
	for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e;) {
		routers_t::iterator ii(i);
		++i;
		if (!ii->second->is_dead())
			continue;
		if (true)
			std::cout << "router " << (unsigned long)ii->second.operator->() << " session " << ii->first << " remove" << std::endl;
		m_routers.erase(ii);
	}
	if (m_routers.empty())
		m_rtrtimeoutconn.disconnect();
}

bool SocketServer::router_timeout_handler(void)
{
	Glib::TimeVal tvacc, tvrun;
	tvacc.assign_current_time();
	tvrun = tvacc;
	tvacc.subtract_seconds(m_timeout);
	tvrun.subtract_seconds(m_maxruntime);
	for (routers_t::iterator i(m_routers.begin()), e(m_routers.end()); i != e; ++i)
		i->second->handle_timeout(tvacc, tvrun);
	return true;
}

void SocketServer::start_x11(void)
{
	if (m_x11srvrun)
		return;
	if (m_xdisplay < 0)
		return;
	try {
		std::vector<std::string> argv, env;
#ifdef __MINGW32__
		std::string workdir;
		{
			gchar *instdir(g_win32_get_package_installation_directory_of_module(0));
			if (instdir) {
				workdir = instdir;
				argv.push_back(Glib::build_filename(instdir, "bin", "Xvfb.exe"));
			} else {
				argv.push_back(Glib::build_filename(PACKAGE_PREFIX_DIR "/bin", "Xvfb.exe"));
				workdir = PACKAGE_PREFIX_DIR;
			}
		}
#else
		std::string workdir("/tmp");
		argv.push_back("/usr/bin/Xvfb");
		env.push_back("PATH=/bin:/usr/bin");
		{
			std::ostringstream disp;
			disp << ':' << m_xdisplay;
			argv.push_back(disp.str());
		}
		argv.push_back("-screen");
		argv.push_back("0");
		argv.push_back("100x100x16");
#endif
		if (false) {
			std::cerr << "Spawn: workdir " << workdir << " argv:" << std::endl;
			for (std::vector<std::string>::const_iterator i(argv.begin()), e(argv.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
			std::cerr << "env:" << std::endl;
			for (std::vector<std::string>::const_iterator i(env.begin()), e(env.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
		}
#ifdef __MINGW32__
		// inherit environment; otherwise, this may fail on WindowsXP with STATUS_SXS_ASSEMBLY_NOT_FOUND (C0150004)
		Glib::spawn_async_with_pipes(workdir, argv,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_x11srvpid, 0, 0, 0);
#else
		Glib::spawn_async_with_pipes(workdir, argv, env,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_x11srvpid, 0, 0, 0);
#endif
		m_x11srvrun = true;
	} catch (Glib::SpawnError& e) {
		m_x11srvrun = false;
		std::cerr << "Cannot run X Server: " + e.what() << std::endl;
		return;
	}
	m_x11srvconnwatch = Glib::signal_child_watch().connect(sigc::mem_fun(this, &SocketServer::x11srv_child_watch), m_x11srvpid);
}

void SocketServer::x11srv_child_watch(GPid pid, int child_status)
{
	if (!m_x11srvrun)
		return;
	if (m_x11srvpid != pid) {
		Glib::spawn_close_pid(pid);
		return;
	}
	Glib::spawn_close_pid(pid);
	m_x11srvrun = false;
	m_x11srvconnwatch.disconnect();
}

void SocketServer::stop_x11(void)
{
	if (!m_x11srvrun)
		return;
	kill(m_x11srvpid, SIGHUP);
}

void SocketServer::sockclient(int fd)
{
	if (!m_autoroute)
		return;
	m_listensock = Gio::Socket::create_from_fd(fd);
        m_listensock->property_blocking() = true;
	Glib::signal_io().connect(sigc::mem_fun(*this, &SocketServer::sockcli_input), m_listensock->get_fd(), Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	m_autoroute->signal_statuschange().connect(sigc::mem_fun(*this, &SocketServer::sockcli_statuschange));
	m_autoroute->signal_log().connect(sigc::mem_fun(*this, &SocketServer::sockcli_autoroutelog));
	if (m_xdisplay >= 0) {
		std::ostringstream disp;
		disp << ':' << m_xdisplay;
		Glib::setenv("DISPLAY", disp.str(), true);
	}
}

void SocketServer::sockcli_close(void)
{
	m_listensock.reset();
	if (m_mainloop)
		m_mainloop->quit();
}

void SocketServer::sockcli_send(const Json::Value& v)
{
	Json::FastWriter writer;
	std::string msg(writer.write(v));
	try {
		gssize r(m_listensock->send(msg.c_str(), msg.size()));
		if (r == -1) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket output return value: " << r << std::endl;
			sockcli_close();
			return;
		}
		if (false)
			std::cerr << "sockcli " << getpid() << " tx " << msg << std::endl;
	} catch (const Gio::Error& error) {
		if (true)
			std::cerr << "sockcli " << getpid() << " socket output error: " << error.what() << std::endl;
		sockcli_close();
		return;
	}
}

bool SocketServer::sockcli_input(Glib::IOCondition iocond)
{
	Json::Value root;
	Json::Value reply;
	try {
		char buf[16384];
 		gssize r(m_listensock->receive(buf, sizeof(buf)));
		if (r <= 0) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input return value: " << r << std::endl;
			sockcli_close();
			return true;
		}
		if (r >= sizeof(buf)) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input return value: " << r << std::endl;
			reply["error"] = std::string("invalid message size");
			sockcli_send(reply);
			return true;
		}
		{
			Json::Reader reader;
			if (!reader.parse(buf, buf + r, root, false)) {
				if (true)
					std::cerr << "sockcli " << getpid() << " cannot parse JSON: " << std::string(buf, buf + r);
				reply["error"] = std::string("cannot parse JSON");
				sockcli_send(reply);
				return true;
			}
		}
		if (false)
			std::cout << "sockcli " << getpid() << " rx " << std::string(buf, buf + r) << std::endl;
	} catch (const Gio::Error& error) {
		if (error.code () == Gio::Error::WOULD_BLOCK) {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input would block" << std::endl;
			return true;
		} else {
			if (true)
				std::cerr << "sockcli " << getpid() << " socket input error: " << error.what() << std::endl;
			sockcli_close();
			return true;
		}
	}
	try {
		if (root.isMember("cmdseq"))
			reply["cmdseq"] = root["cmdseq"];
		std::string cmdname(root.get("cmdname", "").asString());
		reply["cmdname"] = cmdname.empty() ? "autoroute" : cmdname;
		Json::Value reply1(reply);
		cmdlist_t::const_iterator cmdi(m_cmdlist.find(cmdname));
                if (cmdi == m_cmdlist.end()) {
			reply1["error"] = "command not found";
                } else {
                        (this->*(cmdi->second))(root, reply1);
                }
                sockcli_send(reply1);
        } catch (const std::runtime_error& e) {
		reply["error"] = e.what();
                sockcli_send(reply);
        } catch (const Glib::Exception& e) {
		reply["error"] = (std::string)e.what();
                sockcli_send(reply);
        }
	return true;
}

void SocketServer::sockcli_statuschange(CFMUAutoroute::statusmask_t status)
{
        if (status & (CFMUAutoroute::statusmask_starting |
                      CFMUAutoroute::statusmask_stoppingdone |
                      CFMUAutoroute::statusmask_stoppingerror)) {
		Json::Value reply;
		reply["cmdname"] = "autoroute";
                if (status & CFMUAutoroute::statusmask_starting) {
                        reply["status"] =  "starting";
                } else if (status & (CFMUAutoroute::statusmask_stoppingdone |
                                     CFMUAutoroute::statusmask_stoppingerror)) {
                        reply["status"] = "stopping";
                        reply["routesuccess"] = !!(status & CFMUAutoroute::statusmask_stoppingdone);
                }
                sockcli_send(reply);
        }
}

void SocketServer::sockcli_autoroutelog(CFMUAutoroute::log_t item, const Glib::ustring& line)
{
	Json::Value reply;
        if (item != CFMUAutoroute::log_fplproposal) {
	        reply["cmdname"] = "log";
                reply["item"] = (std::string)to_str(item);
                reply["text"] = (std::string)line;
                sockcli_send(reply);
                return;
        }
        // new flight plan
	reply["cmdname"] = "fpl";
	if (m_autoroute) {
		const FPlanRoute& route(m_autoroute->get_route());
		reply["note"] = (std::string)route.get_note();
		reply["curwpt"] = route.get_curwpt();
		reply["offblock"] = (Json::LargestInt)route.get_time_offblock_unix();
		reply["save_offblock"] = (Json::LargestInt)route.get_save_time_offblock_unix();
		reply["onblock"] = (Json::LargestInt)route.get_time_onblock_unix();
		reply["save_onblock"] = (Json::LargestInt)route.get_save_time_onblock_unix();
		{
			Json::Value& fplan(reply["fplan"]);
			for (unsigned int i = 0; i < route.get_nrwpt(); ++i) {
				const FPlanWaypoint& wpt(route[i]);
				Json::Value fplwpt;
				fplwpt["icao"] = (std::string)wpt.get_icao();
				fplwpt["name"] = (std::string)wpt.get_name();
				fplwpt["pathname"] = (std::string)wpt.get_pathname();
				fplwpt["pathcode"] = (std::string)wpt.get_pathcode_name();
				fplwpt["note"] = (std::string)wpt.get_note();
				fplwpt["time"] = (Json::LargestInt)wpt.get_time_unix();
				fplwpt["save_time"] = (Json::LargestInt)wpt.get_save_time_unix();
				fplwpt["frequency"] = wpt.get_frequency();
				set_point(fplwpt, "coord", wpt.get_coord());
				fplwpt["altitude"] = wpt.get_altitude();
				fplwpt["flags"] = wpt.get_flags();
				fplwpt["winddir"] = wpt.get_winddir_deg();
				fplwpt["windspeed"] = wpt.get_windspeed_kts();
				fplwpt["qff"] = wpt.get_qff_hpa();
				fplwpt["isaoffset"] = wpt.get_isaoffset_kelvin();
				fplan.append(fplwpt);
			}
		}
		reply["gcdist"] = m_autoroute->get_gcdistance();
		reply["routedist"] = m_autoroute->get_routedistance();
		reply["mintime"] = m_autoroute->get_mintime();
		reply["routetime"] = m_autoroute->get_routetime();
		reply["minfuel"] = m_autoroute->get_minfuel();
		reply["routefuel"] = m_autoroute->get_routefuel();
		reply["fpl"] = (std::string)m_autoroute->get_simplefpl();
		reply["iteration"] = m_autoroute->get_iterationcount();
		reply["localiteration"] = m_autoroute->get_local_iterationcount();
		reply["remoteiteration"] = m_autoroute->get_remote_iterationcount();
	} else {
		reply["error"] = "no autoroute instance";
	}
	sockcli_send(reply);
}

Point SocketServer::get_point(const Json::Value& cmd, const std::string& name)
{
	Point pt;
	pt.set_invalid();
	if (!cmd.isMember(name))
		return pt;
	std::string s(cmd.get(name, "").asString());
	std::string lat, lon;
	if (s.empty()) {
		{
			const Json::Value& lat(cmd[name + "lat"]);
			const Json::Value& lon(cmd[name + "lon"]);
			if (lat.isIntegral() && lon.isIntegral()) {
				pt.set_lat(lat.asInt());
				pt.set_lon(lon.asInt());
				return pt;
			}
		}
		{
			const Json::Value& lat(cmd[name + "latdeg"]);
			const Json::Value& lon(cmd[name + "londeg"]);
			if (lat.isNumeric() && lon.isNumeric()) {
				pt.set_lat_deg_dbl(lat.asDouble());
				pt.set_lon_deg_dbl(lon.asDouble());
				return pt;
			}
		}
		lat = cmd.get(name + "latstr", "").asString();
		lon = cmd.get(name + "lonstr", "").asString();
		unsigned int x(pt.set_str(lat) | pt.set_str(lon));
		if (!(x & ~(Point::setstr_lat | Point::setstr_lon)))
			return pt;
		pt.set_invalid();
		return pt;
	} else {
		std::string::size_type n(s.find(','));
		if (n == std::string::npos)
			return pt;
		lat = s.substr(0, n);
		lon = s.substr(n + 1);
	}
	if (lat.empty() || lon.empty())
		return pt;
	char *cp1(0), *cp2(0);
	pt.set_lat(strtol(lat.c_str(), &cp1, 10));
	pt.set_lon(strtol(lon.c_str(), &cp2, 10));
	if (*cp1 || *cp2)
		pt.set_invalid();
	return pt;
}

void SocketServer::set_point(Json::Value& cmd, const std::string& name, const Point& pt)
{
	std::ostringstream oss;
	if (!pt.is_invalid()) {
		oss << pt.get_lat() << ',' << pt.get_lon();
		cmd[name + "lat"] = pt.get_lat();
		cmd[name + "lon"] = pt.get_lon();
		cmd[name + "latdeg"] = pt.get_lat_deg_dbl();
		cmd[name + "londeg"] = pt.get_lon_deg_dbl();
		cmd[name + "latstr"] = (std::string)pt.get_lat_str2();
		cmd[name + "lonstr"] = (std::string)pt.get_lon_str2();
	}
	cmd[name] = oss.str();
}

void SocketServer::cmd_nop(const Json::Value& cmdin, Json::Value& cmdout)
{
}

void SocketServer::cmd_quit(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (m_mainloop)
		m_mainloop->quit();
}

void SocketServer::cmd_departure(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("icao") || cmdin.isMember("name")) {
		std::string icao(cmdin.get("icao", "").asString());
		std::string name(cmdin.get("name", "").asString());
		if (!m_autoroute->set_departure(icao, name))
			throw std::runtime_error("departure aerodrome " + icao + " " + name + " not found");
	}
	if (cmdin.isMember("ifr")) {
		const Json::Value& x(cmdin["ifr"]);
		m_autoroute->set_departure_ifr(!x.isBool() || x.asBool());
	}
	if (cmdin.isMember("vfr")) {
		const Json::Value& x(cmdin["vfr"]);
		m_autoroute->set_departure_ifr(x.isBool() && !x.asBool());
	}
	{
		Point pt(get_point(cmdin, "sid"));
		if (pt.is_invalid()) {
			if (cmdin.isMember("sidident"))
				m_autoroute->set_sid(cmdin.get("sidident", "").asString());
		} else {
			m_autoroute->set_sid(pt);
		}
	}
	if (cmdin.isMember("sidlimit"))
		m_autoroute->set_sidlimit(cmdin["sidlimit"].asDouble());
	if (cmdin.isMember("siddb"))
		m_autoroute->set_siddb(cmdin["siddb"].asBool());
	if (cmdin.isMember("time"))
		m_autoroute->set_deptime(cmdin["time"].asUInt());
	if (cmdin.isMember("date")) {
		Glib::TimeVal tv;
		if (tv.assign_from_iso8601(cmdin["date"].asString()))
			m_autoroute->set_deptime(tv.tv_sec);
	}
	if (m_autoroute->get_departure().is_valid()) {
		cmdout["icao"] = (std::string)m_autoroute->get_departure().get_icao();
		cmdout["name"] = (std::string)m_autoroute->get_departure().get_name();
		set_point(cmdout, "coord", m_autoroute->get_departure().get_coord());
	}
	if (m_autoroute->get_sid().is_invalid()) {
		cmdout.removeMember("sid");
	} else {
	        set_point(cmdout, "sid", m_autoroute->get_sid());
		cmdout["sidtype"] = to_str(m_autoroute->get_sidtype());
		cmdout["sidident"] = m_autoroute->get_sidident();
	}
	cmdout["sidlimit"] = m_autoroute->get_sidlimit();
	cmdout["siddb"] = m_autoroute->get_siddb();
	if (m_autoroute->get_departure_ifr()) {		
		cmdout["ifr"] = true;
		cmdout.removeMember("vfr");
	} else {		
		cmdout["vfr"] = true;
		cmdout.removeMember("ifr");
	}
	cmdout["time"] = (Json::LargestInt)m_autoroute->get_deptime();
	{
		Glib::TimeVal tv(m_autoroute->get_deptime(), 0);
		cmdout["date"] = (std::string)tv.as_iso8601();
	}
}

void SocketServer::cmd_destination(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("icao") || cmdin.isMember("name")) {
		std::string icao(cmdin.get("icao", "").asString());
		std::string name(cmdin.get("name", "").asString());
		if (!m_autoroute->set_destination(icao, name))
			throw std::runtime_error("destination aerodrome " + icao + " " + name + " not found");
	}
	if (cmdin.isMember("ifr")) {
		const Json::Value& x(cmdin["ifr"]);
		m_autoroute->set_destination_ifr(!x.isBool() || x.asBool());
	}
	if (cmdin.isMember("vfr")) {
		const Json::Value& x(cmdin["vfr"]);
		m_autoroute->set_destination_ifr(x.isBool() && !x.asBool());
	}
	{
		Point pt(get_point(cmdin, "star"));
		if (pt.is_invalid()) {
			if (cmdin.isMember("starident"))
				m_autoroute->set_star(cmdin.get("starident", "").asString());
		} else {
			m_autoroute->set_star(pt);
		}
	}
	if (cmdin.isMember("starlimit"))
		m_autoroute->set_starlimit(cmdin["starlimit"].asDouble());
	if (cmdin.isMember("stardb"))
		m_autoroute->set_stardb(cmdin["stardb"].asBool());
	if (m_autoroute->get_destination().is_valid()) {
		cmdout["icao"] = (std::string)m_autoroute->get_destination().get_icao();
		cmdout["name"] = (std::string)m_autoroute->get_destination().get_name();
		set_point(cmdout, "coord", m_autoroute->get_destination().get_coord());
	}
	if (m_autoroute->get_star().is_invalid()) {
		cmdout.removeMember("star");
	} else {
	        set_point(cmdout, "star", m_autoroute->get_star());
		cmdout["startype"] = to_str(m_autoroute->get_startype());
		cmdout["starident"] = m_autoroute->get_starident();
	}
	cmdout["starlimit"] = m_autoroute->get_starlimit();
	cmdout["stardb"] = m_autoroute->get_stardb();
	if (m_autoroute->get_destination_ifr()) {		
		cmdout["ifr"] = true;
		cmdout.removeMember("vfr");
	} else {		
		cmdout["vfr"] = true;
		cmdout.removeMember("ifr");
	}
}

void SocketServer::cmd_crossing(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("crossing")) {
		const Json::Value& crossing(cmdin["crossing"]);
		if (crossing.isArray()) {
			unsigned int index = 0;
			m_autoroute->set_crossing_size(0);
			for (unsigned int i = 0; i < crossing.size(); ++i) {
				const Json::Value& el(crossing[i]);
				if (!el.isObject())
					continue;
				Point pt(get_point(el, "star"));
				if (pt.is_invalid()) {
					if (el.isMember("ident"))
						if (m_autoroute->set_crossing(index, el.get("ident", "").asString()))
							++index;
				} else {
					m_autoroute->set_crossing(index, pt);
					++index;
				}
				if (el.isMember("radius"))
					m_autoroute->set_crossing_radius(index, el["radius"].asDouble());
				if (el.isMember("minlevel") || el.isMember("maxlevel")) {
					int minlevel(m_autoroute->get_crossing_minlevel(index));
					int maxlevel(m_autoroute->get_crossing_maxlevel(index));
					if (el.isMember("minlevel"))
						minlevel = el["minlevel"].asInt();
					if (el.isMember("maxlevel"))
						maxlevel = el["maxlevel"].asInt();
					m_autoroute->set_crossing_level(index, minlevel, maxlevel);
				}
			}
		}
	}
	Json::Value crossings;
	for (unsigned int i = 0; i < m_autoroute->get_crossing_size(); ++i) {
		Json::Value crossing;
		set_point(crossing, "coord", m_autoroute->get_crossing(i));
		crossing["type"] = to_str(m_autoroute->get_crossing_type(i));
		crossing["ident"] = m_autoroute->get_crossing_ident(i);
		crossing["radius"] = m_autoroute->get_crossing_radius(i);
		crossing["minlevel"] = m_autoroute->get_crossing_minlevel(i);
		crossing["maxlevel"] = m_autoroute->get_crossing_maxlevel(i);
		crossings.append(crossing);
	}
	cmdout["crossing"] = crossings;
}

void SocketServer::cmd_enroute(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("dctlimit"))
		m_autoroute->set_dctlimit(cmdin["dctlimit"].asDouble());
	if (cmdin.isMember("dctpenalty"))
		m_autoroute->set_dctpenalty(cmdin["dctpenalty"].asDouble());
	if (cmdin.isMember("vfraspclimit"))
		m_autoroute->set_vfraspclimit(cmdin["vfraspclimit"].asDouble());
	if (cmdin.isMember("forceenrouteifr"))
		m_autoroute->set_force_enroute_ifr(!!cmdin["forceenrouteifr"].asBool());
	if (cmdin.isMember("honourawylevels"))
		m_autoroute->set_honour_awy_levels(!!cmdin["honourawylevels"].asBool());
	cmdout["dctlimit"] = m_autoroute->get_dctlimit();
	cmdout["dctpenalty"] = m_autoroute->get_dctpenalty();
	cmdout["vfraspclimit"] = m_autoroute->get_vfraspclimit();
	cmdout["forceenrouteifr"] = !!m_autoroute->get_force_enroute_ifr();
	cmdout["honourawylevels"] = !!m_autoroute->get_honour_awy_levels();
}

void SocketServer::cmd_levels(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	{
		int b(m_autoroute->get_baselevel());
		int t(m_autoroute->get_toplevel());
		if (cmdin.isMember("base"))
			b = cmdin["base"].asInt();
		if (cmdin.isMember("top"))
			t = cmdin["top"].asInt();
		m_autoroute->set_levels(b, t);
	}
	cmdout["base"] = m_autoroute->get_baselevel();
	cmdout["top"] = m_autoroute->get_toplevel(); 
}

void SocketServer::cmd_exclude(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("clear")) {
		const Json::Value& v(cmdin["clear"]);
		if (!v.isBool() || v.asBool())
			m_autoroute->clear_excluderegions();
	}
	bool indexf(cmdin.isMember("index"));
	unsigned int indexv(0);
	if (indexf)
		indexv = cmdin["index"].asUInt();
	{
		int ob(0), ot(999);
		if (cmdin.isMember("base"))
			ob = cmdin["base"].asInt();
		if (cmdin.isMember("top"))
			ot = cmdin["top"].asInt();
		if (cmdin.isMember("aspcid")) {
			m_autoroute->add_excluderegion(CFMUAutoroute::ExcludeRegion(cmdin["aspcid"].asString(), cmdin.get("aspctype", "").asString(), ob, ot));
			if (!indexf) {
				indexv = m_autoroute->get_excluderegions().size() - 1;
				indexf = true;
			}
		}
		Point ptsw(get_point(cmdin, "sw"));
		Point ptne(get_point(cmdin, "ne"));
		if (!ptsw.is_invalid() && !ptne.is_invalid()) {
			m_autoroute->add_excluderegion(CFMUAutoroute::ExcludeRegion(Rect(ptsw, ptne), ob, ot));
			if (!indexf) {
				indexv = m_autoroute->get_excluderegions().size() - 1;
				indexf = true;
			}
		}
	}
	cmdout["count"] = (Json::LargestUInt)m_autoroute->get_excluderegions().size();
	if (indexf && indexv < m_autoroute->get_excluderegions().size()) {
		const CFMUAutoroute::ExcludeRegion& er(m_autoroute->get_excluderegions()[indexv]);
		cmdout["base"] = er.get_minlevel();
		cmdout["top"] = er.get_maxlevel();
		if (er.is_airspace()) {
			cmdout["aspcid"] = er.get_airspace_id();
			cmdout["aspctype"] = er.get_airspace_type();
		} else {
			set_point(cmdout, "sw", er.get_bbox().get_southwest());
			set_point(cmdout, "ne", er.get_bbox().get_northeast());
		}
	}
}

void SocketServer::cmd_tfr(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("enabled"))
		m_autoroute->set_tfr_enabled(!!cmdin["enabled"].asBool());
	if (cmdin.isMember("trace"))
		m_autoroute->set_tfr_trace(cmdin["trace"].asString());
	if (cmdin.isMember("disable"))
		m_autoroute->set_tfr_disable(cmdin["disable"].asString());
	if (cmdin.isMember("precompgraph"))
		m_autoroute->set_precompgraph_enabled(!!cmdin["precompgraph"].asBool());
	cmdout["enabled"] = !!m_autoroute->get_tfr_enabled();
	cmdout["available"] = !!m_autoroute->get_tfr_available();
	cmdout["trace"] = m_autoroute->get_tfr_trace();
	cmdout["disable"] = m_autoroute->get_tfr_disable();
	cmdout["precompgraph"] = !!m_autoroute->get_precompgraph_enabled();
}

void SocketServer::cmd_atmosphere(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("qnh"))
		m_autoroute->set_qnh(cmdin["qnh"].asDouble());
	if (cmdin.isMember("isa"))
		m_autoroute->set_isaoffs(cmdin["qnh"].asDouble());
	if (cmdin.isMember("wind"))
		m_autoroute->set_wind_enabled(!!cmdin["wind"].asBool());
	cmdout["qnh"] = m_autoroute->get_qnh();
	cmdout["isa"] = m_autoroute->get_isaoffs();
	cmdout["wind"] = !!m_autoroute->get_wind_enabled();
}

void SocketServer::cmd_cruise(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("rpm"))
		m_autoroute->set_engine_rpm(cmdin["rpm"].asDouble());
	if (cmdin.isMember("mp"))
		m_autoroute->set_engine_mp(cmdin["mp"].asDouble());
	if (cmdin.isMember("bhp"))
		m_autoroute->set_engine_bhp(cmdin["bhp"].asDouble());
	cmdout["rpm"] = m_autoroute->get_engine_rpm();
	cmdout["mp"] = m_autoroute->get_engine_mp();
	cmdout["bhp"] = m_autoroute->get_engine_bhp();
}

void SocketServer::cmd_optimization(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("target")) {
		std::string tgt(cmdin["target"].asString());
		static const CFMUAutoroute::opttarget_t ot[] = {
			CFMUAutoroute::opttarget_time,
			CFMUAutoroute::opttarget_fuel,
			CFMUAutoroute::opttarget_preferred
		};
		unsigned int i;
		for (i = 0; i < sizeof(ot)/sizeof(ot[0]); ++i)
			if (::to_str(ot[i]) == tgt)
				break;
		if (i >= sizeof(ot)/sizeof(ot[0])) {
			cmdout["error"] = "invalid target=" + tgt;
			return;
		}
		m_autoroute->set_opttarget(ot[i]);
	}
	cmdout["target"] = (std::string)to_str(m_autoroute->get_opttarget());
}

void SocketServer::cmd_preferred(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("level"))
		m_autoroute->set_preferred_level(cmdin["level"].asInt());
	if (cmdin.isMember("penalty"))
		m_autoroute->set_preferred_penalty(cmdin["penalty"].asDouble());
	if (cmdin.isMember("climb"))
		m_autoroute->set_preferred_climb(cmdin["climb"].asDouble());
	if (cmdin.isMember("descent"))
		m_autoroute->set_preferred_descent(cmdin["descent"].asDouble());
	cmdout["level"] = m_autoroute->get_preferred_level();
	cmdout["penalty"] = m_autoroute->get_preferred_penalty();
	cmdout["climb"] = m_autoroute->get_preferred_climb();
	cmdout["descent"] = m_autoroute->get_preferred_descent();
}

void SocketServer::cmd_aircraft(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	if (cmdin.isMember("file"))
		cmdout["status"] = !!m_autoroute->get_aircraft().load_file(cmdin["file"].asString());
	if (cmdin.isMember("data"))
		cmdout["status"] = !!m_autoroute->get_aircraft().load_string(cmdin["data"].asString());
	if (cmdin.isMember("registration"))
		m_autoroute->get_aircraft().set_callsign(cmdin["registration"].asString());
	if (cmdin.isMember("type"))
		m_autoroute->get_aircraft().set_icaotype(cmdin["type"].asString());
	if (cmdin.isMember("equipment"))
		m_autoroute->get_aircraft().set_equipment(cmdin["equipment"].asString());
	if (cmdin.isMember("transponder"))
		m_autoroute->get_aircraft().set_transponder(cmdin["transponder"].asString());
	if (cmdin.isMember("pbn"))
		m_autoroute->get_aircraft().set_pbn(cmdin["pbn"].asString());
	cmdout["registration"] = (std::string)m_autoroute->get_aircraft().get_callsign();
	cmdout["aircrafttype"] = (std::string)m_autoroute->get_aircraft().get_aircrafttype();
	cmdout["type"] = (std::string)m_autoroute->get_aircraft().get_icaotype();
	cmdout["equipment"] = (std::string)m_autoroute->get_aircraft().get_equipment();
	cmdout["transponder"] = (std::string)m_autoroute->get_aircraft().get_transponder();
	cmdout["pbn"] = (std::string)m_autoroute->get_aircraft().get_pbn();
}

void SocketServer::cmd_start(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->start();
}

void SocketServer::cmd_stop(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->stop();
}

void SocketServer::cmd_continue(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->cont();
}

void SocketServer::cmd_preload(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->preload();
}

void SocketServer::cmd_clear(const Json::Value& cmdin, Json::Value& cmdout)
{
	if (!m_autoroute) {
		cmdout["error"] = "no autoroute instance";
		return;
	}
	m_autoroute->clear();
}
