#ifndef ARSOCKSERVER_H
#define ARSOCKSERVER_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sysdeps.h"
#include <glibmm.h>
#include <giomm.h>
#include <giomm/socketservice.h>
#include <json/json.h>
#include <set>

#include "cfmuautoroute.hh"

class SocketServer {
public:
	SocketServer(CFMUAutoroute *autoroute, const Glib::RefPtr<Glib::MainLoop>& mainloop, unsigned int connlimit = ~0U,
		     unsigned int timeout = ~0U, unsigned int maxruntime = ~0U, int xdisplay = -1);
	~SocketServer();
	void listen(const std::string& path, uid_t socketuid, gid_t socketgid, mode_t socketmode);
	void sockclient(int fd);
	bool is_sockclient(void) const { return !m_listenservice; }

protected:
	CFMUAutoroute *m_autoroute;
	Glib::RefPtr<Glib::MainLoop> m_mainloop;
	Glib::RefPtr<Gio::Socket> m_listensock;
	Glib::RefPtr<Gio::UnixSocketAddress> m_listensockaddr;
	Glib::RefPtr<Gio::SocketService> m_listenservice;
	sigc::connection m_rtrtimeoutconn;
	sigc::connection m_x11srvconnwatch;
        Glib::Pid m_x11srvpid;
        unsigned int m_connectlimit;
        unsigned int m_timeout;
	unsigned int m_maxruntime;
	int m_xdisplay;
	bool m_x11srvrun;

	void remove_listensockaddr(void);
	bool on_connect(const Glib::RefPtr<Gio::SocketConnection>& connection, const Glib::RefPtr<Glib::Object>& source_object);

	class Router;

	class Client : public sigc::trackable {
	public:
		typedef Glib::RefPtr<Client> ptr_t;
		typedef Glib::RefPtr<const Client> const_ptr_t;
		Client(SocketServer *server, const Glib::RefPtr<Gio::SocketConnection>& connection);
		~Client();
		void reference(void) const;
		void unreference(void) const;
		void close(void);
		void zap(void);
		void send(const Json::Value& v);
		void msgavailable(void);
		ptr_t get_ptr(void) { reference(); return ptr_t(this); }
		const_ptr_t get_ptr(void) const { reference(); return const_ptr_t(this); }

	protected:
		SocketServer *m_server;
		Glib::RefPtr<Gio::SocketConnection> m_connection;
		sigc::connection m_connin;
		sigc::connection m_connout;
		std::list<Json::Value> m_sendqueue;
		Glib::RefPtr<Router> m_longpoll;
		typedef std::set<std::string> stringset_t;
		stringset_t m_logfilter;
		stringset_t m_logdiscard;
		mutable gint m_refcount;
		unsigned int m_closeprotect;
		bool m_firstreply;

		bool on_input(Glib::IOCondition iocond);
		bool on_output(Glib::IOCondition iocond);
	};

	typedef std::set<Client::ptr_t> clients_t;
	clients_t m_clients;

	void add_client(const Client::ptr_t& p);
	void remove_client(const Client::ptr_t& p);

	class Router : public sigc::trackable {
	public:
		typedef Glib::RefPtr<Router> ptr_t;
		typedef Glib::RefPtr<const Router> const_ptr_t;
		Router(SocketServer *server, pid_t pid, const Glib::RefPtr<Gio::Socket>& sock, const std::string& sess);
		~Router();
		void reference(void) const;
		void unreference(void) const;
		void send(const Json::Value& v);
		typedef std::set<std::string> stringset_t;
		Json::Value receive(const stringset_t& logfilter = stringset_t(), const stringset_t& logdiscard = stringset_t());
		void zap(void);
		void close(void);
		bool is_dead(void) const { return m_pid == (pid_t)-1; }
		bool handle_timeout(const Glib::TimeVal& tvacc, const Glib::TimeVal& tvstart);
		void set_longpoll(const Client::ptr_t& clnt);
		void unset_longpoll(const Client::ptr_t& clnt);
		const std::string& get_session(void) const { return m_session; }

	protected:
		SocketServer *m_server;
		Glib::RefPtr<Gio::Socket> m_socket;
		sigc::connection m_connin;
		sigc::connection m_connout;
		sigc::connection m_connchildwatch;
		std::list<Json::Value> m_sendqueue;
		std::list<Json::Value> m_recvqueue;
		std::list<Json::Value> m_logqueue;
		Glib::TimeVal m_accesstime;
		Glib::TimeVal m_starttime;
		Client::ptr_t m_longpoll;
		std::string m_session;
		mutable gint m_refcount;
		pid_t m_pid;

		bool on_input(Glib::IOCondition iocond);
		bool on_output(Glib::IOCondition iocond);
		void child_watch(GPid pid, int child_status);
		static void child_reaper(GPid pid, int child_status);
		void notify_longpoll(void);
	};

	typedef std::map<std::string,Router::ptr_t> routers_t;
	routers_t m_routers;

	Router::ptr_t find_router(const std::string& session);
	bool stop_router(const std::string& session);
	void reap_died_routers(void);
	bool router_timeout_handler(void);

	void start_x11(void);
	void stop_x11(void);
	void x11srv_child_watch(GPid pid, int child_status);

	void sockcli_close(void);
	void sockcli_send(const Json::Value& v);
	bool sockcli_input(Glib::IOCondition iocond);
	void sockcli_statuschange(CFMUAutoroute::statusmask_t status);
	void sockcli_autoroutelog(CFMUAutoroute::log_t item, const Glib::ustring& line);

	typedef void (SocketServer::*cmd_t)(const Json::Value& cmdin, Json::Value& cmdout);
	typedef std::map<std::string,cmd_t> cmdlist_t;
	cmdlist_t m_cmdlist;

	static Point get_point(const Json::Value& cmd, const std::string& name);
	static void set_point(Json::Value& cmd, const std::string& name, const Point& pt);
	void cmd_nop(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_quit(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_departure(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_destination(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_crossing(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_enroute(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_levels(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_exclude(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_tfr(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_atmosphere(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_cruise(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_optimization(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_preferred(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_aircraft(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_start(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_stop(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_continue(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_preload(const Json::Value& cmdin, Json::Value& cmdout);
	void cmd_clear(const Json::Value& cmdin, Json::Value& cmdout);
};

#endif /* ARSOCKSERVER_H */
