#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>

#include "cfmuautoroute.hh"
#include "icaofpl.h"
#include "airdata.h"

const double CFMUAutoroute::Crossing::maxradius;

CFMUAutoroute::Performance::Performance(int alt, double secpernmi, double fuelpersec, double timepenalty, double fuelpenalty, double prefpenalty,
					Glib::RefPtr<GRIB2::LayerResult> windu, Glib::RefPtr<GRIB2::LayerResult> windv,
					Glib::RefPtr<GRIB2::LayerResult> temp)
	: m_windu(windu), m_windv(windv), m_temp(temp), m_secpernmi(secpernmi), m_fuelpersec(fuelpersec),
	  m_timepenalty(timepenalty), m_fuelpenalty(fuelpenalty), m_preferredpenalty(prefpenalty), m_altitude(alt)
{
}

std::string CFMUAutoroute::Performance::get_fpltas(void) const
{
	std::ostringstream oss;
	oss << 'N' << std::setw(4) << std::setfill('0') << Point::round<int,double>(get_tas());
	return oss.str();
}

Wind<double> CFMUAutoroute::Performance::get_wind(const Point& pt) const
{
	Wind<double> wnd;
	if (!m_windu || !m_windv) {
		wnd.set_wind(0, 0);
		return wnd;
	}
	float wu(m_windu->operator()(pt));
	float wv(m_windv->operator()(pt));
	{
		std::pair<float,float> w(m_windu->get_layer()->get_grid()->transform_axes(wu, wv));
		wu = w.first;
		wv = w.second;
	}
	// convert from m/s to kts
	wu *= (-1e-3f * Point::km_to_nmi * 3600);
	wv *= (-1e-3f * Point::km_to_nmi * 3600);
	wnd.set_wind((M_PI - atan2f(wu, wv)) * Wind<double>::from_rad, sqrtf(wu * wu + wv * wv));
	return wnd;
}

double CFMUAutoroute::Performance::get_temp(const Point& pt) const
{
	if (!m_temp)
		return std::numeric_limits<double>::quiet_NaN();
	return m_temp->operator()(pt);
}

const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchall;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchnone;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchawy;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchdctawy;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchdctawysidstar;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchdctawysidstarlevel;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchsidstar;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaymatchlevel;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaylevelup;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwayleveldown;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaysid;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaystar;
const CFMUAutoroute::lgraphairwayindex_t CFMUAutoroute::airwaydct;

CFMUAutoroute::CFMUAutoroute()
	: m_route(*(FPlan *)0), m_routetime(0), m_routefuel(0), m_deptimepenalty(0), m_depfuelpenalty(0),
	  m_dctlimit(50.0), m_dctpenalty(1.0), m_vfraspclimit(100.0), m_qnh(IcaoAtmosphere::std_sealevel_pressure), m_isaoffs(0),
	  m_enginerpm(2400), m_enginemp(std::numeric_limits<double>::quiet_NaN()), m_enginebhp(130), m_starttime(0), m_preferredlevel(100), m_preferredpenalty(1.1),
	  m_preferredclimb(3), m_preferreddescent(1), m_childtimeoutcount(0), m_pathprobe(-1),
	  m_opttarget(opttarget_time), m_childrun(false), m_forceenrifr(false), m_tfravailable(true), m_tfrenable(true), m_tfrloaded(false),
	  m_honourawylevels(true), m_precompgraphenable(true), m_windenable(false), m_grib2loaded(false), m_running(false), m_done(false)
{
	m_iterationcount[0] = m_iterationcount[1] = 0;
	for (unsigned int i = 0; i < 2; ++i) {
		m_airport[i] = AirportsDb::Airport();
		m_airportconn[i].set_invalid();
		m_airportconnlimit[i] = 40.0;
		m_airportconnident[i].clear();
		m_airportconntype[i] = Engine::AirwayGraphResult::Vertex::type_undefined;
		m_airportconndb[i] = true;
		m_airportifr[i] = true;
	}
	m_levels[0] = 50;
	m_levels[1] = 120;
	time(&m_starttime);
	clear();
	// 10 o'clock
	set_deptime(3600*10);
	check_tfr_available();
}

CFMUAutoroute::~CFMUAutoroute()
{
	child_close();
	clear();
}

void CFMUAutoroute::set_opttarget(opttarget_t t)
{
	if (m_opttarget != t)
		clear();
	m_opttarget = t;
}

const Glib::ustring& to_str(CFMUAutoroute::opttarget_t t)
{
	switch (t) {
	case CFMUAutoroute::opttarget_time:
	{
		static const Glib::ustring r("time");
		return r;
	}

	case CFMUAutoroute::opttarget_fuel:
	{
		static const Glib::ustring r("fuel");
		return r;
	}

	case CFMUAutoroute::opttarget_preferred:
	{
		static const Glib::ustring r("preferred");
		return r;
	}

	default:
	{
		static const Glib::ustring r("??");
		return r;
	}
	}
}

const Glib::ustring& to_str(CFMUAutoroute::log_t l)
{
	switch (l) {
	case CFMUAutoroute::log_fplproposal:
	{
		static const Glib::ustring r("fplproposal");
		return r;
	}

	case CFMUAutoroute::log_fplvalidation:
	{
		static const Glib::ustring r("fplvalidation");
		return r;
	}

	case CFMUAutoroute::log_graphrule:
	{
		static const Glib::ustring r("graphrule");
		return r;
	}

	case CFMUAutoroute::log_graphruledesc:
	{
		static const Glib::ustring r("graphruledesc");
		return r;
	}

	case CFMUAutoroute::log_graphruleoprgoal:
	{
		static const Glib::ustring r("graphruleoprgoal");
		return r;
	}

	case CFMUAutoroute::log_graphchange:
	{
		static const Glib::ustring r("graphchange");
		return r;
	}

	case CFMUAutoroute::log_normal:
	{
		static const Glib::ustring r("normal");
		return r;
	}

	case CFMUAutoroute::log_debug0:
	{
		static const Glib::ustring r("debug0");
		return r;
	}

	case CFMUAutoroute::log_debug1:
	{
		static const Glib::ustring r("debug1");
		return r;
	}

	default:
	{
		static const Glib::ustring r("??");
		return r;
	}
	}
}

void CFMUAutoroute::set_departure(const AirportsDb::Airport& el)
{
	if (m_airport[0].is_valid() != el.is_valid() ||
	    m_airport[0].get_icao() != el.get_icao() ||
	    m_airport[0].get_name() != el.get_name() ||
	    m_airport[0].get_coord() != el.get_coord())
		clear();
	m_airport[0] = el;
}

bool CFMUAutoroute::set_departure(const Glib::ustring& icao, const Glib::ustring& name)
{
	AirportsDb::Airport el;
	if (!find_airport(el, icao, name))
		return false;
	set_departure(el);
	return true;
}

void CFMUAutoroute::set_departure_ifr(bool ifr)
{
	ifr = !!ifr;
	if (m_airportifr[0] != ifr)
		clear();
	m_airportifr[0] = ifr;
}

void CFMUAutoroute::set_sid(const Point& pt)
{
	if (m_airportconn[0] != pt)
		clear();
	m_airportconn[0] = pt;
	m_airportconnident[0].clear();
	m_airportconntype[0] = Engine::AirwayGraphResult::Vertex::type_undefined;
}

bool CFMUAutoroute::set_sid(const Glib::ustring& name)
{
	Point ptnear(m_airport[0].get_coord());
	if (!m_airport[0].is_valid())
		ptnear.set_invalid();
	Point pt;
	Glib::ustring ident;
	Engine::AirwayGraphResult::Vertex::type_t typ;
	if (!find_point(pt, ident, typ, name, m_airportifr[0], ptnear))
		return false;
	set_sid(pt);
	m_airportconnident[0] = ident;
	m_airportconntype[0] = typ;
	return true;
}

void CFMUAutoroute::set_sidlimit(double l)
{
	if (std::isnan(l))
		l = 1.;
	l = std::max(l, 0.);
	if (m_airportconnlimit[0] != l)
		clear();
	m_airportconnlimit[0] = l;
}

void CFMUAutoroute::set_siddb(bool db)
{
	db = !!db;
	if (m_airportconndb[0] != db)
		clear();
	m_airportconndb[0] = db;
}

void CFMUAutoroute::set_destination(const AirportsDb::Airport& el)
{
	if (m_airport[1].is_valid() != el.is_valid() ||
	    m_airport[1].get_icao() != el.get_icao() ||
	    m_airport[1].get_name() != el.get_name() ||
	    m_airport[1].get_coord() != el.get_coord())
		clear();
	m_airport[1] = el;
}

bool CFMUAutoroute::set_destination(const Glib::ustring& icao, const Glib::ustring& name)
{
	AirportsDb::Airport el;
	if (!find_airport(el, icao, name))
		return false;
	set_destination(el);
	return true;
}

void CFMUAutoroute::set_destination_ifr(bool ifr)
{
	ifr = !!ifr;
	if (m_airportifr[1] != ifr)
		clear();
	m_airportifr[1] = ifr;
}

void CFMUAutoroute::set_star(const Point& pt)
{
	if (m_airportconn[1] != pt)
		clear();
	m_airportconn[1] = pt;
	m_airportconnident[1].clear();
	m_airportconntype[1] = Engine::AirwayGraphResult::Vertex::type_undefined;
}

bool CFMUAutoroute::set_star(const Glib::ustring& name)
{
	Point ptnear(m_airport[1].get_coord());
	if (!m_airport[1].is_valid())
		ptnear.set_invalid();
	Point pt;
	Glib::ustring ident;
	Engine::AirwayGraphResult::Vertex::type_t typ;
	if (!find_point(pt, ident, typ, name, m_airportifr[1], ptnear))
		return false;
	set_star(pt);
	m_airportconnident[1] = ident;
	m_airportconntype[1] = typ;
	return true;
}

void CFMUAutoroute::set_starlimit(double l)
{
	if (std::isnan(l))
		l = 0.;
	l = std::max(l, 0.);
	if (m_airportconnlimit[1] != l)
		clear();
	m_airportconnlimit[1] = l;
}

void CFMUAutoroute::set_stardb(bool db)
{
	db = !!db;
	if (m_airportconndb[1] != db)
		clear();
	m_airportconndb[1] = db;
}

void CFMUAutoroute::set_crossing(unsigned int index, const Point& pt)
{
	if (index <= get_crossing_size())
		set_crossing_size(index + 1);
	if (m_crossing[index].get_coord() != pt)
		clear();
	m_crossing[index].set_coord(pt);
	m_crossing[index].set_ident();
	m_crossing[index].set_type();
}

bool CFMUAutoroute::set_crossing(unsigned int index, const std::string& ident)
{
	if (index <= get_crossing_size())
		set_crossing_size(index + 1);
	Point ptnear;
	ptnear.set_invalid();
	if (m_airport[0].is_valid() && m_airport[1].is_valid())
		ptnear = m_airport[0].get_coord().halfway(m_airport[1].get_coord());
	else if (m_airport[0].is_valid())
		ptnear = m_airport[0].get_coord();
	else if (m_airport[1].is_valid())
		ptnear = m_airport[1].get_coord();
	Point pt;
	Glib::ustring ident2;
	Engine::AirwayGraphResult::Vertex::type_t typ;
	if (!find_point(pt, ident2, typ, ident, m_airportifr[1], ptnear))
		return false;
	set_crossing(index, pt);
	m_crossing[index].set_ident(ident2);
	m_crossing[index].set_type(typ);
	return true;
}

void CFMUAutoroute::set_crossing_radius(unsigned int index, double r)
{
	if (index <= get_crossing_size())
		set_crossing_size(index + 1);
	if (m_crossing[index].get_radius() != r)
		clear();
	m_crossing[index].set_radius(r);
}

void CFMUAutoroute::set_crossing_level(unsigned int index, int minlevel, int maxlevel)
{
	if (index <= get_crossing_size())
		set_crossing_size(index + 1);
	if (m_crossing[index].get_minlevel() != minlevel || m_crossing[index].get_maxlevel() != maxlevel)
		clear();
	m_crossing[index].set_level(minlevel, maxlevel);
}

const Point& CFMUAutoroute::get_crossing(unsigned int index) const
{
	if (index >= get_crossing_size())
		return Point::invalid;
	return m_crossing[index].get_coord();
}

const std::string& CFMUAutoroute::get_crossing_ident(unsigned int index) const
{
	static const std::string empty;
	if (index >= get_crossing_size())
		return empty;
	return m_crossing[index].get_ident();
}

double CFMUAutoroute::get_crossing_radius(unsigned int index) const
{
	if (index >= get_crossing_size())
		return 0;
	return m_crossing[index].get_radius();
}

int CFMUAutoroute::get_crossing_minlevel(unsigned int index) const
{
	if (index >= get_crossing_size())
		return 0;
	return m_crossing[index].get_minlevel();
}

int CFMUAutoroute::get_crossing_maxlevel(unsigned int index) const
{
	if (index >= get_crossing_size())
		return 600;
	return m_crossing[index].get_maxlevel();
}

Engine::AirwayGraphResult::Vertex::type_t CFMUAutoroute::get_crossing_type(unsigned int index) const
{
	if (index >= get_crossing_size())
		return Engine::AirwayGraphResult::Vertex::type_undefined;
	return m_crossing[index].get_type();
}

void CFMUAutoroute::set_crossing_size(unsigned int sz)
{
	if (sz == get_crossing_size())
		return;
	m_crossing.resize(sz);
	clear();
}

void CFMUAutoroute::set_dctlimit(double l)
{
	if (std::isnan(l))
		l = 0.;
	l = std::max(l, 0.);
	if (m_dctlimit != l)
		clear();
	m_dctlimit = l;
}

void CFMUAutoroute::set_dctpenalty(double l)
{
	if (std::isnan(l))
		l = 1.;
	l = std::max(std::min(l, 1.e6), 1.e-6);
	if (m_dctpenalty != l)
		clear();
	m_dctpenalty = l;
}

void CFMUAutoroute::set_vfraspclimit(double l)
{
	if (std::isnan(l))
		l = 0.;
	l = std::max(l, 0.);
	if (m_vfraspclimit != l)
		clear();
	m_vfraspclimit = l;
}

void CFMUAutoroute::set_preferred_level(int level)
{
	if (m_preferredlevel != level)
		clear();
	m_preferredlevel = level;
}

void CFMUAutoroute::set_preferred_penalty(double p)
{
	if (std::isnan(p))
		p = 1.1;
	p = std::max(1.0, std::min(10.0, p));
	if (m_preferredpenalty != p)
		clear();
	m_preferredpenalty = p;
}

void CFMUAutoroute::set_preferred_climb(double c)
{
	if (std::isnan(c))
		c = 10.;
	c = std::max(0.1, std::min(10.0, c));
	double d(m_preferreddescent);
	d = std::min(0.75 * c, d);
	if (m_preferredclimb != c || m_preferreddescent != d)
		clear();
	m_preferredclimb = c;
	m_preferreddescent = d;
}

void CFMUAutoroute::set_preferred_descent(double d)
{
	if (std::isnan(d))
		d = 0.;
	d = std::max(0.0, std::min(0.75 * m_preferredclimb, d));
	if (m_preferreddescent != d)
		clear();
	m_preferreddescent = d;
}

void CFMUAutoroute::set_excluderegions(const excluderegions_t& r)
{
	m_excluderegions = r;
	clear();
}

void CFMUAutoroute::add_excluderegion(const ExcludeRegion& r)
{
	m_excluderegions.push_back(r);
	clear();
}

void CFMUAutoroute::clear_excluderegions(void)
{
	m_excluderegions.clear();
	clear();
}

void CFMUAutoroute::set_force_enroute_ifr(bool ifr)
{
	ifr = !!ifr;
	if (m_forceenrifr != ifr)
		clear();
	m_forceenrifr = ifr;
}

void CFMUAutoroute::check_tfr_available(void)
{
	std::string fnamex(get_tfr_xml_filename());
	std::string fnameb(get_tfr_bin_filename());
	if (false)
		std::cerr << "TFR Filename " << fnamex << " exists "
			  << (Glib::file_test(fnamex, Glib::FILE_TEST_EXISTS) ? "yes" : "no") << " dir "
			  << (Glib::file_test(fnamex, Glib::FILE_TEST_IS_DIR) ? "yes" : "no") << std::endl
			  << "TFR Filename " << fnameb << " exists "
			  << (Glib::file_test(fnameb, Glib::FILE_TEST_EXISTS) ? "yes" : "no") << " dir "
			  << (Glib::file_test(fnameb, Glib::FILE_TEST_IS_DIR) ? "yes" : "no") << std::endl;
	m_tfravailable = (Glib::file_test(fnamex, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fnamex, Glib::FILE_TEST_IS_DIR)) ||
		(Glib::file_test(fnameb, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fnameb, Glib::FILE_TEST_IS_DIR));
	m_tfrenable = get_tfr_available();
}

void CFMUAutoroute::set_tfr_enabled(bool ena)
{
	m_tfrenable = ena && get_tfr_available();	
}

void CFMUAutoroute::set_tfr_trace(const std::string& rules)
{
	m_tfr.tracerules_clear();
	std::string::size_type nd(0);
	for (;;) {
		std::string::size_type nb(nd);
		while (nb < rules.size() && std::isspace(rules[nb]))
			++nb;
		if (nb >= rules.size())
			break;
		std::string::size_type ne(rules.find(',', nb));
		if (ne == std::string::npos)
			nd = ne = rules.size();
		else
			nd = ne + 1;
		while (ne > nb && std::isspace(rules[ne - 1]))
			--ne;
		m_tfr.tracerules_add(rules.substr(nb, ne - nb));
	}
}

std::string CFMUAutoroute::get_tfr_trace(void) const
{
	std::string r;
	bool subseq(false);
	for (TrafficFlowRestrictions::tracerules_const_iterator_t ti(m_tfr.tracerules_begin()), te(m_tfr.tracerules_end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	return r;
}

void CFMUAutoroute::set_tfr_disable(const std::string& rules)
{
	m_tfr.disabledrules_clear();
	std::string::size_type nd(0);
	for (;;) {
		std::string::size_type nb(nd);
		while (nb < rules.size() && std::isspace(rules[nb]))
			++nb;
		if (nb >= rules.size())
			break;
		std::string::size_type ne(rules.find(',', nb));
		if (ne == std::string::npos)
			nd = ne = rules.size();
		else
			nd = ne + 1;
		while (ne > nb && std::isspace(rules[ne - 1]))
			--ne;
		m_tfr.disabledrules_add(rules.substr(nb, ne - nb));
	}
}

std::string CFMUAutoroute::get_tfr_disable(void) const
{
	std::string r;
	bool subseq(false);
	for (TrafficFlowRestrictions::disabledrules_const_iterator_t ti(m_tfr.disabledrules_begin()), te(m_tfr.disabledrules_end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	return r;
}

void CFMUAutoroute::set_precompgraph_enabled(bool ena)
{
	m_precompgraphenable = ena;
}

void CFMUAutoroute::set_wind_enabled(bool ena)
{
	m_windenable = ena;
}

void CFMUAutoroute::set_qnh(double q)
{
	if (std::isnan(q))
		q = IcaoAtmosphere::std_sealevel_pressure;
	q = std::max(std::min(q, 1300.), 700.);
	if (m_qnh != q)
		clear();
	m_qnh = q;
}

void CFMUAutoroute::set_isaoffs(double t)
{
	if (std::isnan(t))
		t = 0.;
	t = std::max(std::min(t, 100.), -100.);
	if (m_isaoffs != t)
		clear();
	m_isaoffs = t;
}

void CFMUAutoroute::set_engine_rpm(double l)
{
	if (m_enginerpm != l)
		clear();
	m_enginerpm = l;
}

void CFMUAutoroute::set_engine_mp(double l)
{
	if (m_enginemp != l)
		clear();
	m_enginemp = l;
}

void CFMUAutoroute::set_engine_bhp(double l)
{
	if (m_enginebhp != l)
		clear();
	m_enginebhp = l;
}

void CFMUAutoroute::set_levels(int base, int top)
{
	if (m_levels[0] != base ||
	    m_levels[1] != top)
		clear();
	m_levels[0] = base;
	m_levels[1] = top;
}

void CFMUAutoroute::set_honour_awy_levels(bool awylvl)
{
	awylvl = !!awylvl;
	if (m_honourawylevels != awylvl)
		clear();
	m_honourawylevels = awylvl;
}

void CFMUAutoroute::set_deptime(time_t t)
{
	if (m_route.get_time_offblock_unix() != t)
		clear();
	m_route.set_time_offblock_unix(t);
}

void CFMUAutoroute::set_db_directories(const std::string& dir_main, Engine::auxdb_mode_t auxdbmode, const std::string& dir_aux)
{
	closedb();
	m_dbdir_main = dir_main;
        switch (auxdbmode) {
	case Engine::auxdb_prefs:
		m_dbdir_aux = (Glib::ustring)m_prefs.globaldbdir;
		break;

	case Engine::auxdb_override:
		m_dbdir_aux = dir_aux;
		break;

	default:
		m_dbdir_aux.clear();
		break;
        }
	check_tfr_available();
}

void CFMUAutoroute::opendb(void)
{
	if (!m_airportdb.is_open()) {
		try {
			m_airportdb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_airportdb.is_dbfile_exists(get_db_auxdir()))
				m_airportdb.attach(get_db_auxdir());
			if (m_airportdb.is_aux_attached())
				std::cerr << "Auxillary airports database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening airports database: ") + e.what());
		}
	}
	if (!m_navaiddb.is_open()) {
		try {
			m_navaiddb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_navaiddb.is_dbfile_exists(get_db_auxdir()))
				m_navaiddb.attach(get_db_auxdir());
			if (m_navaiddb.is_aux_attached())
				std::cerr << "Auxillary navaids database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening navaids database: ") + e.what());
		}
	}
	if (!m_waypointdb.is_open()) {
		try {
			m_waypointdb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_waypointdb.is_dbfile_exists(get_db_auxdir()))
				m_waypointdb.attach(get_db_auxdir());
			if (m_waypointdb.is_aux_attached())
				std::cerr << "Auxillary waypoints database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening waypoints database: ") + e.what());
		}
	}
	if (!m_airspacedb.is_open()) {
		try {
			m_airspacedb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_airspacedb.is_dbfile_exists(get_db_auxdir()))
				m_airspacedb.attach(get_db_auxdir());
			if (m_airspacedb.is_aux_attached())
				std::cerr << "Auxillary airspaces database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening airspaces database: ") + e.what());
		}
	}
	if (!m_airwaydb.is_open()) {
		try {
			m_airwaydb.open(get_db_maindir());
			if (!get_db_auxdir().empty() && m_airwaydb.is_dbfile_exists(get_db_auxdir()))
				m_airwaydb.attach(get_db_auxdir());
			if (m_airwaydb.is_aux_attached())
				std::cerr << "Auxillary airways database attached" << std::endl;
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening airways database: ") + e.what());
		}
	}
	if (!m_mapelementdb.is_open()) {
		try {
			m_mapelementdb.open(get_db_auxdir().empty() ? Engine::get_default_aux_dir() : get_db_auxdir());
		} catch (const std::exception& e) {
			m_signal_log(log_normal, std::string("Error opening mapelements database: ") + e.what());
		}
	}
	try {
		m_topodb.open(get_db_auxdir().empty() ? Engine::get_default_aux_dir() : get_db_auxdir());
	} catch (const std::exception& e) {
		m_signal_log(log_normal, std::string("Error opening topo database: ") + e.what());
	}
}

void CFMUAutoroute::closedb(void)
{
	if (m_airportdb.is_open())
		m_airportdb.close();
	if (m_navaiddb.is_open())
		m_navaiddb.close();
	if (m_waypointdb.is_open())
		m_waypointdb.close();
	if (m_airwaydb.is_open())
		m_airwaydb.close();
	if (m_airspacedb.is_open())
		m_airspacedb.close();
	if (m_mapelementdb.is_open())
		m_mapelementdb.close();
	m_topodb.close();
}

bool CFMUAutoroute::find_airport(AirportsDb::Airport& el, const Glib::ustring& icao, const Glib::ustring& name)
{
	opendb();
	AirportsDb::elementvector_t ev;
	if (!icao.empty())
		ev = m_airportdb.find_by_icao(icao, 0, AirportsDb::comp_contains, 0, AirportsDb::Airport::subtables_all);
	if (!name.empty()) {
		AirportsDb::elementvector_t ev1(m_airportdb.find_by_name(name, 0, AirportsDb::comp_contains, 0, AirportsDb::Airport::subtables_all));
		ev.insert(ev.end(), ev1.begin(), ev1.end());
	}
	if (ev.empty())
		return false;
	// remove duplicate airports
	for (unsigned int i = 0; i < ev.size(); ++i)
		for (unsigned int j = i + 1; j < ev.size();) {
			if (ev[i].get_sourceid() != ev[j].get_sourceid()) {
				++j;
				continue;
			}
			ev.erase(ev.begin() + j);
		}
	if (ev.size() != 1) {
		typedef std::multimap<int,unsigned int> priomap_t;
		priomap_t priomap;
		Glib::ustring icaou(icao.uppercase()), nameu(name.uppercase());
		for (unsigned int i = 0; i < ev.size(); ++i) {
			const AirportsDb::Airport& arpt(ev[i]);
			Glib::ustring iu(arpt.get_icao().uppercase());
			Glib::ustring nu(arpt.get_name().uppercase());
			int prio(0);
			if (iu == icaou)
				prio -= 2;
			else if (iu.find(icaou) != Glib::ustring::npos)
				--prio;
			if (nu == nameu)
				prio -= 2;
			else if (nu.find(nameu) != Glib::ustring::npos)
				--prio;
			priomap.insert(std::make_pair(prio, i));
		}
		priomap_t::const_iterator it(priomap.begin());
		if (it == priomap.end() || it->first == 0)
			return false;
		priomap_t::const_iterator it2(it);
		++it2;
		if (it2 != priomap.end() && it2->first == it->first)
			return false;
		el = ev[it->second];
		return true;
	}
	el = ev[0];
	return true;
}

bool CFMUAutoroute::find_point(Point& pt, Glib::ustring& ident, Engine::AirwayGraphResult::Vertex::type_t& typ, const Glib::ustring& name, bool ifr, const Point& ptnear)
{
	ident.clear();
	typ = Engine::AirwayGraphResult::Vertex::type_undefined;
	if (name.empty())
		return false;
	typedef std::vector<Point> result_t;
	result_t result;
	typedef std::vector<Engine::AirwayGraphResult::Vertex::type_t> resulttype_t;
	resulttype_t resulttype;
	typedef std::vector<Glib::ustring> resultident_t;
	resultident_t resultident;
	Glib::ustring nameu(name.uppercase());
	{
		NavaidsDb::elementvector_t ev(m_navaiddb.find_by_text(nameu, 0, NavaidsDb::comp_contains, 0, NavaidsDb::Navaid::subtables_all));
		for (NavaidsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
			const NavaidsDb::Navaid& el(*ei);
			if (!el.is_valid() || (el.get_icao().uppercase().find(nameu) == Glib::ustring::npos &&
					       el.get_name().uppercase().find(nameu) == Glib::ustring::npos))
				continue;
			result.push_back(el.get_coord());
			resulttype.push_back(Engine::AirwayGraphResult::Vertex::type_navaid);
			resultident.push_back(el.get_icao());
		}
	}
	{
		WaypointsDb::elementvector_t ev(m_waypointdb.find_by_text(nameu, 0, WaypointsDb::comp_contains, 0, WaypointsDb::Waypoint::subtables_all));
		for (WaypointsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
			const WaypointsDb::Waypoint& el(*ei);
			if (!el.is_valid() || el.get_name().uppercase().find(nameu) == Glib::ustring::npos)
				continue;
			result.push_back(el.get_coord());
			resulttype.push_back(Engine::AirwayGraphResult::Vertex::type_intersection);
			resultident.push_back(el.get_name());
		}
	}
	if (!ifr) {
		{
			Point pt;
			if (!((Point::setstr_lat | Point::setstr_lon) & ~pt.set_str(name)) && !pt.is_invalid()) {
				result.push_back(pt);
				resulttype.push_back(Engine::AirwayGraphResult::Vertex::type_undefined);
				resultident.push_back(Glib::ustring());
			}
		}
		{
			AirportsDb::elementvector_t ev(m_airportdb.find_by_text(nameu, 0, AirportsDb::comp_contains, 0, AirportsDb::Airport::subtables_all));
			for (AirportsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
				const AirportsDb::Airport& el(*ei);
				if (!el.is_valid() || (el.get_icao().uppercase().find(nameu) == Glib::ustring::npos &&
						       el.get_name().uppercase().find(nameu) == Glib::ustring::npos))
					continue;
				result.push_back(el.get_coord());
				resulttype.push_back(Engine::AirwayGraphResult::Vertex::type_airport);
				resultident.push_back(el.get_icao());
			}
		}
		{
			MapelementsDb::elementvector_t ev(m_mapelementdb.find_by_text(nameu, 0, MapelementsDb::comp_contains, 0, MapelementsDb::Mapelement::subtables_all));
			for (MapelementsDb::elementvector_t::const_iterator ei(ev.begin()), ee(ev.end()); ei != ee; ++ei) {
				const MapelementsDb::Mapelement& el(*ei);
				if (!el.is_valid() || el.get_name().uppercase().find(nameu) == Glib::ustring::npos)
					continue;
				switch (el.get_typecode()) {
				case MapelementsDb::Mapelement::maptyp_city:
				case MapelementsDb::Mapelement::maptyp_village:
				case MapelementsDb::Mapelement::maptyp_spot:
				case MapelementsDb::Mapelement::maptyp_landmark:
				case MapelementsDb::Mapelement::maptyp_lake:
				case MapelementsDb::Mapelement::maptyp_lake_t:
					break;

				default:
					continue;
				}
				result.push_back(el.get_coord());
				resulttype.push_back(Engine::AirwayGraphResult::Vertex::type_mapelement);
				resultident.push_back(el.get_name());
			}
		}
	}
	if (ptnear.is_invalid()) {
		if (result.size() == 1) {
			pt = result[0];
			ident = resultident[0];
			typ = resulttype[0];
			return true;
		}
		return false;
	}
	if (result.empty())
		return false;
	uint64_t dist(ptnear.simple_distance_rel(result[0]));
	unsigned int idx(0);
	for (unsigned i = 1, j = result.size(); i < j; ++i) {
		uint64_t d(ptnear.simple_distance_rel(result[i]));
		if (d >= dist)
			continue;
		dist = d;
		idx = i;
	}
	pt = result[idx];
	ident = resultident[idx];
	typ = resulttype[idx];
	return true;
}

Glib::ustring CFMUAutoroute::get_simplefpl(void) const
{
	if (m_performance.empty() || m_route.get_nrwpt() < 2)
		return "";
	std::ostringstream fpl;
	std::ostringstream otherinfo;
	char flightrules(m_route.get_flightrules());
	const FPlanWaypoint& wptdep(m_route[0]);
	const FPlanWaypoint& wptdest(m_route[m_route.get_nrwpt() - 1]);
	int alt(std::numeric_limits<int>::min());
	uint16_t flags(wptdep.get_flags());
	if (flightrules == 'V') {
		alt = m_route.max_altitude(flags);
	} else if (flags & FPlanWaypoint::altflag_ifr) {
		if (m_route.get_nrwpt() == 2) {
			alt = std::max(m_route[0].get_altitude(), m_route[1].get_altitude());
			if (alt >= 5000)
				alt += 5000;
			alt += 1999;
			alt /= 1000;
			alt *= 1000;
		} else {
			alt = m_route[1].get_altitude();
			flags &= ~FPlanWaypoint::altflag_standard;
			flags |= m_route[1].get_flags() & FPlanWaypoint::altflag_standard;
			if (m_pathprobe >= 0) {
				if (m_pathprobe != 0)
					flags &= ~FPlanWaypoint::altflag_ifr;
				if (m_pathprobe >= m_route.get_nrwpt() && 
				    !(m_route[m_pathprobe].get_flags() & FPlanWaypoint::altflag_ifr))
					flightrules = 'V';
				else
					flightrules = (flags & FPlanWaypoint::altflag_ifr) ? 'Y' : 'Z';
			}
		}
	}
	fpl << "-(FPL-";
	if (!m_done || m_aircraft.get_callsign().empty())
		fpl << m_callsign;
	else
		fpl << m_aircraft.get_callsign();
	fpl << '-' << flightrules << "G -1" << m_aircraft.get_icaotype()
	    << '/' << m_aircraft.get_wakecategory() << " -" << m_aircraft.get_equipment() << '/' << m_aircraft.get_transponder()
	    << " -";
	if (AirportsDb::Airport::is_fpl_zzzz(wptdep.get_icao())) {
		fpl << "ZZZZ";
		otherinfo << "DEP/" << wptdep.get_name() << ' ' << wptdep.get_coord().get_fpl_str() << ' ';
	} else {
		fpl << wptdep.get_icao();
	}
	if (!m_aircraft.get_pbn().empty())
		otherinfo << "PBN/" << m_aircraft.get_pbn() << ' ';
	{
		struct tm utm;
		time_t deptime(wptdep.get_time_unix());
		gmtime_r(&deptime, &utm);
		fpl << std::setw(2) << std::setfill('0') << utm.tm_hour
		    << std::setw(2) << std::setfill('0') << utm.tm_min
		    << " -";
		if (deptime >= m_starttime) {
			// also add date if in the future
			otherinfo << "DOF/" << std::setw(2) << std::setfill('0') << (utm.tm_year % 100)
				  << std::setw(2) << std::setfill('0') << (utm.tm_mon + 1)
				  << std::setw(2) << std::setfill('0') << utm.tm_mday << ' ';
		}
	}
	if (alt == std::numeric_limits<int>::min())
		fpl << m_performance[findperfindex(m_route.max_altitude())].get_fpltas() << "VFR";
	else
		fpl << m_performance[findperfindex(alt)].get_fpltas()
		    << ((flags & FPlanWaypoint::altflag_standard) ? 'F' : 'A')
		    << std::setw(3) << std::setfill('0') << ((alt + 50) / 100);
	std::ostringstream eet;
	if (m_route.get_nrwpt() == 2) {
		if (wptdep.get_flags() & wptdest.get_flags() & FPlanWaypoint::altflag_ifr)
			fpl << " DCT";
	} else {
		for (unsigned int idx = 1;;) {
			const FPlanWaypoint& wpt(m_route[idx]);
			++idx;
			if (wpt.get_pathcode() == FPlanWaypoint::pathcode_sid) {
				if (idx + 1 >= m_route.get_nrwpt())
					break;
				continue;
			}
			Glib::ustring ident(wpt.get_icao());
			if (ident.empty() || ident == wptdep.get_icao() || ident == wptdest.get_icao())
				ident = wpt.get_name();
			{
				const FPlanWaypoint& wptprev(m_route[idx-2]);
				if (ident.empty() || 
				    ((Engine::AirwayGraphResult::Vertex::is_ident_numeric(ident) || ident.size() < 2) &&
				     wpt.get_pathcode() == FPlanWaypoint::pathcode_airway &&
				     wptprev.get_pathcode() == FPlanWaypoint::pathcode_airway &&
				     wpt.get_pathname() == wptprev.get_pathname() &&
				     (wpt.get_flags() & wptprev.get_flags() & FPlanWaypoint::altflag_ifr))) {
					if (idx + 1 >= m_route.get_nrwpt())
						break;
					continue;
				}
			}
			fpl << ' ' << ident;
			uint16_t flgchg(flags);
			flags = wpt.get_flags();
			if (m_pathprobe >= 0 && idx != m_pathprobe + 1)
				flags &= ~FPlanWaypoint::altflag_ifr;
			flgchg ^= flags;
			{
				int16_t a(wpt.get_altitude());
				if (alt != a)
					flgchg |= FPlanWaypoint::altflag_standard;
				alt = a;
			}
			if (flgchg & FPlanWaypoint::altflag_standard)
				fpl << '/' << m_performance[findperfindex(alt)].get_fpltas()
				    << FPlanWaypoint::get_altchar(flags)
				    << std::setw(3) << std::setfill('0') << ((alt + 50) / 100);
			if (flgchg & FPlanWaypoint::altflag_ifr) {
				fpl << ' ' << FPlanWaypoint::get_ruleschar(flags) << "FR";
				if (flags & FPlanWaypoint::altflag_ifr) {
					unsigned int tm(wpt.get_time_unix() - wptdep.get_time_unix());
					tm += 30;
					tm /= 60;
					unsigned int hr(tm / 60);
					tm -= hr * 60;
					eet << ident << std::setw(2) << std::setfill('0') << hr
					    << std::setw(2) << std::setfill('0') << tm << ' ';
				}
			}
			if (idx + 1 >= m_route.get_nrwpt() || wpt.get_pathcode() == FPlanWaypoint::pathcode_star)
				break;
			if (true || (flags & FPlanWaypoint::altflag_ifr)) {
				switch (wpt.get_pathcode()) {
				case FPlanWaypoint::pathcode_directto:
					fpl << " DCT";
					break;

				case FPlanWaypoint::pathcode_airway:
					fpl << ' ' << wpt.get_pathname();
					break;

				default:
					break;
				}
			}
		}
		{
			uint16_t flgchg(flags);
			flags = wptdest.get_flags();
			flgchg ^= flags;
			if (flgchg & FPlanWaypoint::altflag_ifr && m_pathprobe < 0)
				fpl << ' ' << FPlanWaypoint::get_ruleschar(flags) << "FR";
		}
	}
	fpl << " -";
	if (AirportsDb::Airport::is_fpl_zzzz(wptdest.get_icao())) {
		fpl << "ZZZZ";
		otherinfo << "DEST/" << wptdest.get_name() << ' ' << wptdest.get_coord().get_fpl_str() << ' ';
	} else {
		fpl << wptdest.get_icao();
	}
	{
		unsigned int tm(wptdest.get_time_unix() - wptdep.get_time_unix());
		tm += 30;
		tm /= 60;
		unsigned int hr(tm / 60);
		tm -= hr * 60;
		fpl << std::setw(2) << std::setfill('0') << hr
		    << std::setw(2) << std::setfill('0') << tm
		    << " -" << otherinfo.str();
	}
	if (!eet.str().empty())
		fpl << "EET/" << eet.str();
	// add remark: POGO
	if (wptdep.is_ifr() && wptdest.is_ifr() && 
	    IcaoFlightPlan::is_route_pogo(wptdep.get_icao(), wptdest.get_icao()))
		fpl << " RMK/POGO";
	fpl << ')';
	return fpl.str();
}

void CFMUAutoroute::updatefpl(void)
{
	m_routefuel = m_depfuelpenalty;
	double rttime(0);
	time_t tm(m_route.get_time_offblock_unix());
	tm += 5 * 60;
	unsigned int prevpi(0);
	for (unsigned int i = 0; i < m_route.get_nrwpt(); ++i) {
		FPlanWaypoint& wpt(m_route[i]);
		wpt.set_qff_hpa(m_qnh);
		wpt.set_isaoffset_kelvin(m_isaoffs);
		if (i > 0) {
			unsigned int pi(findperfindex(wpt.get_altitude()));
			double tmadd(0);
			if (i == 1)
				tmadd = m_deptimepenalty;
			for (; prevpi < pi; ++prevpi) {
				const Performance& perf(m_performance[prevpi]);
				tmadd += perf.get_timepenalty();
				m_routefuel += perf.get_fuelpenalty();
			}
			prevpi = pi;
			const Performance& perf(m_performance[prevpi]);
			double x(m_route[i - 1].get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
			Point pthalfway(m_route[i - 1].get_coord().halfway(wpt.get_coord()));
			if (m_windenable) {
				Wind<double> wnd(perf.get_wind(pthalfway));
				wnd.set_crs_tas(m_route[i - 1].get_coord().spheric_true_course_dbl(wpt.get_coord()), perf.get_tas());
				x *= perf.get_inverse_tas() * wnd.get_gs();
				wpt.set_winddir_deg(wnd.get_winddir());
				wpt.set_windspeed_kts(wnd.get_windspeed());
			}
			x *= perf.get_secpernmi();
			tmadd += x;
			m_routefuel += x * perf.get_fuelpersec();
			tm += Point::round<time_t,double>(tmadd);
			rttime += tmadd;
			if (i + 1 < m_route.get_nrwpt()) {
				double temp(perf.get_temp(wpt.get_coord()));
				if (!std::isnan(temp))
					wpt.set_oat_kelvin(temp);
			}
			if (m_grib2prmsl)
				wpt.set_qff_hpa(m_grib2prmsl->operator()(wpt.get_coord()) * 0.01f);
		}
		wpt.set_time_unix(tm);
	}
	tm += 5 * 60;
	m_route.set_time_onblock_unix(tm);
	m_routetime = Point::round<unsigned int,double>(rttime);
}

bool CFMUAutoroute::is_running(void) const
{
	return m_running;
}

bool CFMUAutoroute::is_errorfree(void) const
{
	return m_done;
}

double CFMUAutoroute::get_gcdistance(void) const
{
	if (!m_airport[0].is_valid() || !m_airport[1].is_valid())
		return std::numeric_limits<double>::quiet_NaN();
	return m_airport[0].get_coord().spheric_distance_nmi_dbl(m_airport[1].get_coord());
}

double CFMUAutoroute::get_routedistance(void) const
{
	if (m_route.get_nrwpt())
		return m_route.total_distance_nmi();
	return std::numeric_limits<double>::quiet_NaN();
}

unsigned int CFMUAutoroute::get_routetime(void) const
{
	if (m_route.get_nrwpt())
		return m_routetime;
	return 0;
}

double CFMUAutoroute::get_routefuel(void) const
{
	if (m_route.get_nrwpt())
		return m_routefuel;
	return std::numeric_limits<double>::quiet_NaN();
}

unsigned int CFMUAutoroute::get_mintime(void) const
{
	if (!m_airport[0].is_valid() || !m_airport[1].is_valid() || m_performance.empty())
		return 0;
	double dist(m_airport[0].get_coord().spheric_distance_nmi_dbl(m_airport[1].get_coord()));
	double crs(m_airport[0].get_coord().spheric_true_course_dbl(m_airport[1].get_coord()));
	Point pthalfway(m_airport[0].get_coord().halfway(m_airport[1].get_coord()));
	double x(std::numeric_limits<double>::max());
	double climb(0);
	for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
		const Performance& perf(m_performance[pi]);
		double y(dist);
		if (m_windenable) {
			Wind<double> wnd(perf.get_wind(pthalfway));
			wnd.set_crs_tas(crs, perf.get_tas());
			y *= perf.get_inverse_tas() * wnd.get_gs();
		}
		y *= perf.get_secpernmi();
		y += climb;
		x = std::min(x, y);
		climb += perf.get_timepenalty();
	}
	if (x == std::numeric_limits<double>::max())
		return 0;
	return Point::round<unsigned int, double>(x + m_deptimepenalty);
}

double CFMUAutoroute::get_minfuel(void) const
{
	if (!m_airport[0].is_valid() || !m_airport[1].is_valid() || m_performance.empty())
		return std::numeric_limits<double>::quiet_NaN();
	double dist(m_airport[0].get_coord().spheric_distance_nmi_dbl(m_airport[1].get_coord()));
	double x(std::numeric_limits<double>::max());
	double climb(0);
	for (unsigned int pi = 0; pi < m_performance.size(); ++pi) {
		const Performance& perf(m_performance[pi]);
		double y(climb + dist * perf.get_secpernmi() * perf.get_fuelpersec());
		x = std::min(x, y);
		climb += perf.get_fuelpenalty();
	}
	if (x == std::numeric_limits<double>::max())
		return std::numeric_limits<double>::quiet_NaN();
	return x + m_depfuelpenalty;
}

void CFMUAutoroute::loadtfr(void)
{
	if (m_tfrloaded)
		return;
	if (!m_tfrenable)
		return;
	opendb();
	std::vector<TrafficFlowRestrictions::Message> msg;
	Glib::TimeVal tv;
	tv.assign_current_time();
	bool ok(false);
	std::string fname(get_tfr_bin_filename());
	if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
		m_signal_log(log_normal, "Loading traffic flow restrictions " + fname);
		ok = m_tfr.load_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
	}
	if (!ok) {
		fname = get_tfr_xml_filename();
		m_signal_log(log_normal, "Loading traffic flow restrictions " + fname);
		ok = m_tfr.load_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
	}
	{
		Glib::TimeVal tv1;
		tv1.assign_current_time();
		tv = tv1 - tv;
	}
	for (unsigned int i = 0; i < msg.size(); ++i)
		logmessage(msg[i]);
	if (!ok) {
		m_tfravailable = false;
		set_tfr_enabled(false);
		m_signal_log(log_normal, "Cannot load traffic flow restrictions " + get_tfr_bin_filename() + ", " + get_tfr_xml_filename());
	} else {
		m_tfr.reset_rules();
		{
			std::ostringstream oss;
			oss << m_tfr.count_rules() << " rules loaded traffic flow restrictions " << fname
			    << ", created " << m_tfr.get_created() << ", effective " << m_tfr.get_effective()
			    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
			m_signal_log(log_normal, oss.str());
		}
	}
	{
		std::string fname(get_tfr_global_rules_filename());
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			if (m_tfr.load_disabled_trace_rules_file(fname)) {
				std::ostringstream oss;
				oss << "Loaded global disabled/trace rules from " << fname;
				m_signal_log(log_normal, oss.str());
			}
		}
	}
	{
		std::string fname(get_tfr_local_rules_filename());
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			if (m_tfr.load_disabled_trace_rules_file(fname)) {
				std::ostringstream oss;
				oss << "Loaded user disabled/trace rules from " << fname;
				m_signal_log(log_normal, oss.str());
			}
		}
	}
	m_tfrloaded = true;
}

void CFMUAutoroute::tfrpreparedct(void)
{
	Glib::TimeVal tv;
	tv.assign_current_time();
	m_tfr.prepare_dct_rules();
	{
		Glib::TimeVal tv1;
		tv1.assign_current_time();
		tv = tv1 - tv;
	}
	std::ostringstream oss;
	oss << m_tfr.count_dct_rules() << " DCT rules, "
	    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
	m_signal_log(log_normal, oss.str());
}

void CFMUAutoroute::preload(void)
{
	child_run();
	loadtfr();
	if (m_windenable && !m_grib2loaded) {
		GRIB2::Parser p(m_grib2);
		std::ostringstream oss;
		std::string path(m_dbdir_main);
		if (path.empty())
			path = FPlan::get_userdbpath();
		path = Glib::build_filename(path, "gfs");
		p.parse(path);
		unsigned int count1(m_grib2.find_layers().size());
		oss << "Loaded " << count1 << " GRIB2 Layers from " << path;
		if (!m_dbdir_aux.empty()) {
			p.parse(m_dbdir_aux);
			unsigned int count2(m_grib2.find_layers().size() - count1);
			oss << " and " << count2 << " GRIB2 Layers from " << m_dbdir_aux;
		}
		m_signal_log(log_normal, oss.str());
		m_grib2loaded = true;
	}
}

void CFMUAutoroute::start(void)
{
	start(false);
}

void CFMUAutoroute::cont(void)
{
	start(true);
}

void CFMUAutoroute::start(bool cont)
{
	stop();
	child_run();
	m_done = false;
	time(&m_starttime);
	bool ok(get_departure().is_valid() && get_destination().is_valid());
	bool perferr(false);
	if (ok) {
		computeperformance();
		perferr = m_performance.empty();
		ok = ok && !perferr;
	}
	if (!ok) {
		m_running = true;
		m_signal_statuschange(statusmask_starting);
		if (perferr)
			m_signal_log(log_normal, "Performance calculation error - aircraft unable to climb to minimum level");
		clear();
		return;
	}
	if (!(m_airportifr[0] || m_airportifr[1] || m_forceenrifr)) {
		clear();
		m_running = true;
		m_signal_statuschange(statusmask_starting);
		m_signal_log(log_normal, "Starting VFR router");
		m_iterationcount[0] = m_iterationcount[1] = 0;
		m_done = start_vfr();
		if (m_done) {
			m_iterationcount[0] = 1;
			m_signal_log(log_fplproposal, get_simplefpl());
		}
		stop();
		return;
	}
	if (!start_ifr(cont))
		stop();
}

void CFMUAutoroute::precompute_graph(const Rect& bbox, const std::string& fn)
{
	m_performance.clear();
	{
		int level(m_levels[0]);
		level = ((level + 9) / 10) * 10;
		for (; level <= m_levels[1]; level += 10)
			m_performance.push_back(Performance(level * 100, 1, 1, 0, 0));
	}
	loadtfr();
	tfrpreparedct();
	//set_dctlimit(50);
	set_opttarget(opttarget_time);
	clearlgraph();
	if (!lgraphload(bbox))
		throw std::runtime_error("cannot load graph");
	lgraphkillinvalidsuper();
	lgraphadddct();
	if (true) {
		std::ostringstream oss;
		oss << "Airway graph after DCT: " << boost::num_vertices(m_lgraph) << " vertices, "
		    << boost::num_edges(m_lgraph) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	std::ofstream os(fn.c_str(), std::ofstream::binary);
	if (!os)
		throw std::runtime_error("cannot create file " + fn);
	lgraphsave(os, bbox);
}

void CFMUAutoroute::computeperformance(void)
{
	m_grib2prmsl = Glib::RefPtr<GRIB2::LayerResult>();
	const GRIB2::Parameter *grib2windu(0);
	const GRIB2::Parameter *grib2windv(0);
	const GRIB2::Parameter *grib2temp(0);
	double halfgcdist(0);
	Rect bbox;
	if (m_windenable) {
		grib2windu = GRIB2::find_parameter(GRIB2::param_meteorology_momentum_ugrd);
		grib2windv = GRIB2::find_parameter(GRIB2::param_meteorology_momentum_vgrd);
		grib2temp = GRIB2::find_parameter(GRIB2::param_meteorology_temperature_tmp);
		halfgcdist = 0.5 * get_gcdistance();
		bbox = get_bbox().oversize_nmi(100.f);
	}
	m_performance.clear();
	m_deptimepenalty = 0;
	m_depfuelpenalty = 0;
	int level(m_levels[0]);
	if (m_airportifr[0] || m_airportifr[1] || m_forceenrifr)
		level = ((level + 9) / 10) * 10;
	else
		level = ((level + 4) / 10) * 10 + 5;
	Aircraft::Climb& climb(m_aircraft.get_climb());
	Aircraft::Cruise& cruise(m_aircraft.get_cruise());
	AirData ad;
	ad.set_qnh_tempoffs(m_qnh, m_isaoffs);
	while (level <= m_levels[1]) {
		ad.set_pressure_altitude(level * 100);
		if (ad.get_density_altitude() > climb.get_ceiling())
			break;
		double tas(0), fuelflow(0);
		// compute cruise performance
                {
                        double bhp(get_engine_bhp()), rpm(get_engine_rpm()), mp(get_engine_mp()), da(ad.get_density_altitude());
                        cruise.calculate(tas, fuelflow, bhp, rpm, mp, da);
                        m_signal_log(log_normal, Glib::ustring::format("Cruise: DA ", std::fixed, std::setprecision(0), da) + 
				     Glib::ustring::format(" TAS ", std::fixed, std::setprecision(0), tas) + 
				     Glib::ustring::format(" Fuelflow ", std::fixed, std::setprecision(1), fuelflow) + 
				     Glib::ustring::format(" RPM ", std::fixed, std::setprecision(0), rpm) + 
				     Glib::ustring::format(" MP ", std::fixed, std::setprecision(1), mp) + 
				     Glib::ustring::format(" BHP ", std::fixed, std::setprecision(0), bhp) + 
				     Glib::ustring::format(" Fuel/NMI ", std::fixed, std::setprecision(4), fuelflow / tas));
                }
		if (tas <= 0)
			break;
		double secpernmi(3600.0 / tas);
		double ff(fuelflow * (1.0 / 3600.0));
		double timepenalty(0), fuelpenalty(0);
		bool climberr(false);
		// compute climb performance
		{
			double t0(climb.altitude_to_time(ad.get_density_altitude()));
			ad.set_pressure_altitude(level * 100 + 1000);
			double t1(climb.altitude_to_time(ad.get_density_altitude()));
			climberr = climb.time_to_climbrate(t1) < 100;
			if (!climberr) {
				timepenalty = t1 - t0;
				double cldist(climb.time_to_distance(t1) - climb.time_to_distance(t0));
				if (cldist > 0) {
					timepenalty = std::max(timepenalty - cldist * secpernmi, 0.);
					fuelpenalty = climb.time_to_fuel(t1) - climb.time_to_fuel(t0);
					fuelpenalty = std::max(fuelpenalty - cldist * secpernmi * ff, 0.);
				}				
			}
		}
		// preferred level model
		double prefpenalty(pow(get_preferred_penalty(), abs(level - get_preferred_level()) * 0.01));
		// wind
		time_t t(get_deptime() + 5 * 60 + halfgcdist * secpernmi);
		float press(0);
		IcaoAtmosphere::std_altitude_to_pressure(&press, 0, level * (100 * Point::ft_to_m));
		press *= 100;
		if (m_performance.empty()) {
			GRIB2::layerlist_t ll(m_grib2.find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_prmsl), t));
                        Glib::RefPtr<GRIB2::LayerInterpolateResult> prmsl(GRIB2::interpolate_results(bbox, ll, t));
			if (prmsl)
				m_grib2prmsl = prmsl->get_results(t);
		}
		Glib::RefPtr<GRIB2::LayerResult> windu, windv, temp;
		if (grib2windu && grib2windv) {
                        {
                                GRIB2::layerlist_t ll(m_grib2.find_layers(grib2windu, t, GRIB2::surface_isobaric_surface, press));
				Glib::RefPtr<GRIB2::LayerInterpolateResult> intwindu(GRIB2::interpolate_results(bbox, ll, t, press));
				if (intwindu)
					windu = intwindu->get_results(t, press);
                        }
                        {
                                GRIB2::layerlist_t ll(m_grib2.find_layers(grib2windv, t, GRIB2::surface_isobaric_surface, press));
                                Glib::RefPtr<GRIB2::LayerInterpolateResult> intwindv(GRIB2::interpolate_results(bbox, ll, t, press));
				if (intwindv)
					windv = intwindv->get_results(t, press);
                        }
		}
		if (grib2temp) {
			GRIB2::layerlist_t ll(m_grib2.find_layers(grib2temp, t, GRIB2::surface_isobaric_surface, press));
			Glib::RefPtr<GRIB2::LayerInterpolateResult> inttemp(GRIB2::interpolate_results(bbox, ll, t, press));
			if (inttemp)
				temp = inttemp->get_results(t, press);
		}
		m_performance.push_back(Performance(level * 100, secpernmi, ff, timepenalty, fuelpenalty, prefpenalty, windu, windv));
		m_signal_log(log_normal, Glib::ustring::format("Climb: DA ", std::fixed, std::setprecision(0), ad.get_density_altitude()) + 
			     Glib::ustring::format(" dTime ", std::fixed, std::setprecision(1), timepenalty) + 
			     Glib::ustring::format(" dFuel ", std::fixed, std::setprecision(3), fuelpenalty));
		if (climberr)
			break;
		level += 10;
	}
	if (m_performance.empty())
		return;
	// compute departure climb performance
	{
		double t0(climb.altitude_to_time(m_airport[0].get_elevation()));
		double t1(climb.altitude_to_time(m_performance[0].get_altitude()));
		m_deptimepenalty = t1 - t0;
		double cldist(climb.time_to_distance(t1) - climb.time_to_distance(t0));
		if (cldist > 0) {
			m_deptimepenalty = std::max(m_deptimepenalty - cldist * m_performance[0].get_secpernmi(), 0.);
			m_depfuelpenalty = climb.time_to_fuel(t1) - climb.time_to_fuel(t0);
			m_depfuelpenalty = std::max(m_depfuelpenalty - cldist * m_performance[0].get_secpernmi() * m_performance[0].get_fuelpersec(), 0.);
		}				
	}

}

unsigned int CFMUAutoroute::findperfindex(int alt) const
{
	int adiff(std::numeric_limits<int>::max());
	unsigned int idx(0);
	for (unsigned int i = 0; i < m_performance.size(); ++i) {
		int ad(abs(alt - m_performance[i].get_altitude()));
		if (ad > adiff)
			continue;
		adiff = ad;
		idx = i;
	}
	return idx;
}

bool CFMUAutoroute::start_vfr(void)
{
	VFRRouter r(*this);
	if (!r.load_graph(get_bbox())) {
		std::ostringstream oss;
		oss << "VFR router: cannot load routing graph for bbox " << get_bbox();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	{
		std::ostringstream oss;
		oss << "VFR area graph after loading: " << boost::num_vertices(r.get_graph()) << " vertices, "
		    << boost::num_edges(r.get_graph()) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	if (!r.set_departure(get_departure())) {
		std::ostringstream oss;
		oss << "VFR router: cannot find departure airport " << get_departure().get_icao_name();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	if (!r.set_destination(get_destination())) {
		std::ostringstream oss;
		oss << "VFR router: cannot find destination airport " << get_destination().get_icao_name();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	bool havesid(r.add_sid(get_sid()));
	if (!havesid && !get_sid().is_invalid()) {
		std::ostringstream oss;
		oss << "VFR router: cannot find SID point " << get_sid().get_lat_str2() << ' ' << get_sid().get_lon_str2();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	havesid = havesid || r.add_sid(get_departure());
	bool havestar(r.add_star(get_star()));
	if (!havestar && !get_star().is_invalid()) {
		std::ostringstream oss;
		oss << "VFR router: cannot find STAR point " << get_star().get_lat_str2() << ' ' << get_star().get_lon_str2();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	havestar = havestar || r.add_star(get_destination());
	{
		std::ostringstream oss;
		oss << "VFR area graph after SID/STAR: " << boost::num_vertices(r.get_graph()) << " vertices, "
		    << boost::num_edges(r.get_graph()) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	r.add_dct(get_dctlimit(),
		  havesid ? std::numeric_limits<double>::quiet_NaN() : get_sidlimit(),
		  havestar ? std::numeric_limits<double>::quiet_NaN() : get_starlimit());
	{
		std::ostringstream oss;
		oss << "VFR area graph after DCT: " << boost::num_vertices(r.get_graph()) << " vertices, "
		    << boost::num_edges(r.get_graph()) << " edges";
		m_signal_log(log_debug0, oss.str());
	}
	r.set_metric();
	r.exclude_airspace(m_levels[0], m_levels[1], get_vfraspclimit());
	r.exclude_regions();
	if (!r.autoroute(m_route)) {
		std::ostringstream oss;
		oss << "VFR router: cannot find route " << get_departure().get_icao_name()
		    << "->" << get_destination().get_icao_name();
		m_signal_log(log_debug0, oss.str());
		return false;
	}
	{
		const Performance& perf(m_performance[m_performance.size() / 2U]);
		for (unsigned int i = 0; i < m_route.get_nrwpt(); ++i) {
			FPlanWaypoint& wpt(m_route[i]);
			uint16_t flags(wpt.get_flags());
			flags &= ~FPlanWaypoint::altflag_ifr;
			if (i > 0 && i + 1 < m_route.get_nrwpt()) {
				flags |= FPlanWaypoint::altflag_standard;
				wpt.set_altitude(perf.get_altitude());
			}
			wpt.set_flags(flags);
		}
	}
	updatefpl();
	return true;
}

bool CFMUAutoroute::start_ifr(bool cont, bool optfuel)
{
	if (start_ifr_specialcase()) {
		m_running = true;
		m_signal_statuschange(statusmask_starting);
		m_signal_log(log_normal, "Starting special case IFR router");
		m_done = true;
		m_iterationcount[0] = 1;
		m_signal_log(log_fplproposal, get_simplefpl());
		return false;
	}
	cont = cont && (m_superdep < m_lsupervertices.size())
		&& (m_superdest < m_lsupervertices.size())
		&& boost::num_vertices(m_lgraph);
	if (!cont) {
		clear();
		m_callsign.clear();
		for (unsigned int i = 0; i < 5; ++i) {
			char ch('A' + (rand() % ('Z' - 'A' + 1)));
			m_callsign.push_back(ch);
		}
		preload();
		if (get_tfr_enabled()) {
			//m_tfr.tracerules_add("LS1A");
			//m_tfr.tracerules_add("ED1A");
			m_tfr.reset_rules();
			{
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules, created " << m_tfr.get_created() << ", effective " << m_tfr.get_effective();
				m_signal_log(log_normal, oss.str());
			}
			{
				std::ostringstream oss;
				oss << "Rules: Disabled: " << get_tfr_disable() << " Trace: " << get_tfr_trace();
				m_signal_log(log_normal, oss.str());
			}
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				m_tfr.simplify_rules_typeofflight('G');
				m_tfr.simplify_rules_mil(false);
				m_tfr.simplify_rules_gat(true);
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing type of flight G CIV GAT, "
				    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_signal_log(log_normal, oss.str());
			}
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				if (m_airport[0].is_valid() && m_airport[0].get_icao().size() == 4 && m_airport[0].get_icao() != "ZZZZ")
					m_tfr.simplify_rules_dep(m_airport[0].get_icao(), m_airport[0].get_coord());
				if (m_airport[1].is_valid() && m_airport[1].get_icao().size() == 4 && m_airport[1].get_icao() != "ZZZZ")
					m_tfr.simplify_rules_dest(m_airport[1].get_icao(), m_airport[1].get_coord());
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing departure " << m_airport[0].get_icao()
				    << " destination " << m_airport[1].get_icao()
				    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_signal_log(log_normal, oss.str());
			}
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				m_tfr.simplify_rules_aircrafttype(m_aircraft.get_icaotype());
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing aircraft type " << m_aircraft.get_icaotype()
				    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_signal_log(log_normal, oss.str());
			}
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				m_tfr.simplify_rules_equipment(m_aircraft.get_equipment());
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing aircraft equipment " << m_aircraft.get_equipment()
				    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_signal_log(log_normal, oss.str());
			}
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				m_tfr.simplify_rules_altrange(m_levels[0] * 100, m_levels[1] * 100);
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing cruise levels F"
				    << std::setw(3) << std::setfill('0') << m_levels[0] << "..F"
				    << std::setw(3) << std::setfill('0') << m_levels[1]
				    << ", " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_signal_log(log_normal, oss.str());
			}
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				Rect bbox(get_bbox().oversize_nmi(200.f));
				m_tfr.simplify_rules_bbox(bbox);
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing area ["
				    << bbox.get_southwest().get_lat_str2() << ' ' << bbox.get_southwest().get_lon_str2() << ' '
				    << bbox.get_northeast().get_lat_str2() << ' ' << bbox.get_northeast().get_lon_str2()
				    << "], " << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_signal_log(log_normal, oss.str());
			}
			tfrpreparedct();
		} else {
			m_signal_log(log_normal, "Traffic flow restrictions disabled");
		}
	}
	m_running = true;
	m_signal_statuschange(statusmask_starting);
	m_signal_log(log_normal, cont ? "Restarting IFR router" : "Starting IFR router");
	if (!cont) {
		m_iterationcount[0] = m_iterationcount[1] = 0;
		{
			clearlgraph();
			Rect bbox(get_bbox());
			bbox = bbox.oversize_nmi(100.f);
			if (!(get_precompgraph_enabled() && lgraphbinload(bbox))) {
				clearlgraph();
				if (!lgraphload(bbox))
					return false;
				lgraphkillinvalidsuper();
				lgraphadddct();
			}
		}
		if (true) {
			std::ostringstream oss;
			oss << "Airway graph after DCT: " << boost::num_vertices(m_lgraph) << " vertices, "
			    << boost::num_edges(m_lgraph) << " edges";
			m_signal_log(log_debug0, oss.str());
		}
		if (!lgraphaddsidstar()) {
			if (true)
				m_signal_log(log_debug0, "Cannot add SID/STAR to Airway graph");
			return false;
		}
		if (true) {
			std::ostringstream oss;
			oss << "Airway graph after SID/STAR: " << boost::num_vertices(m_lgraph) << " vertices, "
			    << boost::num_edges(m_lgraph) << " edges";
	       		m_signal_log(log_debug0, oss.str());
		}
		lgraphexcluderegions();
		if (true) {
			std::ostringstream oss;
			oss << "Airway graph after manually excluded regions: " << boost::num_vertices(m_lgraph) << " vertices, "
			    << boost::num_edges(m_lgraph) << " edges";
	       		m_signal_log(log_debug0, oss.str());
		}
		lgraphcrossing();
	}
	lgraphmodified();
	lgraphroute();
	return true;
}

void CFMUAutoroute::stop(void)
{
	m_childtimeoutcount = 0;
	m_connchildtimeout.disconnect();
	m_pathprobe = -1;
	if (m_running) {
		m_running = false;
		if (m_done)
			m_signal_statuschange(statusmask_stoppingdone);
		else
			m_signal_statuschange(statusmask_stoppingerror);
	}
}

void CFMUAutoroute::clear(void)
{
	stop();
	while (m_route.get_nrwpt())
		m_route.delete_wpt(0);
	m_routetime = 0;
	m_routefuel = 0;
	m_validationresponse.clear();
	clearlgraph();
	m_done = false;
}

bool CFMUAutoroute::is_stopped(void)
{
	// dispatch events, check for stop command
	if (true) {
		Glib::RefPtr<Glib::MainContext> ctx(Glib::MainContext::get_default());
		while (ctx->iteration(false));
	}
	if (m_running)
		return false;
	m_signal_log(log_debug0, "Interrupted by user, stopping...");
	return true;
}

void CFMUAutoroute::child_run(void)
{
	if (m_childrun)
		return;
	m_signal_log(log_normal, "Connecting to validation server...");
	int childstdin(-1), childstdout(-1);
	try {
		std::vector<std::string> argv, env;
#ifdef __MINGW32__
		std::string workdir;
		{
			gchar *instdir(g_win32_get_package_installation_directory_of_module(0));
			if (instdir) {
				workdir = instdir;
				argv.push_back(Glib::build_filename(instdir, "bin", "cfmuvalidateserver.exe"));
			} else {
				argv.push_back(Glib::build_filename(PACKAGE_PREFIX_DIR "/bin", "cfmuvalidateserver.exe"));
				workdir = PACKAGE_PREFIX_DIR;
			}
		}
#else
		std::string workdir(PACKAGE_DATA_DIR);
		argv.push_back(Glib::build_filename(PACKAGE_LIBEXEC_DIR, "cfmuvalidateserver"));
		env.push_back("PATH=/bin:/usr/bin");
		{
			bool found(false);
			std::string disp(Glib::getenv("DISPLAY", found));
			if (found)
				env.push_back("DISPLAY=" + disp);
		}
#endif
		if (false) {
			std::cerr << "Spawn: workdir " << workdir << " argv:" << std::endl;
			for (std::vector<std::string>::const_iterator i(argv.begin()), e(argv.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
			std::cerr << "env:" << std::endl;
			for (std::vector<std::string>::const_iterator i(env.begin()), e(env.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
		}
#ifdef __MINGW32__
		// inherit environment; otherwise, this may fail on WindowsXP with STATUS_SXS_ASSEMBLY_NOT_FOUND (C0150004)
		Glib::spawn_async_with_pipes(workdir, argv,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_childpid, &childstdin, &childstdout, 0);
#else
		Glib::spawn_async_with_pipes(workdir, argv, env,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_childpid, &childstdin, &childstdout, 0);
#endif
		m_childrun = true;
	} catch (Glib::SpawnError& e) {
		m_childrun = false;
		m_signal_log(log_normal, "Cannot run validator: " + e.what());
		return;
	}
	m_connchildwatch = Glib::signal_child_watch().connect(sigc::mem_fun(this, &CFMUAutoroute::child_watch), m_childpid);
	m_childchanstdin = Glib::IOChannel::create_from_fd(childstdin);
	m_childchanstdout = Glib::IOChannel::create_from_fd(childstdout);
	m_childchanstdin->set_encoding();
	m_childchanstdout->set_encoding();
	m_childchanstdin->set_close_on_unref(true);
	m_childchanstdout->set_close_on_unref(true);
	m_connchildstdout = Glib::signal_io().connect(sigc::mem_fun(this, &CFMUAutoroute::child_stdout_handler), m_childchanstdout,
						      Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
}

bool CFMUAutoroute::child_stdout_handler(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(m_childchanstdout->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
		if (iostat == Glib::IO_STATUS_ERROR)
			m_signal_log(log_debug0, "Child stdout error, stopping...");
		else
			m_signal_log(log_debug0, "Child stdout eof, stopping...");
                child_close();
		stop();
		return true;
        }
	{
		Glib::ustring::size_type n(line.size());
		while (n > 0) {
			--n;
			if (!isspace(line[n]) && line[n] != '\n' && line[n] != '\r') {
				++n;
				break;
			}
		}
		line = Glib::ustring(line, 0, n);
	}
	if (!line.empty())
		m_validationresponse.push_back(line);
	m_signal_log(log_fplvalidation, line);
	if (line.empty()) {
		parse_response();
	}
        return true;
}

void CFMUAutoroute::child_watch(GPid pid, int child_status)
{
	if (!m_childrun)
		return;
	if (m_childpid != pid) {
		Glib::spawn_close_pid(pid);
		return;
	}
	Glib::spawn_close_pid(pid);
	m_childrun = false;
	m_connchildwatch.disconnect();
	m_connchildstdout.disconnect();
	child_close();
	child_timeout();
}

void CFMUAutoroute::child_close(void)
{
	m_connchildstdout.disconnect();
	if (m_childchanstdin) {
		m_childchanstdin->close();
		m_childchanstdin.reset();
	}
	if (m_childchanstdout) {
		m_childchanstdout->close();
		m_childchanstdout.reset();
	}
	if (!m_childrun)
		return;
	m_childrun = false;
	m_connchildwatch.disconnect();
	Glib::signal_child_watch().connect(sigc::hide(sigc::ptr_fun(&Glib::spawn_close_pid)), m_childpid);
}

Rect CFMUAutoroute::get_bbox(void) const
{
	if (!get_departure().is_valid() && !get_destination().is_valid())
		return Rect();
	if (!get_departure().is_valid())
		return Rect(get_destination().get_coord(), get_destination().get_coord());
	if (!get_destination().is_valid())
		return Rect(get_departure().get_coord(), get_departure().get_coord());
	Point pt1(get_departure().get_coord());
	Point pt2(get_destination().get_coord() - pt1);
	return Rect(Point(std::min(pt2.get_lon(), (Point::coord_t)0), std::min(pt2.get_lat(), (Point::coord_t)0)) + pt1,
		    Point(std::max(pt2.get_lon(), (Point::coord_t)0), std::max(pt2.get_lat(), (Point::coord_t)0)) + pt1);
}


void CFMUAutoroute::logmessage(const TrafficFlowRestrictions::Message& msg)
{
	std::ostringstream oss;
	oss << msg.get_type_char() << ':';
	if (!msg.get_rule().empty())
		oss << " R:" << msg.get_rule();
	if (!msg.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(msg.get_vertexset().begin()), e(msg.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!msg.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(msg.get_edgeset().begin()), e(msg.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	oss << ' ' << msg.get_text();
	log_t lt(log_debug1);
	switch (msg.get_type()) {
	case TrafficFlowRestrictions::Message::type_error:
		lt = log_normal;
		break;

	case TrafficFlowRestrictions::Message::type_warning:
		lt = log_debug0;
		break;

	default:
		lt = log_debug1;
		break;
	}
	m_signal_log(lt, oss.str());
}

CFMUAutoroute::VFRRouter::VFRRouter(CFMUAutoroute& ar)
	: m_ar(ar)
{
}

bool CFMUAutoroute::VFRRouter::load_graph(const Rect& bbox)
{
	m_bbox = bbox.oversize_nmi(100.f);
	if (false) {
		std::ostringstream oss;
		oss << "Loading VFR area graph: " << m_bbox;
		m_ar.m_signal_log(log_debug0, oss.str());
	}
	m_ar.opendb();
	m_graph.clear();
	m_graph.add(m_ar.m_airportdb.find_by_rect(m_bbox, ~0, AirportsDb::Airport::subtables_none));
	m_graph.add(m_ar.m_navaiddb.find_by_rect(m_bbox, ~0, NavaidsDb::Navaid::subtables_none));
	m_graph.add(m_ar.m_waypointdb.find_by_rect(m_bbox, ~0, WaypointsDb::Waypoint::subtables_none));
	m_graph.add(m_ar.m_mapelementdb.find_by_rect(m_bbox, ~0, MapelementsDb::Mapelement::subtables_none));
	// We do not need airways for VFR routing
	//m_graph.add(m_ar.m_airwaydb.find_by_area(m_bbox, ~0, AirwaysDb::Airway::subtables_all));
	//m_graph.add(m_ar.m_airwaydb.find_by_rect(m_bbox, ~0, AirwaysDb::Airway::subtables_all));
	if (true) {
		if (false) {
			std::ostringstream oss;
			oss << "VFR area graph: " << m_bbox << ", "
			    << boost::num_vertices(m_graph) << " vertices, " << boost::num_edges(m_graph) << " edges";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
		unsigned int counts[4];
		for (unsigned int i = 0; i < sizeof(counts)/sizeof(counts[0]); ++i)
			counts[i] = 0;
		Graph::vertex_iterator u0, u1;
		for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
			const Vertex& uu(m_graph[*u0]);
			unsigned int i(uu.get_type());
			if (i < sizeof(counts)/sizeof(counts[0]))
				++counts[i];
		}
		{
			std::ostringstream oss;
			oss << "VFR area graph: " << counts[Vertex::type_airport] << " airports, " << counts[Vertex::type_navaid] << " navaids, "
			    << counts[Vertex::type_intersection] << " intersections, " << counts[Vertex::type_mapelement] << " mapelements";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
	}
	return true;
}

CFMUAutoroute::VFRRouter::Graph::vertex_descriptor CFMUAutoroute::VFRRouter::find_airport(const AirportsDb::Airport& arpt)
{
	if (!arpt.is_valid())
		return boost::num_vertices(m_graph);
	Graph::vertex_iterator u0, u1;
	for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
		const Vertex& uu(m_graph[*u0]);
		AirportsDb::Airport a;
		if (!uu.get(a))
			continue;
		if (!a.is_valid())
			continue;
		if (arpt.get_icao() != a.get_icao() ||
		    arpt.get_name() != a.get_name() ||
		    arpt.get_coord() != a.get_coord())
			continue;
		return *u0;
	}
	return boost::num_vertices(m_graph);
}

bool CFMUAutoroute::VFRRouter::set_departure(const AirportsDb::Airport& arpt)
{
	m_dep = find_airport(arpt);
	return m_dep < boost::num_vertices(m_graph);
}

bool CFMUAutoroute::VFRRouter::set_destination(const AirportsDb::Airport& arpt)
{
	m_dest = find_airport(arpt);
	return m_dest < boost::num_vertices(m_graph);
}

void CFMUAutoroute::VFRRouter::add_dct(double dctlim, double sidlim, double starlim)
{
	Graph::vertex_iterator u0, u1;
	for (boost::tie(u0, u1) = boost::vertices(m_graph); u0 != u1; ++u0) {
		Graph::vertex_descriptor u(*u0);
		const Vertex& uu(m_graph[u]);
		Graph::vertex_iterator v0, v1;
		for (boost::tie(v0, v1) = boost::vertices(m_graph); v0 != v1; ++v0) {
			Graph::vertex_descriptor v(*v0);
			if (u == v)
				continue;
			const Vertex& vv(m_graph[*v0]);
			bool kill(u == m_dep && v == m_dest);
			kill = kill || (uu.get_type() == Vertex::type_intersection && uu.is_ident_numeric());
			kill = kill || (vv.get_type() == Vertex::type_intersection && vv.is_ident_numeric());
			double dist, lim(dctlim);
			if (!kill) {
				dist = uu.get_coord().spheric_distance_nmi_dbl(vv.get_coord());
				if (u == m_dep)
					lim = sidlim;
				if (v == m_dest)
					lim = starlim;
				if (!std::isnan(lim))
					kill = dist > lim;
			}
			if (kill) {
				for (;;) {
					bool done(true);
					Graph::out_edge_iterator e0, e1;
					for (boost::tie(e0, e1) = boost::out_edges(u, m_graph); e0 != e1;) {
						Graph::out_edge_iterator ex(e0);
						++e0;
						if (boost::target(*ex, m_graph) != v)
							continue;
						boost::remove_edge(ex, m_graph);
						if (lgraphrestartedgescan) {
							done = false;
							break;
						}
					}
					if (done)
						break;
				}
				continue;
			}
			if (std::isnan(lim))
				continue;
			Graph::edge_descriptor e;
			bool hase;
			tie(e, hase) = boost::edge(u, v, m_graph);
			if (hase)
				continue;
			boost::add_edge(u, v, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, dist), m_graph);
		}
	}
}

void CFMUAutoroute::VFRRouter::remove_sid(void)
{
	for (;;) {
		bool done(true);
		Graph::out_edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::out_edges(m_dep, m_graph); e0 != e1;) {
			Graph::out_edge_iterator ex(e0);
			++e0;
			boost::remove_edge(ex, m_graph);
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
}

bool CFMUAutoroute::VFRRouter::add_sid(const Point& pt)
{
	if (pt.is_invalid())
		return false;
	Graph::vertex_descriptor v;
	if (!m_graph.find_nearest(v, pt, Vertex::typemask_all))
		return false;
	if (v == m_dest)
		return false;
	const Vertex& vv(m_graph[v]);
	double dist(pt.spheric_distance_nmi_dbl(vv.get_coord()));
	if (dist > 0.1)
		return false;
	m_ar.m_airportconnident[0] = vv.get_ident();
	m_ar.m_airportconntype[0] = vv.get_type();
	// remove departure out edges
	remove_sid();
	// add edge
	boost::add_edge(m_dep, v, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, dist), m_graph);
	m_sid.clear();
	return true;
}

bool CFMUAutoroute::VFRRouter::add_sid(const AirportsDb::Airport& arpt)
{
	bool added(false);
	m_sid.clear();
	for (unsigned int i = 0, j = arpt.get_nrvfrrte(); i < j; ++i) {
		const AirportsDb::Airport::VFRRoute& rte(arpt.get_vfrrte(i));
		if (!rte.size() ||
		    rte.get_name().empty())
			continue;
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[rte.size()-1]);
		if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure)
			continue;
		Graph::vertex_descriptor v;
		{
			bool found(m_graph.find_nearest(v, rtept.get_coord(), Vertex::typemask_all));
			if (found) {
				const Vertex& vv(m_graph[v]);
				double dist(rtept.get_coord().spheric_distance_nmi_dbl(vv.get_coord()));
				if (dist > 0.1)
					found = false;
			}
			if (!found)
				v = boost::add_vertex(Vertex(rte.get_name(), rtept.get_coord(), Vertex::type_vfrreportingpt), m_graph);
		}
		if (!added)
			remove_sid();
		added = true;
		boost::add_edge(m_dep, v, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, rte.compute_distance()), m_graph);
		m_sid.push_back(rte);
	}
	return added;
}

void CFMUAutoroute::VFRRouter::remove_star(void)
{
	for (;;) {
		bool done(true);
		Graph::in_edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::in_edges(m_dest, m_graph); e0 != e1;) {
			Graph::in_edge_iterator ex(e0);
			++e0;
			boost::remove_edge(*ex, m_graph);
			if (lgraphrestartedgescan) {
				done = false;
				break;
			}
		}
		if (done)
			break;
	}
}

bool CFMUAutoroute::VFRRouter::add_star(const Point& pt)
{
	if (pt.is_invalid())
		return false;
	Graph::vertex_descriptor u;
	if (!m_graph.find_nearest(u, pt, Vertex::typemask_all))
		return false;
	if (u == m_dep)
		return false;
	const Vertex& uu(m_graph[u]);
	double dist(pt.spheric_distance_nmi_dbl(uu.get_coord()));
	if (dist > 0.1)
		return false;
	m_ar.m_airportconnident[1] = uu.get_ident();
	m_ar.m_airportconntype[1] = uu.get_type();
	// remove destination in edges
	remove_star();
	// add edge
	boost::add_edge(u, m_dest, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, dist), m_graph);
	m_star.clear();
	return true;
}

bool CFMUAutoroute::VFRRouter::add_star(const AirportsDb::Airport& arpt)
{
	bool added(false);
	m_star.clear();
	for (unsigned int i = 0, j = arpt.get_nrvfrrte(); i < j; ++i) {
		const AirportsDb::Airport::VFRRoute& rte(arpt.get_vfrrte(i));
		if (!rte.size() || rte.get_name().empty())
			continue;
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
		if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival)
			continue;
		Graph::vertex_descriptor v;
		{
			bool found(m_graph.find_nearest(v, rtept.get_coord(), Vertex::typemask_all));
			if (found) {
				const Vertex& vv(m_graph[v]);
				double dist(rtept.get_coord().spheric_distance_nmi_dbl(vv.get_coord()));
				if (dist > 0.1)
					found = false;
			}
			if (!found)
				v = boost::add_vertex(Vertex(rte.get_name(), rtept.get_coord(), Vertex::type_vfrreportingpt), m_graph);
		}
		if (!added)
			remove_star();
		added = true;
		boost::add_edge(v, m_dest, Edge("", 0, 600, AirwaysDb::Airway::airway_invalid, rte.compute_distance()), m_graph);
		m_star.push_back(rte);
	}
	return added;
}

void CFMUAutoroute::VFRRouter::set_metric(void)
{
	Graph::edge_iterator e0, e1;
	for (boost::tie(e0, e1) = boost::edges(m_graph); e0 != e1; ++e0) {
		Edge& ee(m_graph[*e0]);
		ee.set_metric(ee.get_dist());
	}
}

void CFMUAutoroute::VFRRouter::exclude_airspace(int baselevel, int toplevel, double maxarea)
{
	if (maxarea <= 0)
		return;
	AirspacesDb::elementvector_t airspaces(m_ar.m_airspacedb.find_by_rect(m_bbox, 0, AirspacesDb::element_t::subtables_none));
	baselevel *= 100;
	toplevel *= 100;
	for (AirspacesDb::elementvector_t::const_iterator ai(airspaces.begin()), ae(airspaces.end()); ai != ae; ++ai) {
		const AirspacesDb::Airspace& aspc(*ai);
		{
			char bc(aspc.get_bdryclass());
			uint8_t tc(aspc.get_typecode());
			switch (tc) {
			default:
				// do not care about airspace classes E, F, G
				if (bc < 'A' || bc > 'D') {
					if (false) {
						std::ostringstream oss;
						oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
							  << " altitude: " << baselevel << "..." << toplevel << " due to class" << std::endl;
						m_ar.m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				break;

			case AirspacesDb::Airspace::typecode_specialuse:
				// Special Use airspaces; only care about prohibited, reserved and danger areas
				if (bc != 'P' && bc != 'R' && bc != 'D') {
					if (false) {
						std::ostringstream oss;
						oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
						    << " altitude: " << baselevel << "..." << toplevel << " due to class" << std::endl;
						m_ar.m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				break;

			case AirspacesDb::Airspace::typecode_ead:
				// EAD airspaces; only care about prohibited, reserved and danger areas
                                if (bc != AirspacesDb::Airspace::bdryclass_ead_d_amc && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_d && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_d_other && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_p && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_r_amc && 
				    bc != AirspacesDb::Airspace::bdryclass_ead_r) {
					if (false) {
						std::ostringstream oss;
						oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
						    << " altitude: " << baselevel << "..." << toplevel << " due to class" << std::endl;
						m_ar.m_signal_log(log_debug0, oss.str());
					}
					continue;
				}
				break;
			}
		}
		if (baselevel >= aspc.get_altupr_corr() || toplevel <= aspc.get_altlwr_corr()) {
			if (false) {
				std::ostringstream oss;
				oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
				    << " due to altitude: " << baselevel << "..." << toplevel << " fully outside "
				    << aspc.get_altlwr_corr() << "..." << aspc.get_altupr_corr() << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		if (false && !(baselevel >= aspc.get_altlwr_corr() && toplevel <= aspc.get_altupr_corr())) {
			if (false) {
				std::ostringstream oss;
				oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
				    << " due to altitude: " << baselevel << "..." << toplevel << " not inside "
				    << aspc.get_altlwr_corr() << "..." << aspc.get_altupr_corr() << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		Rect bbox(aspc.get_bbox());
		const MultiPolygonHole& poly(aspc.get_polygon());
		double area(fabs(poly.get_simple_area_nmi2_dbl()));
		if (area > maxarea) {
			if (false) {
				std::ostringstream oss;
				oss << "VFR: exclude_airspace: skipping " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
				    << " due to area: " << area << "nmi^2" << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			continue;
		}
		unsigned int killcnt(0);
		Graph::edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::edges(m_graph); e0 != e1; ++e0) {
			Edge& ee(m_graph[*e0]);
			if (ee.get_metric() == std::numeric_limits<double>::max())
				continue;
			Graph::vertex_descriptor u(boost::source(*e0, m_graph));
			if (u == m_dep)
				continue;
			Graph::vertex_descriptor v(boost::target(*e0, m_graph));
			if (v == m_dest)
				continue;
			const Vertex& uu(m_graph[u]);
			const Vertex& vv(m_graph[v]);
			bool uui(bbox.is_inside(uu.get_coord()));
			bool vvi(bbox.is_inside(vv.get_coord()));
			{
				Rect r(uu.get_coord(), uu.get_coord());
				r.add(vv.get_coord());
				if (!r.is_intersect(bbox))
					continue;
			}
			if ((uui && poly.windingnumber(uu.get_coord())) ||
			    (vvi && poly.windingnumber(vv.get_coord())) ||
			    poly.is_intersection(uu.get_coord(), vv.get_coord())) {
				//ee.set_metric(std::numeric_limits<double>::max());
				ee.set_metric(ee.get_dist() * 1000);
				++killcnt;
				if (false) {
					std::ostringstream oss;
					oss << "VFR: airspace: kill edge " << uu.get_ident() << "->" << vv.get_ident() << std::endl;
					m_ar.m_signal_log(log_debug0, oss.str());
				}
			}
		}
		if (false || killcnt) {
			std::ostringstream oss;
			oss << "VFR: exclude_airspace: " << aspc.get_icao_name() << ' ' << aspc.get_class_string()
			    << " altitude: " << aspc.get_altlwr_corr() << "..." << aspc.get_altupr_corr()
			    << " area: " << area << "nmi^2: removed " << killcnt << " edges";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
	}
}

void CFMUAutoroute::VFRRouter::exclude_regions(void)
{
	for (excluderegions_t::const_iterator ei(m_ar.m_excluderegions.begin()), ee(m_ar.m_excluderegions.end()); ei != ee; ++ei) {
		AirspacesDb::Airspace aspc;
		Rect bbox(ei->get_bbox());
		if (ei->is_airspace()) {
			AirspacesDb::elementvector_t ev(m_ar.m_airspacedb.find_by_icao(ei->get_airspace_id(), 0, AirspacesDb::comp_exact, 0, AirspacesDb::Airspace::subtables_all));
			int eadidx(-1);
			for (unsigned int evi = 0; evi < ev.size(); ) {
				if (!ev[evi].is_valid() || ev[evi].get_class_string() != ei->get_airspace_type()) {
					ev.erase(ev.begin() + evi);
					continue;
				}
				if (ev[evi].get_typecode() == AirspacesDb::Airspace::typecode_ead && eadidx == -1)
					eadidx = evi;
				++evi;
			}
			if (eadidx == -1)
				eadidx = 0;
			if (ev.empty()) {
				std::ostringstream oss;
				oss << "Exclude Regions: Airspace " << ei->get_airspace_id() << '/' << ei->get_airspace_type()
				    << " not found";
				m_ar.m_signal_log(log_debug0, oss.str());
				continue;
			}
			aspc = ev[eadidx];
			bbox = aspc.get_bbox();
			if (ev.size() > 1) {
				std::ostringstream oss;
				oss << "Exclude Regions: Multiple Airspaces " << ei->get_airspace_id() << '/' << ei->get_airspace_type()
				    << " found";
				m_ar.m_signal_log(log_debug0, oss.str());
			}
		}
		unsigned int killcnt(0);
		Graph::edge_iterator e0, e1;
		for (boost::tie(e0, e1) = boost::edges(m_graph); e0 != e1; ++e0) {
			Edge& ee(m_graph[*e0]);
			if (ee.get_metric() == std::numeric_limits<double>::max())
				continue;
			Graph::vertex_descriptor u(boost::source(*e0, m_graph));
			if (u == m_dep)
				continue;
			Graph::vertex_descriptor v(boost::target(*e0, m_graph));
			if (v == m_dest)
				continue;
			const Vertex& uu(m_graph[u]);
			const Vertex& vv(m_graph[v]);
			bool uui(bbox.is_inside(uu.get_coord()));
			bool vvi(bbox.is_inside(vv.get_coord()));
			if (aspc.is_valid()) {
				{
					Rect r(uu.get_coord(), uu.get_coord());
					r.add(vv.get_coord());
					if (!r.is_intersect(bbox))
						continue;
				}
				if ((!uui || !aspc.get_polygon().windingnumber(uu.get_coord())) &&
				    (!vvi || !aspc.get_polygon().windingnumber(vv.get_coord())) &&
				    !aspc.get_polygon().is_intersection(uu.get_coord(), vv.get_coord()))
					continue;
			} else {
				if (!uui && !vvi && !bbox.is_intersect(uu.get_coord(), vv.get_coord()))
					continue;
			}
			//ee.set_metric(std::numeric_limits<double>::max());
			ee.set_metric(ee.get_dist() * 1000);
			++killcnt;
			if (false) {
				std::ostringstream oss;
				oss << "VFR: airspace: kill edge " << uu.get_ident() << "->" << vv.get_ident() << std::endl;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
		}
		if (false || killcnt) {
			std::ostringstream oss;
			oss << "VFR: exclude_regions: ";
			if (aspc.is_valid())
				oss << aspc.get_icao_name() << ' ' << aspc.get_class_string();
			else
				oss << bbox;
			oss << " removed " << killcnt << " edges";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
	}
}

bool CFMUAutoroute::VFRRouter::autoroute(FPlanRoute& route)
{
	while (route.get_nrwpt())
		route.delete_wpt(0);
	if (m_dep == m_dest) {
		if (true) {
			std::ostringstream oss;
			oss << "VFR: departure and destination are the same";
			m_ar.m_signal_log(log_debug0, oss.str());
		}
		return false;
	}
	std::vector<Graph::vertex_descriptor> predecessors;
	m_graph.shortest_paths_metric(m_dep, predecessors);
	{
		const Vertex& vv(m_graph[m_dest]);
		if (!vv.insert(route, 0)) {
			if (true) {
				std::ostringstream oss;
				oss << "VFR: cannot insert " << vv.get_ident() << " into fpl" ;
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
	}
	Graph::vertex_descriptor v(m_dest);
	while (v != m_dep) {
		Graph::vertex_descriptor u(predecessors[v]);
		if (u == v) {
			if (true) {
				std::ostringstream oss;
				oss << "VFR: no route found; "
				    << boost::out_degree(m_dep, m_graph) << " SID edges, "
				    << boost::in_degree(m_dest, m_graph) << " STAR edges";
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
		const Vertex& uu(m_graph[u]);
		const Vertex& vv(m_graph[v]);
		bool rteadded(false);
		if (v == m_dest && uu.is_match(Vertex::typemask_vfrreportingpt)) {
			for (sidstar_t::const_iterator si(m_star.begin()), se(m_star.end()); si != se; ++si) {
				const AirportsDb::Airport::VFRRoute& rte(*si);
				if (!rte.size() || rte.get_name() != uu.get_ident())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[0]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival)
					continue;
				AirportsDb::Airport arpt;
				m_graph[m_dest].get(arpt);
				for (unsigned int i = 0, j = rte.size() - 1U; i < j; ++i) {
					const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[i]);
					FPlanWaypoint wpt;
					if (rtept.is_at_airport() && arpt.is_valid()) {
						Engine::DbObject::set(wpt, arpt);
					} else {
						if (arpt.is_valid())
							wpt.set_icao(arpt.get_icao());
						wpt.set_name(rtept.get_name());
						wpt.set_coord(rtept.get_coord());
						wpt.set_altitude(rtept.get_altitude());
					}
					wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
					wpt.set_pathname(rte.get_name());
					route.insert_wpt(i, wpt);
				}
				rteadded = true;
				break;
			}
		}
		if (u == m_dep && vv.is_match(Vertex::typemask_vfrreportingpt)) {
			for (sidstar_t::const_iterator si(m_sid.begin()), se(m_sid.end()); si != se; ++si) {
				const AirportsDb::Airport::VFRRoute& rte(*si);
				if (!rte.size() || rte.get_name() != vv.get_ident())
					continue;
				const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[rte.size()-1]);
				if (rtept.get_pathcode() != AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure)
					continue;
				AirportsDb::Airport arpt;
				m_graph[m_dep].get(arpt);
				for (unsigned int i = 0, j = rte.size() - 1U; i < j; ++i) {
					const AirportsDb::Airport::VFRRoute::VFRRoutePoint& rtept(rte[i]);
					FPlanWaypoint wpt;
					if (rtept.is_at_airport() && arpt.is_valid()) {
						Engine::DbObject::set(wpt, arpt);
					} else {
						if (arpt.is_valid())
							wpt.set_icao(arpt.get_icao());
						wpt.set_name(rtept.get_name());
						wpt.set_coord(rtept.get_coord());
						wpt.set_altitude(rtept.get_altitude());
					}
					wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
					wpt.set_pathname(rte.get_name());
					route.insert_wpt(i, wpt);
				}
				rteadded = true;
				break;
			}
		}
		if (!rteadded && uu.is_match(Vertex::typemask_vfrreportingpt)) {
			// FIXME!!
			std::ostringstream oss;
			oss << "VFR: do not know what to do with VFR arr/dep point " << uu.get_ident();
			m_ar.m_signal_log(log_debug0, oss.str());
			rteadded = true;
		}
		if (!rteadded && !uu.insert(route, 0)) {
			if (true) {
				std::ostringstream oss;
				oss << "VFR: cannot insert " << uu.get_ident() << " into fpl";
				m_ar.m_signal_log(log_debug0, oss.str());
			}
			return false;
		}
		v = u;
	}
	return true;
}
