#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>

#include "cfmuautoroute.hh"

const char CFMUAutoroute::GraphSerializeBase::signature[] = "vfrnav CFMU Routing Graph V1\n";

CFMUAutoroute::GraphSerializeBase::GraphSerializeBase(void)
	: m_minlevel(std::numeric_limits<int16_t>::min()),
	  m_maxlevel(std::numeric_limits<int16_t>::min())
{
	Point pt;
	pt.set_invalid();
	m_bbox = Rect(pt, pt);
}

bool CFMUAutoroute::GraphSerializeBase::is_valid(void) const
{
	if (m_bbox.get_southwest().is_invalid() ||
	    m_bbox.get_northeast().is_invalid() ||
	    m_minlevel > m_maxlevel || m_minlevel < 0 ||
	    m_minlevel % 10 || m_maxlevel % 10)
		return false;
	return true;
}

CFMUAutoroute::GraphWrite::GraphWrite(std::ostream& os)
	: m_os(os)
{
}

template <> void CFMUAutoroute::GraphWrite::io<int8_t>(const int8_t& v)
{
        m_os.write(reinterpret_cast<const char *>(&v), sizeof(v));
        if (!m_os)
                throw std::runtime_error("io<int8_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<uint8_t>(const uint8_t& v)
{
        m_os.write(reinterpret_cast<const char *>(&v), sizeof(v));
        if (!m_os)
                throw std::runtime_error("io<uint8_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<int16_t>(const int16_t& v)
{
        uint8_t buf[2];
        buf[0] = v;
        buf[1] = v >> 8;
        m_os.write(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_os)
                throw std::runtime_error("io<int16_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<uint16_t>(const uint16_t& v)
{
        uint8_t buf[2];
        buf[0] = v;
        buf[1] = v >> 8;
        m_os.write(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_os)
                throw std::runtime_error("io<uint16_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<int32_t>(const int32_t& v)
{
        uint8_t buf[4];
        buf[0] = v;
        buf[1] = v >> 8;
        buf[2] = v >> 16;
        buf[3] = v >> 24;
        m_os.write(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_os)
                throw std::runtime_error("io<int32_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<uint32_t>(const uint32_t& v)
{
        uint8_t buf[4];
        buf[0] = v;
        buf[1] = v >> 8;
        buf[2] = v >> 16;
        buf[3] = v >> 24;
        m_os.write(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_os)
                throw std::runtime_error("io<uint32_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<int64_t>(const int64_t& v)
{
        uint8_t buf[8];
        buf[0] = v;
        buf[1] = v >> 8;
        buf[2] = v >> 16;
        buf[3] = v >> 24;
        buf[4] = v >> 32;
        buf[5] = v >> 40;
        buf[6] = v >> 48;
        buf[7] = v >> 56;
	m_os.write(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_os)
                throw std::runtime_error("io<int64_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<uint64_t>(const uint64_t& v)
{
        uint8_t buf[8];
        buf[0] = v;
        buf[1] = v >> 8;
        buf[2] = v >> 16;
        buf[3] = v >> 24;
        buf[4] = v >> 32;
        buf[5] = v >> 40;
        buf[6] = v >> 48;
        buf[7] = v >> 56;
	m_os.write(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_os)
                throw std::runtime_error("io<uint64_t>: write error");	
}

template <> void CFMUAutoroute::GraphWrite::io<float>(const float& v)
{
        union {
                float f;
                uint32_t u;
        } x;
        x.f = v;
	io(x.u);
}

template <> void CFMUAutoroute::GraphWrite::io<double>(const double& v)
{
	union {
                double d;
                uint64_t u;
        } x;
        x.d = v;
	io(x.u);
}

template <> void CFMUAutoroute::GraphWrite::io<std::string>(const std::string& v)
{
        iouint32(v.size());
        m_os.write(v.c_str(), v.size());
        if (!m_os)
                throw std::runtime_error("io<std::string>: write error");
}

template <> void CFMUAutoroute::GraphWrite::io<Glib::ustring>(const Glib::ustring& v)
{
        iouint32(v.size());
        m_os.write(v.c_str(), v.size());
        if (!m_os)
                throw std::runtime_error("io<Glib::ustring>: write error");
}

template <> void CFMUAutoroute::GraphWrite::io<char>(const char& v)
{
	ioint8(v);
}

template <> void CFMUAutoroute::GraphWrite::io<bool>(const bool& v)
{
	uint8_t vv(!!v);
	io(vv);
}

void CFMUAutoroute::GraphWrite::header(void)
{
	m_os.write(signature, sizeof(signature));
	if (!m_os)
		throw std::runtime_error("cannot write signature");
	if (!is_valid())
		throw std::runtime_error("write header: invalid parameters");
	m_bbox.hibernate_binary(*this);
	io(m_minlevel);
	io(m_maxlevel);
}

CFMUAutoroute::GraphRead::GraphRead(std::istream& is)
	: m_is(is)
{
}

template <> void CFMUAutoroute::GraphRead::io<int8_t>(int8_t& v)
{
        m_is.read(reinterpret_cast<char *>(&v), sizeof(v));
        if (!m_is)
                throw std::runtime_error("io<int8_t>: short read");
}

template <> void CFMUAutoroute::GraphRead::io<uint8_t>(uint8_t& v)
{
        m_is.read(reinterpret_cast<char *>(&v), sizeof(v));
        if (!m_is)
                throw std::runtime_error("io<uint8_t>: short read");
}

template <> void CFMUAutoroute::GraphRead::io<int16_t>(int16_t& v)
{
        uint8_t buf[2];
        m_is.read(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_is)
                throw std::runtime_error("io<int16_t>: short read");
        v = buf[0] | (buf[1] << 8);
}

template <> void CFMUAutoroute::GraphRead::io<uint16_t>(uint16_t& v)
{
        uint8_t buf[2];
        m_is.read(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_is)
                throw std::runtime_error("io<uint16_t>: short read");
        v = buf[0] | (buf[1] << 8);
}

template <> void CFMUAutoroute::GraphRead::io<int32_t>(int32_t& v)
{
        uint8_t buf[4];
        m_is.read(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_is)
                throw std::runtime_error("io<int32_t>: short read");
        v = buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24);
}

template <> void CFMUAutoroute::GraphRead::io<uint32_t>(uint32_t& v)
{
        uint8_t buf[4];
        m_is.read(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_is)
                throw std::runtime_error("io<uint32_t>: short read");
        v = buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24);
}

template <> void CFMUAutoroute::GraphRead::io<int64_t>(int64_t& v)
{
        uint8_t buf[8];
        m_is.read(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_is)
                throw std::runtime_error("io<int64_t>: short read");
        uint64_t r(buf[4] | (buf[5] << 8) | (buf[6] << 16) | (buf[7] << 24));
        r <<= 32;
        r |= buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24);
        v = r;
}

template <> void CFMUAutoroute::GraphRead::io<uint64_t>(uint64_t& v)
{
        uint8_t buf[8];
        m_is.read(reinterpret_cast<char *>(buf), sizeof(buf));
        if (!m_is)
                throw std::runtime_error("io<uint64_t>: short read");
        uint64_t r(buf[4] | (buf[5] << 8) | (buf[6] << 16) | (buf[7] << 24));
        r <<= 32;
        r |= buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24);
        v = r;
}

template <> void CFMUAutoroute::GraphRead::io<float>(float& v)
{
        union {
                float f;
                uint32_t u;
        } x;
	io(x.u);
        v = x.f;
}

template <> void CFMUAutoroute::GraphRead::io<double>(double& v)
{
	union {
                double d;
                uint64_t u;
        } x;
	io(x.u);
        v = x.d;
}

template <> void CFMUAutoroute::GraphRead::io<std::string>(std::string& v)
{
	v.clear();
        uint32_t len(0);
	io(len);
        if (!len)
                return;
        char buf[len];
        m_is.read(buf, len);
        if (!m_is)
                throw std::runtime_error("io<std::string>: short read");
        v = std::string(buf, buf + len);
}

template <> void CFMUAutoroute::GraphRead::io<Glib::ustring>(Glib::ustring& v)
{
	v.clear();
        uint32_t len(0);
	io(len);
        if (!len)
                return;
        char buf[len];
        m_is.read(buf, len);
        if (!m_is)
                throw std::runtime_error("io<Glib::ustring>: short read");
        v = std::string(buf, buf + len);
}

template <> void CFMUAutoroute::GraphRead::io<char>(char& v)
{
	ioint8(v);
}

template <> void CFMUAutoroute::GraphRead::io<bool>(bool& v)
{
	uint8_t vv(0);
	io(vv);
	v = !!vv;
}

void CFMUAutoroute::GraphRead::header(void)
{
	{
		char buf[sizeof(signature)];
		m_is.read(buf, sizeof(signature));
		if (!m_is)
			throw std::runtime_error("cannot read signature");
		if (memcmp(buf, signature, sizeof(signature)))
			throw std::runtime_error("invalid signature");
	}
	m_bbox.hibernate_binary(*this);
	io(m_minlevel);
	io(m_maxlevel);
	if (!is_valid())
		throw std::runtime_error("read header: invalid parameters");
}

void CFMUAutoroute::lgraphsave(std::ostream& os, const Rect& bbox)
{
	unsigned int nrperf(m_performance.size());
	if (!nrperf)
		throw std::runtime_error("no performance table");
	if (!m_lsupervertices.size())
		throw std::runtime_error("no super vertices");
	if (nrperf * m_lsupervertices.size() != boost::num_vertices(m_lgraph)) {
		std::ostringstream oss;
		oss << "invalid graph structure: levels " << nrperf << " supervertices " << m_lsupervertices.size()
		    << " vertices " << boost::num_vertices(m_lgraph);
		throw std::runtime_error(oss.str());
	}
	for (LSuperVertices::const_iterator i(m_lsupervertices.begin()), e(m_lsupervertices.end()); i != e; ++i) {
		if (i->get_v1() - i->get_v0() == nrperf - 1)
			continue;
		std::ostringstream oss;
		oss << "invalid graph structure: levels " << nrperf << " vertex " << i->get_ident()
		    << " v0 " << i->get_v0() << " v1 " << i->get_v1();
		throw std::runtime_error(oss.str());
	}
	typedef std::vector<unsigned int> superremap_t;
	superremap_t superremap(m_lsupervertices.size() + 1, 0);
	for (LGraph::vertex_descriptor v(0); v < boost::num_vertices(m_lgraph); ++v) {
		if (!boost::out_degree(v, m_lgraph) && !boost::in_degree(v, m_lgraph))
			continue;
		superremap[v / nrperf] = 1;
	}
	for (unsigned int i = 0; i < m_lsupervertices.size(); ++i) {
		const LSuperVertex& super(m_lsupervertices[i]);
		if (!bbox.is_inside(super.get_coord()))
			superremap[i] = 0;
	}
	{
		unsigned int idx(0);
		for (superremap_t::iterator i(superremap.begin()), e(superremap.end()); i != e; ++i) {
			unsigned int x(idx);
			std::swap(x, *i);
			idx += x;
		}
	}
       	GraphWrite gw(os);
	gw.set_bbox(bbox);
	gw.set_minlevel(m_performance[0].get_altitude() / 100);
	gw.set_maxlevel(m_performance.back().get_altitude() / 100);
	if (gw.get_maxlevel() - gw.get_minlevel() != (nrperf - 1U) * 10U) {
		std::ostringstream oss;
		oss << "invalid performance table: levels " << nrperf
		    << " F" << std::setfill('0') << std::setw(3) << gw.get_minlevel()
		    << "..F" << std::setfill('0') << std::setw(3) << gw.get_maxlevel();
		throw std::runtime_error(oss.str());
	}
	if (true) {
		Rect bbox(gw.get_bbox());
		bool subseq(false);
		Rect bbox1;
		for (unsigned int i = 0; i < m_lsupervertices.size(); ++i) {
			if (superremap[i] == superremap[i + 1])
				continue;
			const LSuperVertex& super(m_lsupervertices[i]);
			if (subseq) {
				bbox1 = bbox1.add(super.get_coord());
				continue;
			}
			bbox1 = Rect(super.get_coord(), super.get_coord());
			subseq = true;
		}
		if (!subseq)
			throw std::runtime_error("Graph does not contain a non-skipped vertex");
		std::ostringstream oss;
		oss << "Graph bbox ["
		    << bbox.get_southwest().get_lat_str2() << ' ' << bbox.get_southwest().get_lon_str2()
		    << ' ' << bbox.get_northeast().get_lat_str2() << ' ' << bbox.get_northeast().get_lon_str2()
		    << "] actual bbox ["
		    << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
		    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2()
		    << "] F" << std::setfill('0') << std::setw(3) << gw.get_minlevel()
		    << "..F" << std::setfill('0') << std::setw(3) << gw.get_maxlevel();
		m_signal_log(log_debug0, oss.str());
	}
	gw.header();
	// node table
	gw.iouint32(superremap.back());
	for (unsigned int i = 0; i < m_lsupervertices.size(); ++i) {
		if (superremap[i] == superremap[i + 1])
			continue;
		const LSuperVertex& super(m_lsupervertices[i]);
		Engine::DbObject::ptr_t p(super.get_object());
		Engine::DbObject::hibernate_binary(p, gw);
	}
	// airway table
	gw.iouint32(m_lgraphairwaytable.size());
	for (lgraphairwaytable_t::const_iterator i(m_lgraphairwaytable.begin()), e(m_lgraphairwaytable.end()); i != e; ++i)
		gw.io(*i);
	// edge list
	{
	        LGraph::edge_iterator e0, e1;
		uint32_t nredges(0), nrlvledges(0), nrdiscedges(0);
		for (boost::tie(e0, e1) = boost::edges(m_lgraph); e0 != e1; ++e0) {
			const LEdge& ee(m_lgraph[*e0]);
			if (ee.is_match(airwaymatchlevel)) {
				++nrlvledges;
				continue;
			}
			if (!ee.is_match(airwaymatchdctawy))
				continue;
			uint32_t v0(boost::source(*e0, m_lgraph));
			uint32_t v1(boost::target(*e0, m_lgraph));
			uint32_t sv0(v0 / nrperf);
			if (superremap[sv0] == superremap[sv0 + 1]) {
				++nrdiscedges;
				continue;
			}
			uint32_t sv1(v1 / nrperf);
			if (superremap[sv1] == superremap[sv1 + 1]) {
				++nrdiscedges;
				continue;
			}
			++nredges;
		}
		if (true) {
			std::ostringstream oss;
			oss << "#edges " << boost::num_edges(m_lgraph) << " perf " << nrperf
			    << " supervertices " << m_lsupervertices.size() << " awy/dct edges " << nredges
			    << " lvl edges " << nrlvledges;
			m_signal_log(log_debug0, oss.str());
		}
		gw.iouint32(nredges);
		for (boost::tie(e0, e1) = boost::edges(m_lgraph); e0 != e1; ++e0) {
			const LEdge& ee(m_lgraph[*e0]);
			if (!ee.is_match(airwaymatchdctawy))
				continue;
			uint32_t v0(boost::source(*e0, m_lgraph));
			uint32_t v1(boost::target(*e0, m_lgraph));
			uint32_t sv0(v0 / nrperf);
			if (superremap[sv0] == superremap[sv0 + 1])
				continue;
			uint32_t sv1(v1 / nrperf);
			if (superremap[sv1] == superremap[sv1 + 1])
				continue;
			v0 -= (sv0 - superremap[sv0]) * nrperf;
			v1 -= (sv1 - superremap[sv1]) * nrperf;
			gw.io(v0);
			gw.io(v1);
			gw.iouint32(ee.get_ident());
			gw.io(ee.get_metric());
		}
		if (nredges + nrlvledges + nrdiscedges != boost::num_edges(m_lgraph)) {
			std::ostringstream oss;
			oss << "boost #edges / edge iterator inconsistency: " << nredges << " + " << nrlvledges << " + " << nrdiscedges
			    << " != " << boost::num_edges(m_lgraph);
			throw std::runtime_error(oss.str());
		}
	}
}

void CFMUAutoroute::lgraphbinload(GraphRead& gr, const Rect& bbox)
{
	// header must already have been consumed
	clearlgraph();
	unsigned int nrperf(m_performance.size());
	if (!nrperf)
		throw std::runtime_error("no performance table");
	if (!gr.is_valid())
		throw std::runtime_error("header not valid");
	if (!gr.get_bbox().is_inside(bbox))
		throw std::runtime_error("area not contained in graph");
	if (m_performance[0].get_altitude() < gr.get_minlevel() * 100U ||
	    m_performance.back().get_altitude() > gr.get_maxlevel() * 100U)
		throw std::runtime_error("levels not contained in graph");
	unsigned int nrperffile((gr.get_maxlevel() - gr.get_minlevel()) / 10U + 1U);
	unsigned int poffs((m_performance[0].get_altitude() - gr.get_minlevel() * 100U) / 1000U);
	typedef std::vector<unsigned int> superremap_t;
	superremap_t superremap;
	// node table
	{
		uint32_t nrnodes(0), nodeidx(0);
		gr.iouint32(nrnodes);
		for (unsigned int i = 0; i < nrnodes; ++i) {
			Engine::DbObject::ptr_t p;
			Engine::DbObject::hibernate_binary(p, gr);
			superremap.push_back(nodeidx);
			if (!p || !bbox.is_inside(p->get_coord()))
				continue;
			++nodeidx;
			lgraphaddvertex(p, true, true);
		}
		superremap.push_back(nodeidx);
	}
	// airway table
	{
		uint32_t nr(0);
		gr.iouint32(nr);
		for (unsigned int i = 0; i < nr; ++i) {
			Glib::ustring awyu;
			gr.io(awyu);
			std::pair<lgraphairwayrevtable_t::iterator,bool> ins(m_lgraphairwayrevtable.insert(std::make_pair(awyu, m_lgraphairwaytable.size())));
			if (ins.second)
				m_lgraphairwaytable.push_back(awyu);
			else
				throw std::runtime_error("duplicate airway " + awyu);
		}
	}
	// edge list
	{
		uint32_t nr(0);
		gr.iouint32(nr);
		for (unsigned int i = 0; i < nr; ++i) {
			uint32_t v0(0), v1(0), ident(0);
			double metric(0);
			gr.io(v0);
			gr.io(v1);
			gr.io(ident);
			gr.io(metric);
			// compute correct graph node indices
			unsigned int v0s(v0 / nrperffile);
			if (v0s + 1 >= superremap.size() || superremap[v0s] == superremap[v0s + 1])
				continue;
			unsigned int v0p(v0 - v0s * nrperffile);
			v0s = superremap[v0s];
			if (v0s >= m_lsupervertices.size())
				continue;
			const LSuperVertex& super0(m_lsupervertices[v0s]);
			unsigned int v1s(v1 / nrperffile);
			if (v1s + 1 >= superremap.size() || superremap[v1s] == superremap[v1s + 1])
				continue;
			unsigned int v1p(v1 - v1s * nrperffile);
			v1s = superremap[v1s];
			if (v1s >= m_lsupervertices.size())
				continue;
			const LSuperVertex& super1(m_lsupervertices[v1s]);
			if (v0p < poffs)
				continue;
			v0p -= poffs;
			if (v0p >= nrperf)
				continue;
			if (v1p < poffs)
				continue;
			v1p -= poffs;
			if (v1p >= nrperf)
				continue;
			if (v0p != v1p)
				throw std::runtime_error("airway/dct edge crosses level");
			if (ident == airwaydct && metric > get_dctlimit()) {
				if (false) {
					std::ostringstream oss;
					oss << "skipping DCT edge " << lgraphvertexname(super0.get_v0() + v0p)
					    << "->" << lgraphvertexname(super1.get_v0() + v1p) << std::fixed
					    << " dist " << std::setprecision(1) << metric
					    << " limit " << std::setprecision(1) << get_dctlimit();
					m_signal_log(log_normal, oss.str());
				}
				continue;
			}
			const Performance& perf(m_performance[v0p]);
			boost::add_edge(super0.get_v0() + v0p, super1.get_v0() + v1p,
					LEdge(ident, lgraphmetric(super0.get_coord(), super1.get_coord(), perf, metric)), m_lgraph);
		}
	}
}

class CFMUAutoroute::GraphFile {
public:
	typedef Glib::RefPtr<GraphFile> ptr_t;

	GraphFile(const std::string& fn);
	void reference(void) const;
	void unreference(void) const;

	GraphRead& get_gr(void) { return m_gr; }
	const GraphRead& get_gr(void) const { return m_gr; }
	std::ifstream& get_is(void) { return m_is; }
	const std::ifstream& get_is(void) const { return m_is; }
	const std::string& get_filename(void) const { return m_filename; }

	struct sort_ptr_area {
	public:
		sort_ptr_area(void) {}
		bool operator()(const ptr_t& p0, const ptr_t& p1) const;
	};

protected:
	std::string m_filename;
	std::ifstream m_is;
	GraphRead m_gr;
	mutable gint m_refcount;
};

CFMUAutoroute::GraphFile::GraphFile(const std::string& fn)
	: m_filename(fn), m_is(m_filename.c_str(), std::ifstream::binary), m_gr(m_is), m_refcount(1)
{
}

void CFMUAutoroute::GraphFile::reference(void) const
{
        g_atomic_int_inc(&m_refcount);
}

void CFMUAutoroute::GraphFile::unreference(void) const
{
        if (!g_atomic_int_dec_and_test(&m_refcount))
                return;
        delete this;
}

bool CFMUAutoroute::GraphFile::sort_ptr_area::operator()(const ptr_t& p0, const ptr_t& p1) const
{
	if (!p1)
		return !!p0;
	if (!p0)
		return false;
	if (!p1->get_gr().is_valid())
		return p0->get_gr().is_valid();
	if (!p0->get_gr().is_valid())
		return false;
	return p0->get_gr().get_bbox().get_simple_area_km2_dbl() < p1->get_gr().get_bbox().get_simple_area_km2_dbl();
}

bool CFMUAutoroute::lgraphbinload(const Rect& bbox)
{
	typedef std::vector<GraphFile::ptr_t> files_t;
	files_t files;
	{
		Glib::Dir dir(get_db_auxdir());
		for (;;) {
			std::string name(dir.read_name());
			if (name.empty())
				break;
			if (name.size() < 13)
				continue;
			if (name.substr(name.size()-4) != ".bin" ||
			    name.substr(0, 9) != "cfmugraph")
				continue;
			std::string fname(Glib::build_filename(get_db_auxdir(), name));
			if (!Glib::file_test(fname, Glib::FILE_TEST_EXISTS) || Glib::file_test(fname, Glib::FILE_TEST_IS_DIR))
				continue;
			GraphFile::ptr_t fp(new GraphFile(fname));
			if (!fp->get_is()) {
				std::ostringstream oss;
				oss << "cannot open file " << fp->get_filename();
				m_signal_log(log_normal, oss.str());
				continue;
			}
			try {
				fp->get_gr().header();
			} catch (const std::exception& e) {
				std::ostringstream oss;
				oss << "cannot read file header from " << fp->get_filename()
				    << ": " << e.what();
				m_signal_log(log_normal, oss.str());
				continue;
			}
			if (!fp->get_gr().is_valid()) {
				std::ostringstream oss;
				oss << "invalid file header from " << fp->get_filename();
				m_signal_log(log_normal, oss.str());
				continue;
			}
			if (!fp->get_gr().get_bbox().is_inside(bbox)) {
				std::ostringstream oss;
				const Rect& bbox1(fp->get_gr().get_bbox());
				oss << "file " << fp->get_filename() << " bbox ["
				    << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
				    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2()
				    << "] does not cover [" << bbox.get_southwest().get_lat_str2() << ' ' << bbox.get_southwest().get_lon_str2()
				    << ' ' << bbox.get_northeast().get_lat_str2() << ' ' << bbox.get_northeast().get_lon_str2()
				    << ']';
				m_signal_log(log_normal, oss.str());
				continue;
			}
			if (m_performance[0].get_altitude() < fp->get_gr().get_minlevel() * 100U ||
			    m_performance.back().get_altitude() > fp->get_gr().get_maxlevel() * 100U) {
				std::ostringstream oss;
				oss << "file " << fp->get_filename()
				    << " F" << std::setfill('0') << std::setw(3) << fp->get_gr().get_minlevel()
				    << "..F" << std::setfill('0') << std::setw(3) << fp->get_gr().get_maxlevel()
				    << " does not cover F" << std::setfill('0') << std::setw(3) << (m_performance[0].get_altitude() / 100U)
				    << "..F" << std::setfill('0') << std::setw(3) << (m_performance.back().get_altitude() / 100U);
				m_signal_log(log_normal, oss.str());
				continue;
			}
			files.push_back(fp);
		}
	}
	std::sort(files.begin(), files.end(), GraphFile::sort_ptr_area());
	for (files_t::const_iterator i(files.begin()), e(files.end()); i != e; ++i) {
		if (!*i)
			continue;
		std::ostringstream oss;
		const Rect& bbox1((*i)->get_gr().get_bbox());
		oss << "CFMU graph file " << (*i)->get_filename() << " bbox ["
		    << bbox1.get_southwest().get_lat_str2() << ' ' << bbox1.get_southwest().get_lon_str2()
		    << ' ' << bbox1.get_northeast().get_lat_str2() << ' ' << bbox1.get_northeast().get_lon_str2()
		    << "] F" << std::setfill('0') << std::setw(3) << (*i)->get_gr().get_minlevel()
		    << "..F" << std::setfill('0') << std::setw(3) << (*i)->get_gr().get_maxlevel()
		    << " area " << (*i)->get_gr().get_bbox().get_simple_area_km2_dbl()
		    << " area2 " << ((PolygonSimple)((*i)->get_gr().get_bbox())).get_simple_area_km2_dbl();
		if (!(*i)->get_gr().is_valid())
			oss << " (INVALID)";
		m_signal_log(log_normal, oss.str());
	}
	for (files_t::iterator i(files.begin()), e(files.end()); i != e; ++i) {
		if (!*i || !(*i)->get_gr().is_valid())
			break;
		try {
			lgraphbinload((*i)->get_gr(), bbox);
		} catch (const std::exception& e) {
			std::ostringstream oss;
			oss << "cannot load file " << (*i)->get_filename() << ": " << e.what();
			m_signal_log(log_normal, oss.str());
			continue;
		}
		std::ostringstream oss;
		oss << "file " << (*i)->get_filename() << " successfully loaded ";
		m_signal_log(log_normal, oss.str());
		return true;
	}
	return false;
}
