#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "interval.hh"

#include <sstream>
//#include <cstdint>
#include <stdint.h>

template <typename T>
IntervalSet<T>::Interval::Interval(void)
	: m_lower(std::numeric_limits<type_t>::min()), m_upper(std::numeric_limits<type_t>::max())
{
}

template <typename T>
IntervalSet<T>::Interval::Interval(type_t val)
	: m_lower(val), m_upper(val + 1)
{
}

template <typename T>
IntervalSet<T>::Interval::Interval(type_t val0, type_t val1)
	: m_lower(val0), m_upper(val1)
{
}

template <typename T>
bool IntervalSet<T>::Interval::operator<(const Interval& x) const
{
	if (m_lower < x.m_lower)
		return true;
	if (m_lower > x.m_lower)
		return false;
	return m_upper < x.m_upper;
}

template <typename T>
bool IntervalSet<T>::Interval::is_empty(void) const
{
	return m_lower >= m_upper;
}

template <typename T>
bool IntervalSet<T>::Interval::is_full(void) const
{
	return m_lower == std::numeric_limits<type_t>::min() &&
		m_upper == std::numeric_limits<type_t>::max();
}
template <typename T>
bool IntervalSet<T>::Interval::is_inside(type_t x) const
{
	return x >= m_lower && x < m_upper;
}

template <typename T>
bool IntervalSet<T>::Interval::is_overlap(const Interval& x) const
{
	if (is_empty() || x.is_empty())
		return false;
	return x.get_upper() > get_lower() && x.get_lower() < get_upper();
}

template <typename T>
bool IntervalSet<T>::Interval::is_adjacent(const Interval& x) const
{
	if (is_empty() || x.is_empty())
		return false;
	return x.get_upper() == get_lower() || get_upper() == x.get_lower();
}

template <typename T>
bool IntervalSet<T>::Interval::is_overlap_or_adjacent(const Interval& x) const
{
	if (is_empty() || x.is_empty())
		return false;
	return x.get_upper() >= get_lower() && x.get_lower() <= get_upper();
}

template <typename T>
void IntervalSet<T>::Interval::set_empty(void)
{
	m_lower = m_upper = std::numeric_limits<type_t>::min();
}

template <typename T>
void IntervalSet<T>::Interval::set_full(void)
{
	m_lower = std::numeric_limits<type_t>::min();
	m_upper = std::numeric_limits<type_t>::max();
}

template <typename T>
std::string IntervalSet<T>::Interval::to_str(void) const
{
	std::ostringstream oss;
	oss << '[' << m_lower << ',' << m_upper << ')';
	return oss.str();
}

template <typename T>
IntervalSet<T>::IntervalSet(void)
{
}

template <typename T>
IntervalSet<T>::IntervalSet(const Interval& x)
{
	m_set.insert(x);
	normalize();
}

template <typename T>
IntervalSet<T>::IntervalSet(const IntervalSet& x)
{
	m_set = x.m_set;
}

template <typename T>
bool IntervalSet<T>::is_empty(void) const
{
	return m_set.empty();
}

template <typename T>
bool IntervalSet<T>::is_full(void) const
{
	if (m_set.empty())
		return false;
	return m_set.begin()->is_full();
}

template <typename T>
bool IntervalSet<T>::is_inside(type_t x) const
{
	for (typename set_t::const_iterator i(m_set.begin()), e(m_set.end()); i != e; ++i)
		if (i->is_inside(x))
			return true;
	return false;
}

template <typename T>
void IntervalSet<T>::set_empty(void)
{
	m_set.clear();
}

template <typename T>
void IntervalSet<T>::set_full(void)
{
	m_set.clear();
	Interval x;
	x.set_full();
	m_set.insert(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator&=(type_t x)
{
	return (*this) &= Interval(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator|=(type_t x)
{
	return (*this) |= Interval(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator^=(type_t x)
{
	return (*this) ^= Interval(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator-=(type_t x)
{
	return (*this) -= Interval(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator&=(const IntervalSet::Interval& x)
{
	return (*this) &= IntervalSet(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator|=(const IntervalSet::Interval& x)
{
	std::pair<typename set_t::iterator,bool> ins(m_set.insert(x));
	if (ins.second)
		normalize();
	return *this;
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator^=(const IntervalSet::Interval& x)
{
	return (*this) ^= IntervalSet(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator-=(const IntervalSet::Interval& x)
{
	return (*this) &= ~IntervalSet(x);
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator&=(const IntervalSet& x)
{
	IntervalSet y;
	y.m_set.swap(m_set);
	for (typename set_t::const_iterator xi(x.m_set.begin()), xe(x.m_set.end()); xi != xe; ++xi) {
		const Interval& xx(*xi);
		if (xx.is_empty())
			continue;
		for (typename set_t::const_iterator yi(y.m_set.begin()), ye(y.m_set.end()); yi != ye; ++yi) {
			const Interval& yy(*yi);
			if (yy.is_empty())
				continue;
			if (yy.get_lower() >= xx.get_upper())
				break;
			if (!yy.is_overlap_or_adjacent(xx))
				continue;
			m_set.insert(Interval(std::max(xx.get_lower(), yy.get_lower()), std::min(xx.get_upper(), yy.get_upper())));
		}
	}
	normalize();
	return *this;
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator|=(const IntervalSet& x)
{
	bool norm(false);
	for (typename set_t::const_iterator i(x.m_set.begin()), e(x.m_set.end()); i != e; ++i) {
		std::pair<typename set_t::iterator,bool> ins(m_set.insert(*i));
		norm = norm || ins.second;
	}
	if (norm)
		normalize();
	return *this;
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator^=(const IntervalSet& x)
{
	IntervalSet z(*this);
	z &= x;
	(*this) |= x;
	return (*this) -= z;
}

template <typename T>
IntervalSet<T>& IntervalSet<T>::operator-=(const IntervalSet& x)
{
	return (*this) &= ~x;
}

template <typename T>
IntervalSet<T> IntervalSet<T>::operator~() const
{
	IntervalSet z;
	type_t v(std::numeric_limits<type_t>::min());
	for (typename set_t::const_iterator i(m_set.begin()), e(m_set.end()); i != e; ++i) {
		const Interval& x(*i);
		if (x.is_empty())
			continue;
		if (v < x.get_lower())
		  z.m_set.insert(Interval(v, x.get_lower()));
		if (v < x.get_upper())
			v = x.get_upper();
	}
	if (v < std::numeric_limits<type_t>::max())
		z.m_set.insert(Interval(v, std::numeric_limits<type_t>::max()));
	return z;
}

template <typename T>
std::string IntervalSet<T>::to_str(void) const
{
	bool subseq(false);
	std::string r;
	for (typename set_t::const_iterator i(m_set.begin()), e(m_set.end()); i != e; ++i) {
		if (i->is_empty())
			continue;
		if (subseq)
			r += "|";
		subseq = true;
		r += i->to_str();
	}
	if (subseq)
		return r;
	return "[)";
}

template <typename T>
void IntervalSet<T>::normalize(void)
{
	for (typename set_t::iterator i(m_set.begin()), e(m_set.end()); i != e; ) {
		const Interval& x(*i);
		typename set_t::iterator in(i);
		++in;
		// remove empty subintervals
		if (x.is_empty()) {
			m_set.erase(i);
			i = in;
			//e = m_set.end();
			continue;
		}
		if (in == e)
			break;
		const Interval& y(*in);
		// remove empty subintervals
		if (y.is_empty()) {
			m_set.erase(in);
			//e = m_set.end();
			continue;
		}
		// merge adjacent
		if (!x.is_overlap_or_adjacent(y)) {
			++i;
			continue;
		}
		Interval z(std::min(x.get_lower(), y.get_lower()), std::max(x.get_upper(), y.get_upper()));
		m_set.erase(i);
		m_set.erase(in);
		std::pair<typename set_t::iterator,bool> ins(m_set.insert(z));
		i = ins.first;
		//e = m_set.end();
	}
}

template class IntervalSet<int32_t>;
