//
// C++ Implementation: FlightDeckWindow
//
// Description: Flight Deck Windows, Gyro Instrument Drawing Areas
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2011, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include "flightdeckwindow.h"

#include <memory>
#include <iomanip>

const int32_t FlightDeckWindow::RouteProfileDrawingArea::WeatherProfilePoint::invalidalt = std::numeric_limits<int32_t>::min();
const float FlightDeckWindow::RouteProfileDrawingArea::WeatherProfilePoint::invalidcover = std::numeric_limits<float>::quiet_NaN();

FlightDeckWindow::RouteProfileDrawingArea::WeatherProfilePoint::WeatherProfilePoint(double dist, int32_t zerodegisotherm,
										    float cldlowcover, int32_t cldlowbase, int32_t cldlowtop,
										    float cldmidcover, int32_t cldmidbase, int32_t cldmidtop,
										    float cldhighcover, int32_t cldhighbase, int32_t cldhightop,
										    float cldconvcover, int32_t cldconvbase, int32_t cldconvtop)
	: m_dist(dist), m_zerodegisotherm(zerodegisotherm), m_cldlowcover(cldlowcover), m_cldlowbase(cldlowbase), m_cldlowtop(cldlowtop),
	  m_cldmidcover(cldmidcover), m_cldmidbase(cldmidbase), m_cldmidtop(cldmidtop), m_cldhighcover(cldhighcover), m_cldhighbase(cldhighbase),
	  m_cldhightop(cldhightop), m_cldconvcover(cldconvcover), m_cldconvbase(cldconvbase), m_cldconvtop(cldconvtop)
{
}

void FlightDeckWindow::RouteProfileDrawingArea::WeatherProfilePoint::scaledist(double ds)
{
        m_dist *= ds;
}

void FlightDeckWindow::RouteProfileDrawingArea::WeatherProfilePoint::adddist(double d)
{
        m_dist += d;
}

void FlightDeckWindow::RouteProfileDrawingArea::WeatherProfile::scaledist(double ds)
{
        for (typename base_t::iterator i(this->begin()), e(this->end()); i != e; ++i)
                i->scaledist(ds);
}

void FlightDeckWindow::RouteProfileDrawingArea::WeatherProfile::adddist(double d)
{
        for (typename base_t::iterator i(this->begin()), e(this->end()); i != e; ++i)
                i->adddist(d);
}

double FlightDeckWindow::RouteProfileDrawingArea::WeatherProfile::get_dist(void) const
{
        if (this->empty())
                return 0;
        return this->operator[](this->size() - 1).get_dist();
}

FlightDeckWindow::RouteProfileDrawingArea::RouteProfileDrawingArea(BaseObjectType *castitem, const Glib::RefPtr<Builder>& refxml)
	: Gtk::DrawingArea(castitem), m_engine(0), m_route(0), m_scaledist(1), m_scaleelev(0.1), m_origindist(0), m_originelev(0),
	  m_dragx(-1), m_dragy(-1)

{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
#ifdef HAVE_GTKMM2
	signal_expose_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_expose_event));
#endif
#ifdef HAVE_GTKMM3
        signal_draw().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_draw));
#endif
        signal_show().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_show));
        signal_hide().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_hide));
        signal_button_press_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_press_event));
        signal_button_release_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_release_event));
        signal_motion_notify_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_motion_notify_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	m_dispatch.connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::async_result));
	//set_size_request(128, 128);
	//add_events(Gdk::EXPOSURE_MASK | Gdk::PROPERTY_CHANGE_MASK | Gdk::POINTER_MOTION_MASK | Gdk::POINTER_MOTION_HINT_MASK);
}

FlightDeckWindow::RouteProfileDrawingArea::RouteProfileDrawingArea(void)
	: m_engine(0), m_route(0), m_scaledist(1), m_scaleelev(0.1), m_origindist(0), m_originelev(0),
	  m_dragx(-1), m_dragy(-1)
{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
#ifdef HAVE_GTKMM2
	signal_expose_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_expose_event));
#endif
#ifdef HAVE_GTKMM3
        signal_draw().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_draw));
#endif
        signal_show().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_show));
        signal_hide().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_hide));
        signal_button_press_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_press_event));
        signal_button_release_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_button_release_event));
        signal_motion_notify_event().connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::on_motion_notify_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	m_dispatch.connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::async_result));
	//set_size_request(128, 128);
	//add_events(Gdk::EXPOSURE_MASK | Gdk::PROPERTY_CHANGE_MASK | Gdk::POINTER_MOTION_MASK | Gdk::POINTER_MOTION_HINT_MASK);
}

FlightDeckWindow::RouteProfileDrawingArea::~RouteProfileDrawingArea()
{
	async_cancel();
}

void FlightDeckWindow::RouteProfileDrawingArea::set_engine(Engine& eng)
{
	async_cancel();
	m_engine = &eng;
	m_profile.clear();
	m_wxprofile.clear();
	route_change();
	
}

void FlightDeckWindow::RouteProfileDrawingArea::set_route(const FPlanRoute& route)
{
	async_cancel();
	m_route = &route;
	m_profile.clear();
	m_wxprofile.clear();
	route_change();
}

void FlightDeckWindow::RouteProfileDrawingArea::route_change(void)
{
	async_cancel();
	m_profile.clear();
	m_wxprofile.clear();
	if (!get_is_drawable())
		return;
	async_startquery();
}

void FlightDeckWindow::RouteProfileDrawingArea::async_startquery(void)
{
	if (!m_engine || !m_route || m_route->get_nrwpt() < 2)
		return;
	async_cancel();
	async_clear();
	m_profilequery = m_engine->async_elevation_profile((*m_route)[0].get_coord(), (*m_route)[1].get_coord(), 5.);
	async_connectdone();
	compute_wxprofile();
}

void FlightDeckWindow::RouteProfileDrawingArea::dist_zoom_in(void)
{
	set_dist_zoom(get_dist_zoom() * 2.);
}

void FlightDeckWindow::RouteProfileDrawingArea::dist_zoom_out(void)
{
	set_dist_zoom(get_dist_zoom() * 0.5);
}

double FlightDeckWindow::RouteProfileDrawingArea::get_dist_zoom(void) const
{
	return m_scaledist;
}

void FlightDeckWindow::RouteProfileDrawingArea::set_dist_zoom(double z)
{
	double zz(std::max(std::min(z, 100.), 0.01));
	if (zz == m_scaledist)
		return;
	m_scaledist = zz;
	if (get_is_drawable())
		queue_draw();
}

void FlightDeckWindow::RouteProfileDrawingArea::elev_zoom_in(void)
{
	set_elev_zoom(get_elev_zoom() * 2.);
}

void FlightDeckWindow::RouteProfileDrawingArea::elev_zoom_out(void)
{
	set_elev_zoom(get_elev_zoom() * 0.5);
}

double FlightDeckWindow::RouteProfileDrawingArea::get_elev_zoom(void) const
{
	return m_scaleelev;
}

void FlightDeckWindow::RouteProfileDrawingArea::set_elev_zoom(double z)
{
	double zz(std::max(std::min(z, 10.), 0.001));
	if (zz == m_scaleelev)
		return;
	m_scaleelev = zz;
	if (get_is_drawable())
		queue_draw();
}

#ifdef HAVE_GTKMM2
bool FlightDeckWindow::RouteProfileDrawingArea::on_expose_event(GdkEventExpose *event)
{
        Glib::RefPtr<Gdk::Window> wnd(get_window());
        if (!wnd)
                return false;
        Cairo::RefPtr<Cairo::Context> cr(wnd->create_cairo_context());
	if (event) {
		GdkRectangle *rects;
		int n_rects;
		gdk_region_get_rectangles(event->region, &rects, &n_rects);
		for (int i = 0; i < n_rects; i++)
			cr->rectangle(rects[i].x, rects[i].y, rects[i].width, rects[i].height);
		cr->clip();
		g_free(rects);
	}
        return on_draw(cr);
}
#endif

bool FlightDeckWindow::RouteProfileDrawingArea::on_draw(const Cairo::RefPtr<Cairo::Context>& cr)
{
	cr->save();
	// background
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->paint();
	// scaling
	Gtk::Allocation allocation = get_allocation();
	const int width = allocation.get_width();
	const int height = allocation.get_height();
        cr->select_font_face("sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
        cr->set_font_size(12);
	int alt_width;
	int dist_height;
	int dist_line_height;
	double elevinc, elevstart;
	double distinc, diststart;
	{
		Cairo::TextExtents ext;
		cr->get_text_extents("99999", ext);
		alt_width = ext.width + 4;
		dist_line_height = ext.height + 4;
		dist_height = 2 * dist_line_height + 5;
		elevinc = pow(10., std::max(0., ceil(log10(ext.height * 5 / m_scaleelev))));
		elevstart = elevinc * ceil(m_originelev / elevinc);
		cr->get_text_extents("999", ext);
		distinc = pow(10., std::max(0., ceil(log10(ext.width * 3 / m_scaledist))));
		diststart = distinc * ceil(m_origindist / distinc);
	}
	// draw location labels
	if (m_route) {
		set_color_fplan(cr);
		for (unsigned int i = 0, j = m_route->get_nrwpt(); i < j; ++i) {
			int x;
			{
				double dist;
				if (!m_profile.size()) {
					continue;
				} else if (i >= m_profile.size()) {
					if (!m_profile.back().size())
						continue;
					dist = m_profile.back().back().get_dist();
				} else if (!m_profile[i].size()) {
					continue;
				} else {
					dist = m_profile[i].front().get_dist();
				}
				x = alt_width + Point::round<int,double>((dist - m_origindist) * m_scaledist);
			}
			if (x < alt_width || x > width)
				continue;
			Glib::ustring txt((*m_route)[i].get_icao_name());
			if (txt.empty())
				continue;
			Cairo::TextExtents ext;
			cr->get_text_extents(txt, ext);
			cr->move_to(x - ext.width / 2U - ext.x_bearing, height - 2U * dist_line_height + 2U - ext.y_bearing);
			cr->show_text(txt);
			cr->begin_new_path();
			cr->move_to(x - 2, height - 2 * dist_line_height);
			cr->rel_line_to(4, 0);
			cr->rel_line_to(-2, -4);
			cr->close_path();
			cr->fill();
		}
	}
	// normalize and draw terrain
	cr->save();
	Cairo::Matrix mnorm;
	cr->get_matrix(mnorm);
	cr->rectangle(alt_width, 0, width - alt_width, height - dist_height);
	cr->clip();
	set_color_sky(cr);
	cr->paint();
	cr->set_line_width(2);
	cr->translate(alt_width, height - dist_height);
	cr->scale(m_scaledist, -m_scaleelev);
	cr->translate(-m_origindist, -m_originelev);
	Cairo::Matrix mscaled;
	cr->get_matrix(mscaled);
	// draw low clouds
	for (WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
		if (std::isnan(pi->get_cldlowcover()) ||
		    pi->get_cldlowbase() == WeatherProfilePoint::invalidalt ||
		    pi->get_cldlowtop() == WeatherProfilePoint::invalidalt) {
			++pi;
			continue;
		}
		WeatherProfile::const_iterator pi2(pi);
		++pi2;
		if (pi2 == pe || std::isnan(pi2->get_cldlowcover()) ||
		    pi2->get_cldlowbase() == WeatherProfilePoint::invalidalt ||
		    pi2->get_cldlowtop() == WeatherProfilePoint::invalidalt) {
			pi = pi2;
			continue;
		}
		cr->move_to(pi->get_dist(), pi->get_cldlowbase());
		cr->line_to(pi->get_dist(), pi->get_cldlowtop());
		cr->line_to(pi2->get_dist(), pi2->get_cldlowtop());
		cr->line_to(pi2->get_dist(), pi2->get_cldlowbase());
		cr->close_path();
		double gray(1.0 - 0.25 * pi->get_cldlowcover());
		cr->set_source_rgba(gray, gray, gray, pi->get_cldlowcover());
		cr->fill();
		pi = pi2;
	}
	// draw mid clouds
	for (WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
		if (std::isnan(pi->get_cldmidcover()) ||
		    pi->get_cldmidbase() == WeatherProfilePoint::invalidalt ||
		    pi->get_cldmidtop() == WeatherProfilePoint::invalidalt) {
			++pi;
			continue;
		}
		WeatherProfile::const_iterator pi2(pi);
		++pi2;
		if (pi2 == pe || std::isnan(pi2->get_cldmidcover()) ||
		    pi2->get_cldmidbase() == WeatherProfilePoint::invalidalt ||
		    pi2->get_cldmidtop() == WeatherProfilePoint::invalidalt) {
			pi = pi2;
			continue;
		}
		cr->move_to(pi->get_dist(), pi->get_cldmidbase());
		cr->line_to(pi->get_dist(), pi->get_cldmidtop());
		cr->line_to(pi2->get_dist(), pi2->get_cldmidtop());
		cr->line_to(pi2->get_dist(), pi2->get_cldmidbase());
		cr->close_path();
		double gray(1.0 - 0.25 * pi->get_cldmidcover());
		cr->set_source_rgba(gray, gray, gray, pi->get_cldmidcover());
		cr->fill();
		pi = pi2;
	}
	// draw high clouds
	for (WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
		if (std::isnan(pi->get_cldhighcover()) ||
		    pi->get_cldhighbase() == WeatherProfilePoint::invalidalt ||
		    pi->get_cldhightop() == WeatherProfilePoint::invalidalt) {
			++pi;
			continue;
		}
		WeatherProfile::const_iterator pi2(pi);
		++pi2;
		if (pi2 == pe || std::isnan(pi2->get_cldhighcover()) ||
		    pi2->get_cldhighbase() == WeatherProfilePoint::invalidalt ||
		    pi2->get_cldhightop() == WeatherProfilePoint::invalidalt) {
			pi = pi2;
			continue;
		}
		cr->move_to(pi->get_dist(), pi->get_cldhighbase());
		cr->line_to(pi->get_dist(), pi->get_cldhightop());
		cr->line_to(pi2->get_dist(), pi2->get_cldhightop());
		cr->line_to(pi2->get_dist(), pi2->get_cldhighbase());
		cr->close_path();
		double gray(1.0 - 0.25 * pi->get_cldhighcover());
		cr->set_source_rgba(gray, gray, gray, pi->get_cldhighcover());
		cr->fill();
		pi = pi2;
	}
	// draw convective
	for (WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
		if (std::isnan(pi->get_cldconvcover()) ||
		    pi->get_cldconvbase() == WeatherProfilePoint::invalidalt ||
		    pi->get_cldconvtop() == WeatherProfilePoint::invalidalt) {
			++pi;
			continue;
		}
		WeatherProfile::const_iterator pi2(pi);
		++pi2;
		if (pi2 == pe || std::isnan(pi2->get_cldconvcover()) ||
		    pi2->get_cldconvbase() == WeatherProfilePoint::invalidalt ||
		    pi2->get_cldconvtop() == WeatherProfilePoint::invalidalt) {
			pi = pi2;
			continue;
		}
		cr->move_to(pi->get_dist(), pi->get_cldconvbase());
		cr->line_to(pi->get_dist(), pi->get_cldconvtop());
		cr->line_to(pi2->get_dist(), pi2->get_cldconvtop());
		cr->line_to(pi2->get_dist(), pi2->get_cldconvbase());
		cr->close_path();
		double gray(1.0 - 0.25 * pi->get_cldconvcover());
		cr->set_source_rgba(gray, gray, gray, pi->get_cldconvcover());
		cr->fill();
		pi = pi2;
	}
	// draw 0degC isotherm
	for (WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
		if (pi->get_zerodegisotherm() == WeatherProfilePoint::invalidalt) {
			++pi;
			continue;
		}
		cr->begin_new_path();
		cr->move_to(pi->get_dist(), pi->get_zerodegisotherm());
		WeatherProfile::const_iterator pi2(pi);
		++pi2;
		while (pi2 != pe && pi2->get_zerodegisotherm() != WeatherProfilePoint::invalidalt) {
			cr->line_to(pi2->get_dist(), pi2->get_zerodegisotherm());
			++pi2;
		}
		if (false && (pi2 - pi) <= 1) {
			pi = pi2;
			continue;
		}
		cr->set_matrix(mnorm);
		cr->set_source_rgb(1.0, 0.0, 0.0);
		cr->stroke();
		cr->set_matrix(mscaled);
		pi = pi2;
	}
	cr->set_matrix(mscaled);
	Cairo::Matrix mscaledmeter;
	cr->scale(1., Point::m_to_ft_dbl);
	cr->get_matrix(mscaledmeter);
	// draw max elev
	{
		cr->begin_new_path();
		cr->move_to(0, 0);
		double dist(0);
		for (profile_t::const_iterator pi(m_profile.begin()), pe(m_profile.end()); pi != pe; ++pi)
			for (Engine::ElevationProfileResult::Profile::const_iterator ppi(pi->begin()), ppe(pi->end()); ppi != ppe; ++ppi)
				cr->line_to(dist = ppi->get_dist(), ppi->get_maxelev());
		cr->line_to(dist, 0);
		cr->close_path();
	}
	cr->set_matrix(mnorm);
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->fill_preserve();
	set_color_terrain_fill(cr);
	cr->fill_preserve();
	set_color_terrain(cr);
	cr->stroke();
	cr->set_matrix(mscaledmeter);
	// draw elev
	{
		cr->begin_new_path();
		cr->move_to(0, 0);
		double dist(0);
		for (profile_t::const_iterator pi(m_profile.begin()), pe(m_profile.end()); pi != pe; ++pi)
			for (Engine::ElevationProfileResult::Profile::const_iterator ppi(pi->begin()), ppe(pi->end()); ppi != ppe; ++ppi)
				cr->line_to(dist = ppi->get_dist(), ppi->get_elev());
		cr->line_to(dist, 0);
		cr->close_path();
	}
	cr->set_matrix(mnorm);
	set_color_terrain_fill(cr);
	cr->fill_preserve();
	set_color_terrain(cr);
	cr->stroke();
	cr->restore();
	// draw scales
	set_color_marking(cr);
	cr->set_line_width(1);
	for (;; elevstart += elevinc) {
		int y(height - dist_height - Point::round<int,double>((elevstart - m_originelev) * m_scaleelev));
		if (y < 0)
			break;
		cr->move_to(alt_width, y);
		cr->line_to(width, y);
		cr->stroke();
		std::ostringstream oss;
		oss << std::fixed << std::setprecision(0) << elevstart;
		Cairo::TextExtents ext;
		cr->get_text_extents(oss.str(), ext);
		cr->move_to(alt_width - 2 - ext.width - ext.x_bearing, y - ext.height / 2 - ext.y_bearing);
		cr->show_text(oss.str());
	}
	for (;; diststart += distinc) {
		int x(alt_width + Point::round<int,double>((diststart - m_origindist) * m_scaledist));
		if (x >= width)
			break;
		cr->move_to(x, 0);
		cr->line_to(x, height - dist_height);
		cr->stroke();
		std::ostringstream oss;
		oss << std::fixed << std::setprecision(0) << diststart;
		Cairo::TextExtents ext;
		cr->get_text_extents(oss.str(), ext);
		cr->move_to(x - ext.width / 2U - ext.x_bearing, height - dist_line_height + 2U - ext.y_bearing);
		cr->show_text(oss.str());
	}
	cr->restore();
	// normalize and draw route
	if (m_route) {
		cr->save();
		cr->rectangle(alt_width, 0, width - alt_width, height - dist_height);
		cr->clip();
		set_color_fplan(cr);
		cr->set_line_width(2);
		cr->set_matrix(mscaled);
		bool first(true);
		cr->begin_new_path();
		for (unsigned int i = 0, j = m_route->get_nrwpt(); i < j; ++i) {
			double dist;
			{
				if (!m_profile.size()) {
					continue;
				} else if (i >= m_profile.size()) {
					if (!m_profile.back().size())
						continue;
					dist = m_profile.back().back().get_dist();
				} else if (!m_profile[i].size()) {
					continue;
				} else {
					dist = m_profile[i].front().get_dist();
				}
			}
			float alt((*m_route)[i].get_true_altitude());
			if (first)
				cr->move_to(dist, alt);
			else
				cr->line_to(dist, alt);
			first = false;
		}
		cr->set_matrix(mnorm);
		cr->stroke();
		cr->restore();
	}
	// end
	cr->restore();
	return true;
}

void FlightDeckWindow::RouteProfileDrawingArea::on_size_allocate(Gtk::Allocation& allocation)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        Gtk::DrawingArea::on_size_allocate(allocation);
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
}

void FlightDeckWindow::RouteProfileDrawingArea::on_show(void)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        Gtk::DrawingArea::on_show();
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	if (m_profile.empty())
	        async_startquery();
}

void FlightDeckWindow::RouteProfileDrawingArea::on_hide(void)
{
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        Gtk::DrawingArea::on_hide();
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	m_dragx = m_dragy = -1;
}

bool FlightDeckWindow::RouteProfileDrawingArea::on_button_press_event(GdkEventButton *event)
{
	if (event->button == 1) {
		m_dragx = event->x;
		m_dragy = event->y;
	}
        if (true)
                std::cerr << "routeprofile button_press " << event->x << ' ' << event->y << ' ' << event->button << std::endl;
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        if (Gtk::DrawingArea::on_button_press_event(event))
                return true;
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        return false;
}

bool FlightDeckWindow::RouteProfileDrawingArea::on_button_release_event(GdkEventButton *event)
{
	m_dragx = m_dragy = -1;
        if (true)
                std::cerr << "routeprofile button_release " << event->x << ' ' << event->y << ' ' << event->button << std::endl;
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        if (Gtk::DrawingArea::on_button_release_event(event))
                return true;
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        return false;
}

bool FlightDeckWindow::RouteProfileDrawingArea::on_motion_notify_event(GdkEventMotion *event)
{
	if (m_dragx >= 0 && m_dragy >= 0) {
		int dx(event->x - m_dragx), dy(event->y - m_dragy);
		m_dragx = event->x;
		m_dragy = event->y;
		m_origindist -= dx / m_scaledist;
		m_originelev += dy / m_scaleelev;
		if (m_profile.size() && m_profile.back().size()) {
			double dmax(m_profile.back().back().get_dist());
			Gtk::Allocation allocation = get_allocation();
			dmax -= (allocation.get_width() - 50) / m_scaledist;
			if (m_origindist > dmax)
				m_origindist = dmax;
		}
		if (m_origindist < 0)
			m_origindist = 0;
		if (m_originelev > 60000)
			m_originelev = 60000;
		if (m_originelev < 0)
			m_originelev = 0;
		queue_draw();
	}
        if (false)
                std::cerr << "routeprofile motion_notify " << event->x << ' ' << event->y << std::endl;
#ifdef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        if (Gtk::DrawingArea::on_motion_notify_event(event))
                return true;
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        return false;
}

void FlightDeckWindow::RouteProfileDrawingArea::async_cancel(void)
{
        Engine::ElevationProfileResult::cancel(m_profilequery);
}

void FlightDeckWindow::RouteProfileDrawingArea::async_clear(void)
{
	m_profilequery = Glib::RefPtr<Engine::ElevationProfileResult>();
}

void FlightDeckWindow::RouteProfileDrawingArea::async_connectdone(void)
{
        if (m_profilequery)
                m_profilequery->connect(sigc::mem_fun(*this, &RouteProfileDrawingArea::async_done));
}

void FlightDeckWindow::RouteProfileDrawingArea::async_done(void)
{
	m_dispatch();
}

void FlightDeckWindow::RouteProfileDrawingArea::async_result(void)
{
	if (!m_profilequery || !m_profilequery->is_done())
		return;
	if (m_profilequery->is_error()) {
		m_profilequery = Glib::RefPtr<Engine::ElevationProfileResult>();
		return;
	}
	m_profile.push_back(m_profilequery->get_result());
	m_profilequery = Glib::RefPtr<Engine::ElevationProfileResult>();
	if (!m_route || !m_engine || m_profile.size() >= m_route->get_nrwpt() || !m_profile.back().size()) {
		m_profile.clear();
		return;
	}
	if (m_profile.size() > 1) {
		unsigned int idx(m_profile.size() - 2U);
		m_profile.back().adddist(m_profile[idx].front().get_dist() + 
					 (*m_route)[idx].get_coord().spheric_distance_nmi_dbl((*m_route)[idx + 1U].get_coord()));
	}
	if (get_is_drawable())
		queue_draw();
	{
		unsigned int idx(m_profile.size());
		if (idx + 1U < m_route->get_nrwpt()) {
			async_cancel();
			async_clear();
			m_profilequery = m_engine->async_elevation_profile((*m_route)[idx].get_coord(), (*m_route)[idx + 1U].get_coord(), 5.);
			async_connectdone();
		}
	}
}

void FlightDeckWindow::RouteProfileDrawingArea::compute_wxprofile(void)
{
	m_wxprofile.clear();
	if (!m_engine || !m_route)
		return;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxzerodegisotherm;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldlowcover;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldlowbase;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldlowtop;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldmidcover;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldmidbase;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldmidtop;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldhighcover;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldhighbase;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldhightop;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldconvcover;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldconvbase;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> wxcldconvtop;
	Rect bbox(m_route->get_bbox().oversize_nmi(100.f));
	double dist(0);
	for (unsigned int wptnr = 1U, nrwpt = m_route->get_nrwpt(); wptnr < nrwpt; ++wptnr) {
		const FPlanWaypoint& wpt0((*m_route)[wptnr - 1U]);
		const FPlanWaypoint& wpt1((*m_route)[wptnr]);
		double distleg(wpt0.get_coord().spheric_distance_nmi_dbl(wpt1.get_coord()));
		if (distleg <= 0)
			continue;
		Point ptorig(wpt0.get_coord());
		Point ptdiff(wpt1.get_coord() - ptorig);
		gint64 timeorig(wpt0.get_time_unix());
		gint64 timediff(wpt1.get_time_unix() - timeorig);
		for (double t = 0, tinc = 1.0 / distleg;; t += tinc) {
			if (t > 1.0)
				t = 1.0;
			Point pt(Point(ptdiff.get_lon() * t, ptdiff.get_lat() * t) + ptorig);
			gint64 efftime(timeorig + t * timediff);
			bool first(wptnr == 1U && t == 0);
			if (first ||
			    (wxzerodegisotherm && (efftime < wxzerodegisotherm->get_minefftime() ||
						   efftime > wxzerodegisotherm->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_hgt),
											   efftime, GRIB2::surface_0degc_isotherm, 0));
                                wxzerodegisotherm = GRIB2::interpolate_results(bbox, ll, efftime, 0);
				if (true)
					std::cerr << "0degC Isotherm: wptnr " << wptnr << " t " << t << " ll " << ll.size()
						  << " interp " << (wxzerodegisotherm ? "yes" : "no") << std::endl;
			}
			if (first ||
			    (wxcldlowcover && (efftime < wxcldlowcover->get_minefftime() ||
					       efftime > wxcldlowcover->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
											   efftime, GRIB2::surface_low_cloud, 0));
                                wxcldlowcover = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldlowbase && (efftime < wxcldlowbase->get_minefftime() ||
					      efftime > wxcldlowbase->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_low_cloud_bottom, 0));
                                wxcldlowbase = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldlowtop && (efftime < wxcldlowtop->get_minefftime() ||
					     efftime > wxcldlowtop->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_low_cloud_top, 0));
                                wxcldlowtop = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldmidcover && (efftime < wxcldmidcover->get_minefftime() ||
					       efftime > wxcldmidcover->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
											   efftime, GRIB2::surface_middle_cloud, 0));
                                wxcldmidcover = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldmidbase && (efftime < wxcldmidbase->get_minefftime() ||
					      efftime > wxcldmidbase->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_middle_cloud_bottom, 0));
                                wxcldmidbase = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldmidtop && (efftime < wxcldmidtop->get_minefftime() ||
					     efftime > wxcldmidtop->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_middle_cloud_top, 0));
                                wxcldmidtop = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldhighcover && (efftime < wxcldhighcover->get_minefftime() ||
						efftime > wxcldhighcover->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
											   efftime, GRIB2::surface_top_cloud, 0));
                                wxcldhighcover = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldhighbase && (efftime < wxcldhighbase->get_minefftime() ||
					       efftime > wxcldhighbase->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_top_cloud_bottom, 0));
                                wxcldhighbase = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldhightop && (efftime < wxcldhightop->get_minefftime() ||
					      efftime > wxcldhightop->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_top_cloud_top, 0));
                                wxcldhightop = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}

			if (first ||
			    (wxcldconvcover && (efftime < wxcldconvcover->get_minefftime() ||
						efftime > wxcldconvcover->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
											   efftime, GRIB2::surface_convective_cloud, 0));
                                wxcldconvcover = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldconvbase && (efftime < wxcldconvbase->get_minefftime() ||
					       efftime > wxcldconvbase->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_convective_cloud_bottom, 0));
                                wxcldconvbase = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			if (first ||
			    (wxcldconvtop && (efftime < wxcldconvtop->get_minefftime() ||
					      efftime > wxcldconvtop->get_maxefftime()))) {
				GRIB2::layerlist_t ll(m_engine->get_grib2_db().find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
											   efftime, GRIB2::surface_convective_cloud_top, 0));
                                wxcldconvtop = GRIB2::interpolate_results(bbox, ll, efftime, 0);
			}
			int32_t zerodegisotherm = WeatherProfilePoint::invalidalt;
			if (wxzerodegisotherm) {
				float x(wxzerodegisotherm->operator()(pt, efftime, 0));
				if (!std::isnan(x))
					zerodegisotherm = x * Point::m_to_ft;
			}
			float cldlowcover = WeatherProfilePoint::invalidcover;
			if (wxcldlowcover) {
				float x(wxcldlowcover->operator()(pt, efftime, 0));
				if (!std::isnan(x))
					cldlowcover = 0.01 * x;
			}
			int32_t cldlowbase = WeatherProfilePoint::invalidalt;
			if (wxcldlowbase) {
				float x(wxcldlowbase->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldlowbase = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			int32_t cldlowtop = WeatherProfilePoint::invalidalt;
			if (wxcldlowtop) {
				float x(wxcldlowtop->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldlowtop = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			float cldmidcover = WeatherProfilePoint::invalidcover;
			if (wxcldmidcover) {
				float x(wxcldmidcover->operator()(pt, efftime, 0));
				if (!std::isnan(x))
					cldmidcover = 0.01 * x;
			}
			int32_t cldmidbase = WeatherProfilePoint::invalidalt;
			if (wxcldmidbase) {
				float x(wxcldmidbase->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldmidbase = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			int32_t cldmidtop = WeatherProfilePoint::invalidalt;
			if (wxcldmidtop) {
				float x(wxcldmidtop->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldmidtop = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			float cldhighcover = WeatherProfilePoint::invalidcover;
			if (wxcldhighcover) {
				float x(wxcldhighcover->operator()(pt, efftime, 0));
				if (!std::isnan(x))
					cldhighcover = 0.01 * x;
			}
			int32_t cldhighbase = WeatherProfilePoint::invalidalt;
			if (wxcldhighbase) {
				float x(wxcldhighbase->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldhighbase = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			int32_t cldhightop = WeatherProfilePoint::invalidalt;
			if (wxcldhightop) {
				float x(wxcldhightop->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldhightop = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			float cldconvcover = WeatherProfilePoint::invalidcover;
			if (wxcldconvcover) {
				float x(wxcldconvcover->operator()(pt, efftime, 0));
				if (!std::isnan(x))
					cldconvcover = 0.01 * x;
			}
			int32_t cldconvbase = WeatherProfilePoint::invalidalt;
			if (wxcldconvbase) {
				float x(wxcldconvbase->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldconvbase = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}
			int32_t cldconvtop = WeatherProfilePoint::invalidalt;
			if (wxcldconvtop) {
				float x(wxcldconvtop->operator()(pt, efftime, 0));
				if (!std::isnan(x)) {
					float alt(0);
					IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, x * 0.01);
					cldconvtop = Point::round<int,float>(alt * Point::m_to_ft);
				}
			}

			m_wxprofile.push_back(WeatherProfilePoint(dist + t * distleg, zerodegisotherm,
								  cldlowcover, cldlowbase, cldlowtop,
								  cldmidcover, cldmidbase, cldmidtop,
								  cldhighcover, cldhighbase, cldhightop,
								  cldconvcover, cldconvbase, cldconvtop));
			if (t == 1.0)
				break;
		}
		dist += distleg;
	}
}
