//
// C++ Implementation: Navigation
//
// Description: Navigation
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include "sensors.h"
#include "fplan.h"
#include "wind.h"
#include "prefs.h"

Sensors::FPlanWptExtra::FPlanWptExtra(time_t ft, double truealt, double tt, double th, double decl,
				      double dist, double fuel, double tas, double rpm, double mp)
	: m_flighttime(ft), m_truealt(truealt), m_tt(tt), m_th(th), m_decl(decl), m_dist(dist), m_fuel(fuel), m_tas(tas), m_rpm(rpm), m_mp(mp)
{
}

double Sensors::FPlanWptExtra::get_magnetictrack(void) const
{
	double x(get_truetrack() - get_declination());
	if (x < 0)
		x += 360;
	if (x >= 360)
		x -= 360;
	return x;
}

double Sensors::FPlanWptExtra::get_magneticheading(void) const
{
	double x(get_trueheading() - get_declination());
	if (x < 0)
		x += 360;
	if (x >= 360)
		x -= 360;
	return x;
}

const double Sensors::track_to_deg;
const double Sensors::deg_to_track;
const double Sensors::track_to_rad;
const double Sensors::rad_to_track;

void Sensors::set_acft_cruise_rpm(double rpm)
{
	m_cruiserpm = rpm;
	get_configfile().set_double("main", "cruiserpm", m_cruiserpm);
	save_config();
	update_sensors_core(change_altitude);
}

void Sensors::set_acft_cruise_mp(double mp)
{
	m_cruisemp = mp;
	get_configfile().set_double("main", "cruisemp", m_cruisemp);
	save_config();
	update_sensors_core(change_altitude);
}

void Sensors::set_acft_cruise_bhp(double bhp)
{
	m_cruisebhp = bhp;
	get_configfile().set_double("main", "cruisebhp", m_cruisebhp);
	save_config();
	update_sensors_core(change_altitude);
}

void Sensors::set_acft_iasmode(iasmode_t iasmode)
{
	m_iasmode = iasmode;
	get_configfile().set_integer("main", "autoias", (unsigned int)m_iasmode);
	save_config();
}

void Sensors::get_acft_cruise(double& rpm, double& mp, double& bhp)
{
	rpm = mp = bhp = std::numeric_limits<double>::quiet_NaN();
	bool constspeed(get_aircraft().get_cruise().is_constantspeed());
	if (get_acft_iasmode() != Sensors::iasmode_bhp_rpm || constspeed)
		rpm = get_acft_cruise_rpm();
	if (get_acft_iasmode() != Sensors::iasmode_bhp_rpm && constspeed)
		mp = get_acft_cruise_mp();
	if (get_acft_iasmode() == Sensors::iasmode_bhp_rpm)
		bhp = get_acft_cruise_bhp();
}

const Sensors::FPlanWptExtra& Sensors::get_route_extrainfo(uint16_t wpt) const
{
	static const FPlanWptExtra dummy;
	if (wpt < m_routeextra.size())
		return m_routeextra[wpt];
	return dummy;
}

void Sensors::nav_fplan_modified(void)
{
	if (!m_route.is_dirty())
		return;
	recompute_wptextra();
	if (m_route.get_nrwpt())
		m_route.save_fp();
	m_sensorschange(change_fplan);
}

void Sensors::nav_fplan_setdeparture(time_t offblock, unsigned int taxitime)
{
	m_route.set_time_offblock_unix(offblock);
	if (m_route.get_nrwpt())
		m_route[0].set_time_unix(offblock + taxitime);
	recompute_wptextra();
	recompute_times(1);
	m_sensorschange(change_fplan);
}

void Sensors::nav_load_fplan(FPlanRoute::id_t id)
{
	close_track();
	m_route.load_fp(id);
	recompute_wptextra();
	m_sensorschange(change_fplan);
}

void Sensors::nav_new_fplan(void)
{
	close_track();
	m_route.new_fp();
	m_routeextra.clear();
}

void Sensors::nav_duplicate_fplan(void)
{
	m_route.duplicate_fp();
}

void Sensors::nav_delete_fplan(void)
{
	close_track();
	m_route.delete_fp();
	m_routeextra.clear();
}

void Sensors::nav_reverse_fplan(void)
{
	m_route.reverse_fp();
	recompute_wptextra();
}

void Sensors::nav_insert_wpt(uint32_t nr, const FPlanWaypoint& wpt)
{
	m_route.insert_wpt(nr, wpt);
	if (nr >= m_routeextra.size())
		m_routeextra.push_back(FPlanWptExtra());
	else
		m_routeextra.insert(m_routeextra.begin() + nr, FPlanWptExtra());
	recompute_wptextra();
}

void Sensors::nav_delete_wpt(uint32_t nr)
{
	m_route.delete_wpt(nr);
	if (nr < m_routeextra.size())
		m_routeextra.erase(m_routeextra.begin() + nr);
	recompute_wptextra();
}

void Sensors::nav_set_brg2(const Glib::ustring& dest)
{
	m_navbrg2dest = dest;
	if (m_navbrg2ena)
		m_sensorschange(change_navigation);
}

void Sensors::nav_set_brg2(const Point& pt)
{
	m_navbrg2 = pt;
	if (m_navbrg2ena)
		if (update_navigation() & change_navigation)
			m_sensorschange(change_navigation);
}

void Sensors::nav_set_brg2(bool enable)
{
	if (m_navbrg2ena == enable)
		return;
	m_navbrg2ena = enable;
	update_navigation();
	m_sensorschange(change_navigation);
}

void Sensors::nav_set_hold(double track)
{
	if (std::isnan(track)) {
		nav_set_hold(false);
		return;
	}
	m_navholdtrack = track * deg_to_track;
	if (m_navmode == navmode_fpl)
		m_navmode = navmode_fplhold;
	m_sensorschange(change_navigation);
}

void Sensors::nav_set_hold(bool enable)
{
	switch (m_navmode) {
	case navmode_fpl:
		if (!enable)
			break;
		m_navmode = navmode_fplhold;
		m_sensorschange(change_navigation);
		break;

	case navmode_fplhold:
		if (enable)
			break;
		m_navmode = navmode_fpl;
		m_sensorschange(change_navigation);
		break;

	case navmode_direct:
		if (!enable)
			break;
		m_navmode = navmode_directhold;
		m_sensorschange(change_navigation);
		break;

	case navmode_directhold:
		if (enable)
			break;
		m_navmode = navmode_direct;
		m_sensorschange(change_navigation);
		break;

	default:
		break;
	}
}

void Sensors::nav_nearestleg(void)
{
	if (m_route.get_nrwpt() <= 0)
		return;
	unsigned int nrsti(0);
	double nrstdist(m_position.spheric_distance_nmi_dbl(m_route[0].get_coord()));
	for (unsigned int i = 1; i < m_route.get_nrwpt(); ++i) {
		double d(m_position.spheric_distance_nmi_dbl(m_route[i].get_coord()));
		if (d > nrstdist)
			continue;
		nrsti = i;
		nrstdist = d;
	}
	double track(std::numeric_limits<double>::quiet_NaN());
	if (nrsti > 0)
		track = m_route[nrsti - 1].get_coord().spheric_true_course(m_route[nrsti].get_coord());
	nav_directto(m_route[nrsti].get_icao_name(), m_route[nrsti].get_coord(), track, m_route[nrsti].get_altitude(), m_route[nrsti].get_flags(), nrsti + 1);
}

void Sensors::nav_directto(const Glib::ustring& dest, const Point& coord, double track, int32_t alt, uint16_t altflags, uint16_t wptnr)
{
	if (std::isnan(track))
		track = m_position.spheric_true_course(coord);
	m_navcurdest = dest;
	m_navcur = m_navnext = coord;
	m_navcuralt = alt;
	m_navcuraltflags = altflags;
	m_navcurtrack = m_navnexttrack = m_navholdtrack = track * deg_to_track;
	m_navcurwpt = ~0U;
	m_navmode = navmode_direct;
	m_navblock = m_position.simple_box_km(0.02);
	if (wptnr <= m_route.get_nrwpt()) {
		m_navcurwpt = wptnr;
		m_navmode = navmode_fpl;
		++wptnr;
		if (wptnr <= m_route.get_nrwpt()) {
			const FPlanWaypoint& wpt(m_route[wptnr - 1]);
			m_navnext = wpt.get_coord();
			m_navnexttrack = m_navcur.spheric_true_course(m_navnext) * deg_to_track;
		}
		if (m_navcurwpt >= 1) {
			if (m_positionok && !std::isnan(m_groundspeed)) {
				m_navcurdist = m_position.spheric_distance_nmi_dbl(m_navcur);
				FPlanWaypoint& wpt(m_route[m_navcurwpt - 1]);
				wpt.save_time();
				wpt.set_time_unix(time(0) + m_navcurdist / std::max(m_groundspeed, 10.0) * 3600.0);
				recompute_times(m_navcurwpt);
			}
		} else {
			m_route.save_time_offblock();
			m_route.set_time_offblock_unix(time(0));
			if (m_route.get_nrwpt() > 0) {
				FPlanWaypoint& wpt(m_route[0]);
				wpt.save_time();
				wpt.set_time_unix(m_route.get_time_offblock_unix() + 5 * 60);
				recompute_times(1);
			}
		}
	} else if (wptnr == m_route.get_nrwpt() + 1) {
		m_navcurwpt = wptnr;
		m_navmode = navmode_fpl;
		m_route.save_time_onblock();
		m_route.set_time_onblock_unix(time(0));
	}
	m_sensorschange((update_navigation() & change_navigation) | change_fplan);
}

void Sensors::nav_finalapproach(const AirportsDb::Airport::FASNavigate& fas)
{
	if (!fas.is_valid())
		return;
	m_navfas = fas;
	m_navcurdest = fas.get_airport() + " " + fas.get_ident();
	m_navcur = m_navnext = fas.get_fpap();
	m_navcuralt = fas.get_thralt();
	m_navcuraltflags = 0;
	m_navcurtrack = m_navnexttrack = m_navholdtrack = fas.get_course();
	m_navcurwpt = ~0U;
	m_navmode = navmode_finalapproach;
	m_navblock = m_position.simple_box_km(0.02);
	m_sensorschange(update_navigation() & change_navigation);
}

Sensors::fplanwind_t Sensors::get_fplan_wind(void) const
{
	if ((m_navmode == navmode_fpl || m_navmode == navmode_fplhold) &&
	    m_navcurwpt >= 1 && m_navcurwpt <= m_route.get_nrwpt()) {
		const FPlanWaypoint& wpt(m_route[m_navcurwpt - 1]);
		return fplanwind_t(wpt.get_winddir_deg(), wpt.get_windspeed_kts());
	}
	return get_fplan_wind(m_position);
}

Sensors::fplanwind_t Sensors::get_fplan_wind(const Point& coord) const
{
	fplanwind_t r(0, 0);
	uint64_t mindist(std::numeric_limits<uint64_t>::max());
	for (unsigned int i = 0; i < m_route.get_nrwpt(); ++i) {
		const FPlanWaypoint& wpt(m_route[i]);
		uint64_t dist(coord.simple_distance_rel(wpt.get_coord()));
		if (dist > mindist)
			continue;
		mindist = dist;
		r.first = wpt.get_winddir_deg();
		r.second = wpt.get_windspeed_kts();
	}
	return r;
}

void Sensors::recompute_wptextra(void)
{
	m_routeextra.resize(m_route.get_nrwpt());
	if (m_routeextra.empty())
		return;
	AirData ad;
	ad.set_qnh_tempoffs(get_altimeter_qnh(), get_altimeter_tempoffs());
        WMM wmm;
        time_t curtime(time(0));
	{
		const FPlanWaypoint& wpt(m_route[0]);
		FPlanWptExtra& wpte(m_routeextra[0]);
		wpte.set_flighttime(0);
		wpte.set_fuel(0);
		wpte.set_rpm(0);
		wpte.set_mp(0);
		if (wpt.get_flags() & FPlanWaypoint::altflag_standard)
			ad.set_pressure_altitude(wpt.get_altitude());
		else
			ad.set_true_altitude(wpt.get_altitude());
		wpte.set_truealt(ad.get_true_altitude());
		{
			double tm(get_aircraft().get_climb().altitude_to_time(ad.get_density_altitude()));
			wpte.set_tas(get_aircraft().get_climb().time_to_tas(tm));
		}
		wmm.compute(ad.get_true_altitude() * (FPlanWaypoint::ft_to_m / 1000.0), wpt.get_coord(), curtime);
		wpte.set_truetrack(0);
		wpte.set_trueheading(0);
		wpte.set_declination(wmm.get_dec());
		wpte.set_dist(0);
	}
	for (unsigned int i = 1; i < m_routeextra.size(); ++i) {
		const FPlanWaypoint& wpt(m_route[i]);
		FPlanWptExtra& wpte(m_routeextra[i]);
		const FPlanWaypoint& wpto(m_route[i - 1]);
		const FPlanWptExtra& wpteo(m_routeextra[i - 1]);
		wpte.set_tas(0);
		wpte.set_rpm(0);
		wpte.set_mp(0);
		double dist(wpto.get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
		double tt(wpto.get_coord().spheric_true_course(wpt.get_coord()));
		wpte.set_dist(dist);
		wpte.set_truetrack(tt);
		double prevda(ad.get_density_altitude());
		double prevta(ad.get_true_altitude());
		if (wpt.get_flags() & FPlanWaypoint::altflag_standard)
			ad.set_pressure_altitude(wpt.get_altitude());
		else
			ad.set_true_altitude(wpt.get_altitude());
		double nextda(ad.get_density_altitude());
		double ltime(wpteo.get_flighttime()), fuel(wpteo.get_fuel());
		double th(tt);
		Wind<double> wind;
		wind.set_wind(wpto.get_winddir_deg(), wpto.get_windspeed_kts());
		// check for climb segment
		bool climbseg(nextda - prevda > 100);
		if (climbseg) {
			double prevt(get_aircraft().get_climb().altitude_to_time(prevda));
			double nextt(get_aircraft().get_climb().altitude_to_time(nextda));
			double prevd(get_aircraft().get_climb().time_to_distance(prevt));
			double nextd(get_aircraft().get_climb().time_to_distance(nextt));
			double climbd(nextd - prevd);
			// check for wind
			double speedfactor(1);
			if (wind.get_windspeed() > 0) {
				double tas(get_aircraft().get_climb().time_to_tas(prevt));
				wind.set_crs_tas(tt, tas);
				if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0) {
					climbd = 0;
				} else {
					th = wind.get_hdg();
					speedfactor = wind.get_gs() / tas;
					climbd *= speedfactor;
					if (false)
						std::cerr << "recompute_wptextra: climb: windcorr: TT " << tt << " TH " << th
							  << " TAS " << tas << " GS " << wind.get_gs() << " Wind " << wind.get_winddir() << '/' << wind.get_windspeed() << std::endl;
				}
			}
			if (climbd <= dist) {
				dist -= climbd;
			} else {
				// climb not finished at waypoint
				nextt = get_aircraft().get_climb().distance_to_time(prevd + dist / speedfactor);
				dist = 0;
				ad.set_density_altitude(get_aircraft().get_climb().time_to_altitude(nextt));
			}
			ltime += nextt - prevt;
			fuel += get_aircraft().get_climb().time_to_fuel(nextt) -
				get_aircraft().get_climb().time_to_fuel(prevt);
			wpte.set_tas(get_aircraft().get_climb().time_to_tas(nextt));
		}
		// check for cruise segment
		if (dist > 0) {
			double tas(0), fuelflow(0), bhp, rpm, mp;
			get_acft_cruise(rpm, mp, bhp);
			double da(ad.get_density_altitude());
			get_aircraft().get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
			if (!climbseg)
				wpte.set_tas(tas);
			double gs(tas);
			if (wind.get_windspeed() > 0) {
				wind.set_crs_tas(tt, tas);
				if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0) {
					gs = 1;
				} else if (!climbseg) {
					th = wind.get_hdg();
					gs = wind.get_gs();
					if (false)
						std::cerr << "recompute_wptextra: cruise: windcorr: TT " << tt << " TH " << th
							  << " TAS " << tas << " GS " << gs << " Wind " << wind.get_winddir() << '/' << wind.get_windspeed() << std::endl;
				}
			}
			double ct(dist / gs);
			ltime += ct * 3600.0;
			fuel += ct * fuelflow;
			if (!std::isnan(rpm) && rpm > 0)
				wpte.set_rpm(rpm);
			if (!std::isnan(mp) && mp > 0)
				wpte.set_mp(mp);
		}
		wpte.set_flighttime(ltime);
		wpte.set_fuel(fuel);
		double nextta(ad.get_true_altitude());
		wpte.set_truealt(nextta);
		{
			double nextta(ad.get_true_altitude());
			wmm.compute((prevta + nextta) * (FPlanWaypoint::ft_to_m / 2000.0), wpto.get_coord().halfway(wpt.get_coord()), curtime);
			wpte.set_declination(wmm.get_dec());
		}
		wpte.set_trueheading(th);
	}
}

void Sensors::recompute_times(unsigned int wptnr)
{
	if (true) {
		recompute_wptextra();
		for (; wptnr < m_route.get_nrwpt(); ++wptnr) {
			FPlanWaypoint& wpt(m_route[wptnr]);
			if (!wptnr) {
				wpt.set_time_unix(time(0));
				continue;
			}
			const FPlanWaypoint& wpto(m_route[wptnr - 1]);
			const FPlanWptExtra& wpte(m_routeextra[wptnr]);
			const FPlanWptExtra& wpteo(m_routeextra[wptnr - 1]);
			wpt.set_time_unix(wpto.get_time_unix() + wpte.get_flighttime() - wpteo.get_flighttime());
		}
		if (wptnr == m_route.get_nrwpt() && m_route.get_nrwpt()) {
			FPlanWaypoint& wpt(m_route[m_route.get_nrwpt() - 1]);
			m_route.set_time_onblock_unix(wpt.get_time_unix() + 5U * 60U);
		}
		return;
	}
	double smul(m_groundspeed);
	if (std::isnan(smul) || smul < m_tkoffldgspeed) {
		if (m_engine)
			smul = m_engine->get_prefs().vcruise;
		else
			smul = 100;
	}
	smul = std::max(smul, 10.0);
	smul = 3600 / smul;
	for (; wptnr < m_route.get_nrwpt(); ++wptnr) {
		FPlanWaypoint& wpt(m_route[wptnr]);
		if (!wptnr) {
			wpt.set_time_unix(time(0));
			continue;
		}
		const FPlanWaypoint& wpto(m_route[wptnr - 1]);
		double dist(wpto.get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
		wpt.set_time_unix(wpto.get_time_unix() + smul * dist);
		if (false)
			std::cerr << "recompute_times[" << wptnr << "] " << wpt.get_icao_name() << " prevtime " << wpto.get_time_unix()
				  << " dist " << dist << " mul " << smul << std::endl;
	}
	if (wptnr == m_route.get_nrwpt() && m_route.get_nrwpt()) {
		FPlanWaypoint& wpt(m_route[m_route.get_nrwpt() - 1]);
		m_route.set_time_onblock_unix(wpt.get_time_unix() + 5 * 60);
	}
}

void Sensors::nav_off(void)
{
	if (m_navmode == navmode_off)
		return;
	m_navcurdist = m_navcurxtk = m_navroc = m_navtimetowpt = std::numeric_limits<double>::quiet_NaN();
	m_navcurtrackerr = m_navnexttrackerr = 0;
	m_sensorschange(change_navigation | change_fplan);
}

double Sensors::nav_get_track_true(void) const
{
	if (!nav_is_on())
		return std::numeric_limits<double>::quiet_NaN();
	return m_navcurtrack * track_to_deg;
}

double Sensors::nav_get_track_mag(void) const
{
	double t(nav_get_track_true());
	if (std::isnan(t))
		return t;
	t -= get_declination();
	if (t < 0)
		t += 360;
	else if (t >= 360)
		t -= 360;
	return t;
}

time_t Sensors::nav_get_flighttime(void) const
{
	if (m_navflighttimerunning)
		return time(0) - m_navflighttime;
	return m_navflighttime;
}

void Sensors::close_track(void)
{
        if (!m_track.is_valid())
		return;
	if (m_engine) {
		m_track.set_modtime(time(0));
		m_engine->get_tracks_db().save(m_track);
        }
        m_track = TracksDb::Track();
}

Sensors::change_t Sensors::update_navigation(change_t changemask)
{
	if (changemask & change_position) {
		changemask |= change_navigation;
		if (!m_navbrg2ena || !m_positionok) {
			m_navbrg2tracktrue = m_navbrg2trackmag = m_navbrg2dist = std::numeric_limits<double>::quiet_NaN();
		} else {
			m_navbrg2tracktrue = m_position.spheric_true_course(m_navbrg2);
			m_navbrg2trackmag = m_navbrg2tracktrue - get_declination();
			if (m_navbrg2trackmag < 0)
				m_navbrg2trackmag += 360;
			else if (m_navbrg2trackmag >= 360)
				m_navbrg2trackmag -= 360;
			m_navbrg2dist = m_position.spheric_distance_nmi_dbl(m_navbrg2);
		}
		if (m_navmode == navmode_off || !m_positionok) {
			m_navcurdist = m_navcurxtk = m_navroc = m_navtimetowpt = std::numeric_limits<double>::quiet_NaN();
			m_navcurtrackerr = m_navnexttrackerr = 0;
		} else {
			if ((m_navmode == navmode_fpl || m_navmode == navmode_fplhold) && m_navcurwpt < 1) {
				if (!m_navblock.is_inside(m_position)) {
					changemask |= change_fplan;
					m_navlastwpttime = time(0);
					m_route.set_time_offblock_unix(m_navlastwpttime);
					m_navcurwpt = 1;
					if (m_route.get_nrwpt() > 0) {
						FPlanWaypoint& wpt(m_route[0]);
						wpt.save_time();
						wpt.set_time_unix(m_route.get_time_offblock_unix() + 5 * 60);
						recompute_times(1);
						m_navcurdest = wpt.get_icao_name();
						m_navcur = m_navnext = wpt.get_coord();
						m_navcuralt = wpt.get_altitude();
						m_navcuraltflags = wpt.get_flags();
						m_navcurtrack = m_navnexttrack = m_navholdtrack = 0;
						if (m_route.get_nrwpt() >= 2) {
							const FPlanWaypoint& wpt1(m_route[1]);
							m_navnext = wpt1.get_coord();
							m_navnexttrack = m_navcur.spheric_true_course(m_navnext) * deg_to_track;
						}
						// open track
						m_track = TracksDb::Track();
						m_track.make_valid();
						m_track.set_sourceid(NewWaypointWindow::make_sourceid());
						{
							const FPlanWaypoint& wpt0(m_route[0]);
							const FPlanWaypoint& wpt1(m_route[m_route.get_nrwpt()-1]);
							m_track.set_fromicao(wpt0.get_icao());
							m_track.set_fromname(wpt0.get_name());
							m_track.set_toicao(wpt1.get_icao());
							m_track.set_toname(wpt1.get_name());
						}
						m_track.set_notes(m_route.get_note());
						m_track.set_offblocktime(m_navlastwpttime);
						m_track.set_takeofftime(m_navlastwpttime);
						m_track.set_landingtime(m_navlastwpttime);
						m_track.set_onblocktime(m_navlastwpttime);
						m_track.set_modtime(m_navlastwpttime);
					}
				}
			} else if ((m_navmode == navmode_fpl || (m_navmode == navmode_fplhold && m_navcurwpt <= 1)) && m_navcurwpt < m_route.get_nrwpt()) {
				bool nextwpt(false);
				time_t deltat(0);
				if (m_navcurwpt <= 1) {
					nextwpt = !std::isnan(m_groundspeed) && m_groundspeed > m_tkoffldgspeed;
					if (nextwpt) {
						m_navflighttime = time(0);
						m_navflighttimerunning = true;
						if (m_track.is_valid()) {
							m_track.set_takeofftime(m_navflighttime);
							m_track.set_landingtime(m_navflighttime);
							m_track.set_onblocktime(m_navflighttime);
						}
					}
				} else {
					m_navcurdist = m_position.spheric_distance_nmi_dbl(m_navcur);
					uint32_t trkcur(m_position.spheric_true_course(m_navcur) * deg_to_track);
					m_navcurtrackerr = m_navcurtrack - trkcur;
					// try to fit a circle segment into triangle m_position - m_navcur - m_navnext
					// determine distance from m_navcur to advance such that a smooth standard rate turn
					// will exactly intercept the new track
					int32_t alpha(m_navnexttrack - trkcur + (1 << 31));
					alpha = std::max(abs(alpha), (int32_t)(2*deg_to_track));
					double stdturnradius(m_groundspeed);
					if (std::isnan(stdturnradius))
						stdturnradius = 0;
					stdturnradius = std::max(stdturnradius, 10.0);
					stdturnradius *= (1.0 / 60.0 / M_PI);
					double distadv(stdturnradius);
					{
						// do not divide by tangent to avoid numeric problems
						double angle(alpha * (0.5 * track_to_rad));
						distadv *= cos(angle) / sin(angle);
					}
					nextwpt = m_navcurdist <= distadv;
					if (nextwpt) {
						// correct for time required to fly half the arc;
						// the time stored for the current waypoint is the current time
						// plus time for half the arc
						double dt(alpha * (track_to_rad / M_PI));
						dt = 1 - dt;
						dt *= 60;
						if (!std::isnan(m_groundspeed) && m_groundspeed > 10)
							dt = std::min(dt,  m_position.spheric_distance_nmi_dbl(m_navnext) / m_groundspeed * (3600 * 0.5));
						deltat = Point::round<time_t,double>(dt);
					}
				}
				if (nextwpt) {
					// advance waypoint
					changemask |= change_fplan;
					m_navlastwpttime = time(0);
					m_route[m_navcurwpt - 1].save_time();
					m_route[m_navcurwpt - 1].set_time_unix(m_navlastwpttime + deltat);
					m_navcuralt = m_route[m_navcurwpt].get_altitude();
					m_navcuraltflags = m_route[m_navcurwpt].get_flags();
					m_navcurdest = m_route[m_navcurwpt].get_icao_name();
					m_navcurtrack = m_navholdtrack = m_navnexttrack;
					m_navcur = m_navnext;
					++m_navcurwpt;
					if (m_navcurwpt < m_route.get_nrwpt()) {
						m_navnext = m_route[m_navcurwpt].get_coord();
						m_navnexttrack = m_navcur.spheric_true_course(m_navnext) * deg_to_track;
					}
					recompute_times(m_navcurwpt - 1);
				}
			} else if ((m_navmode == navmode_fpl || m_navmode == navmode_fplhold) && m_navcurwpt > m_route.get_nrwpt()) {
				if (!m_navblock.is_inside(m_position)) {
					changemask |= change_fplan;
					m_navlastwpttime = time(0);
					m_route.set_time_onblock_unix(m_navlastwpttime);
					m_navblock = m_position.simple_box_km(0.02);
					if (m_track.is_valid())
						m_track.set_onblocktime(m_navlastwpttime);
				}
			} else if ((m_navmode == navmode_fpl || m_navmode == navmode_fplhold) && !std::isnan(m_groundspeed) &&
				   m_groundspeed < m_tkoffldgspeed && m_route.get_nrwpt() > 0) {
				changemask |= change_fplan;
				m_route[m_navcurwpt - 1].save_time();
				m_route[m_navcurwpt - 1].set_time_unix(time(0));
				m_navlastwpttime = time(0);
				m_route.save_time_onblock();
				m_route.set_time_onblock_unix(m_navlastwpttime);
				m_navblock = m_position.simple_box_km(0.02);
				++m_navcurwpt;
				if (m_navflighttimerunning)
					m_navflighttime = time(0) - m_navflighttime;
				m_navflighttimerunning = false;
				if (m_track.is_valid()) {
					m_track.set_landingtime(m_navlastwpttime);
					m_track.set_onblocktime(m_navlastwpttime);
				}
			} else if ((m_navmode == navmode_fpl || m_navmode == navmode_fplhold ||
				    m_navmode == navmode_direct || m_navmode == navmode_directhold) && 
				   (abs(m_navcurtrackerr) >= (1 << 30))) {
				m_navcurtrack = m_navholdtrack;
				if (m_navmode == navmode_direct) {
					m_navmode = navmode_off;
				} else if ((m_navmode == navmode_fpl || m_navmode == navmode_fplhold) && m_navcurwpt < m_route.get_nrwpt()) {
					changemask |= change_fplan;
					m_navlastwpttime = time(0);
					m_route[m_navcurwpt - 1].save_time();
					m_route[m_navcurwpt - 1].set_time_unix(m_navlastwpttime);
					recompute_times(m_navcurwpt);
				}
			}
			if (m_navmode == navmode_finalapproach) {
				m_navcurdist = m_position.spheric_distance_nmi_dbl(m_navfas.get_ftp());
				double dist(m_position.spheric_distance_nmi_dbl(m_navfas.get_fpap()));
				m_navcurtrackerr = m_navcurtrack - (m_position.spheric_true_course(m_navfas.get_fpap()) * deg_to_track);
				m_navcurxtk = sin(m_navcurtrackerr * track_to_rad) * dist;
				dist *= fabs(cos(m_navcurtrackerr * track_to_rad));
				double maxxtk(dist * tan(m_navfas.get_coursedefl_rad()));
				m_navmaxxtk = std::min(std::max(maxxtk, 0.001), 1.);
				double baroalt, truealt;
				get_altitude(baroalt, truealt);
				dist = m_navcurdist * (1000.0 / Point::km_to_nmi_dbl * Point::m_to_ft_dbl);
				m_navgs = (m_navfas.get_glidepath_rad() - atan2(truealt - m_navfas.get_thralt(), dist)) / m_navfas.get_glidepathdefl_rad();
			} else {
				m_navcurdist = m_position.spheric_distance_nmi_dbl(m_navcur);
				m_navcurtrackerr = m_navcurtrack - (m_position.spheric_true_course(m_navcur) * deg_to_track);
				m_navcurxtk = sin(m_navcurtrackerr * track_to_rad) * m_navcurdist;
				m_navmaxxtk = 1.0;
				m_navgs = std::numeric_limits<double>::quiet_NaN();
			}
			if (std::isnan(m_groundspeed) || m_groundspeed < 1.0) {
				m_navroc = m_navtimetowpt = std::numeric_limits<double>::quiet_NaN();
			} else {
				m_navtimetowpt = m_navcurdist / m_groundspeed * 3600.0;
				double baroalt, truealt;
				get_altitude(baroalt, truealt);
				double altdiff(m_navcuralt - ((m_navcuraltflags & FPlanWaypoint::altflag_standard) ? baroalt : truealt));
				if (std::isnan(altdiff) || m_navtimetowpt < 10 || m_navcuralt == std::numeric_limits<int32_t>::min()) {
					m_navroc = std::numeric_limits<double>::quiet_NaN();
				} else {
					m_navroc = altdiff * 60.0 / m_navtimetowpt;
				}
			}
		}
		// Update Track
		if (m_positionok && m_engine && m_track.is_valid()) {
			bool doadd(!m_track.size());
			if (!doadd) {
				Glib::TimeVal tv(m_track[m_track.size()].get_timeval());
				tv -= m_positiontime;
				tv.add_milliseconds(1500);
				doadd = tv.negative();
			}
			if (doadd) {
				TracksDb::Track::TrackPoint tp;
				tp.set_timeval(m_positiontime);
				tp.set_coord(m_position);
				double alt(m_airdata.get_true_altitude());
				if (std::isnan(alt))
					tp.set_alt_invalid();
				else
					tp.set_alt((TracksDb::Track::TrackPoint::alt_t)alt);
				m_track.append(tp);
				if (m_track.get_unsaved_points() >= 60)
					m_engine->get_tracks_db().save_appendlog(m_track);
			}
		}
	}
	{
		Glib::TimeVal tv;
		tv.assign_current_time();
		std::swap(m_fueltime, tv);
		tv -= m_fueltime;
		m_fuel -= (1.0 / 3600.0) * m_fuelflow * tv.as_double();
	}
	if (changemask & (change_altitude | change_course)) {
		double da(m_airdata.get_density_altitude());
		double tas(0), fuelflow(0), bhp, rpm, mp;
		get_acft_cruise(rpm, mp, bhp);
	        m_aircraft.get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
		if (!std::isnan(m_groundspeed) && m_groundspeed > m_tkoffldgspeed) {
			if (m_altrate > 150) {
				const Aircraft::Climb& c(m_aircraft.get_climb());
				double t(c.altitude_to_time(da));
				tas = c.time_to_tas(t);
				fuelflow = c.time_to_fuelflow(t);
				bhp = std::numeric_limits<double>::quiet_NaN();
			}
			if (std::isnan(fuelflow))
				fuelflow = 0;
			m_fuelflow = fuelflow;
			if (m_iasmode != iasmode_manual && !std::isnan(tas)) {
				m_airdata.set_tas(tas);
				changemask |= change_airdata;
			}
			m_bhp = bhp;
		}
	}
	return changemask;
}
