//
// C++ Implementation: fplan
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include <libxml++/libxml++.h>
#include <glibmm.h>
#ifdef HAVE_GIOMM
#include <giomm.h>
#endif
#include <unistd.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <sqlite3x.hpp>
#include <sqlite3.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdexcept>
#include <stdarg.h>
#include <limits>
#include <cstring>
#include <glib.h>
//#include <glib/gstdio.h>

#include "fplan.h"
#include "baro.h"
#include "wmm.h"

FPlanLeg::FPlanLeg(void)
        : m_fromicao(), m_toicao(), m_fromname(), m_toname(), m_fromcoord(), m_tocoord(),
          m_fromalt(0), m_toalt(0), m_fromtruealt(0), m_totruealt(0), m_fromflags(0), m_toflags(0),
          m_fromfreq(0), m_tofreq(0), m_fromtime(0), m_legtime(0), m_winddir(0), m_windspeed(0),
          m_dist(0), m_vdist(0), m_truetrack(0), m_var(0), m_wca(0),
          m_legtype(legtype_invalid), m_future(false),
          m_crosstrack(0), m_disttodest(0), m_vgnd(0), m_legfraction(0),
          m_altdiff(0), m_referencealt(0), m_mapalt(0),
          m_mapcoord(), m_coursetodest(0), m_gpsvalid(false), m_gpsaltvalid(false)
{
}

FPlanLeg::FPlanLeg(const Glib::ustring& fromicao, const Glib::ustring & fromname, const Point & fromcoord,
                   int16_t fromalt, uint16_t fromflags, uint32_t fromtime, uint32_t fromfreq,
                   const Glib::ustring& toicao, const Glib::ustring & toname, const Point & tocoord,
                   int16_t toalt, uint16_t toflags, uint32_t totime, uint32_t tofreq,
		   float winddir, float windspeed, float qnh, float isaoffset,
                   legtype_t legtype, bool future)
        : m_fromicao(fromicao), m_toicao(toicao), m_fromname(fromname), m_toname(toname), m_fromcoord(fromcoord), m_tocoord(tocoord),
          m_fromalt(fromalt), m_toalt(toalt), m_fromtruealt(fromalt), m_totruealt(toalt), m_fromflags(fromflags), m_toflags(toflags),
          m_fromfreq(fromfreq), m_tofreq(tofreq), m_fromtime(fromtime), m_legtime(totime - fromtime), m_winddir(winddir), m_windspeed(windspeed),
          m_dist(0), m_vdist(0), m_truetrack(0), m_var(0), m_wca(0),
          m_legtype(legtype), m_future(future),
          m_crosstrack(0), m_disttodest(0), m_vgnd(0), m_legfraction(0),
          m_altdiff(0), m_referencealt(0), m_mapalt(0),
          m_mapcoord(), m_coursetodest(0), m_gpsvalid(false), m_gpsaltvalid(false)
{
        m_dist = m_fromcoord.spheric_distance_nmi(m_tocoord);
        m_truetrack = (uint16_t)(m_fromcoord.spheric_true_course(m_tocoord) * from_deg);
        WMM wmm;
        if (wmm.compute(get_fromaltitude() * (FPlanWaypoint::ft_to_m * 1e-3), get_fromcoord(), get_fromtime_unix()))
                m_var = (uint16_t)(wmm.get_dec() * from_deg);
	baro_correction(qnh, isaoffset);
}

Glib::ustring FPlanLeg::get_fromicaoname(void) const
{
        Glib::ustring r(m_fromicao);
        if (!m_fromicao.empty() && !m_fromname.empty())
                r += " ";
        r += m_fromname;
        return r;
}

Glib::ustring FPlanLeg::get_toicaoname(void) const
{
        Glib::ustring r(m_toicao);
        if (!m_toicao.empty() && !m_toname.empty())
                r += " ";
        r += m_toname;
        return r;
}

void FPlanLeg::baro_correction(float qnh, float isaoffset)
{
        m_fromtruealt = m_fromalt;
        m_totruealt = m_toalt;
        if (!((m_fromflags | m_toflags) & FPlanWaypoint::altflag_standard))
                return;
	IcaoAtmosphere atmo(qnh, isaoffset + IcaoAtmosphere::std_sealevel_temperature);
        float press, alt;
        if (m_fromflags & FPlanWaypoint::altflag_standard) {
                atmo.std_altitude_to_pressure(&press, 0, get_fromaltitude() * FPlanWaypoint::ft_to_m);
                atmo.pressure_to_altitude(&alt, 0, press);
                m_fromtruealt = (int16_t)(alt * FPlanWaypoint::m_to_ft);
        }
        if (m_toflags & FPlanWaypoint::altflag_standard) {
                atmo.std_altitude_to_pressure(&press, 0, get_toaltitude() * FPlanWaypoint::ft_to_m);
                atmo.pressure_to_altitude(&alt, 0, press);
                m_totruealt = (int16_t)(alt * FPlanWaypoint::m_to_ft);
        }
}

int16_t FPlanLeg::baro_correction(float qnh, float isaoffset, int16_t alt, uint16_t altflags)
{
        if (!(altflags & FPlanWaypoint::altflag_standard))
                return alt;
	IcaoAtmosphere atmo(qnh, isaoffset + IcaoAtmosphere::std_sealevel_temperature);
        float press, alt1;
        atmo.std_altitude_to_pressure(&press, 0, alt * FPlanWaypoint::ft_to_m);
        atmo.pressure_to_altitude(&alt1, 0, press);
        return Point::round<int16_t,float>(alt1 * FPlanWaypoint::m_to_ft);
}

void FPlanLeg::wind_correction(float vcruise_kts)
{
	wind_correction(vcruise_kts, m_winddir, m_windspeed);
}

/*
 * Winddreieck:
 * a = vdist
 * b = vwind
 * c = vcruise
 * alpha = ?
 * beta = WCA
 * gamma = 180 + TT - winddir
 *
 * Sinussatz: vwind/sin(WCA) = vcruise/sin(winddir-TT) = vdist/sin(winddir-TT-WCA)
 * Cosinussatz: vdist^2 = vwind^2 + vcruise^2 - 2*vwind*vcruise*cos(winddir-TT-WCA)
 */

void FPlanLeg::wind_correction(float vcruise_kts, float winddir_deg, float vwind_kts)
{
        uint16_t wdirminustt = (uint16_t)(winddir_deg * from_deg);
        wdirminustt -= m_truetrack;
        m_wca = (uint16_t)(from_rad * asin(sin(wdirminustt * to_rad) * vwind_kts / vcruise_kts));
        m_vdist = sqrt(vwind_kts * vwind_kts + vcruise_kts * vcruise_kts -
                       2.0 * vwind_kts * vcruise_kts * cos(to_rad * (wdirminustt - m_wca)));
        m_vgnd = m_vdist;
        switch (m_legtype) {
                case legtype_offblock_takeoff:
                case legtype_landing_onblock:
                        m_legtime = 5 * 60;
                        break;

                case legtype_normal:
                        m_legtime = (uint32_t)(3600 * m_dist / std::max(m_vdist, 0.1f));
                        break;

                default:
                        m_legtime = 0;
                        break;
        }
}

void FPlanLeg::update( time_t curtime, const Point & coord, bool coordvalid, int16_t alt, bool altvalid, float vgnd )
{
        m_crosstrack = m_disttodest = 0.0;
        m_legfraction = 0;
        m_altdiff = m_mapalt = 0;
        m_coursetodest = m_truetrack;
        m_gpsvalid = m_gpsaltvalid = false;
        CartesianVector3Df cfrom = PolarVector3Df(m_fromcoord, 0);
        CartesianVector3Df cto = PolarVector3Df(m_tocoord, 0);
        m_gpsvalid = coordvalid;
        m_gpsaltvalid = coordvalid && altvalid;
        float t = 0;
        if (m_gpsvalid) {
                m_mapcoord = coord;
                m_vgnd = vgnd;
                if (get_legtype() == legtype_normal) {
                        CartesianVector3Df cm = PolarVector3Df(m_mapcoord, 0);
                        t = cm.nearest_fraction(cfrom, cto);
                        CartesianVector3Df cproj = cfrom + (cto - cfrom) * t;
                        m_crosstrack = (cm - cproj).length();
                        m_coursetodest = (uint16_t)(m_mapcoord.spheric_true_course(m_tocoord) * from_deg);
                        if ((int16_t)(m_coursetodest - m_truetrack) < 0)
                                m_crosstrack = -m_crosstrack;
                }
        } else {
                if (!m_legtime) {
                        t = 0.5;
                } else {
                        t = curtime - get_fromtime_unix();
                        t = std::min(std::max(t / m_legtime, 0.0f), 1.0f);
                }
                m_mapcoord = ((PolarVector3Df)(cfrom + (cto - cfrom) * t)).getcoord();
                m_vgnd = m_vdist;
        }
        m_referencealt = m_fromtruealt + (int16_t)(t * (m_totruealt - m_fromtruealt));
        m_legfraction = (uint16_t)((1U << 15) * t);
        m_mapalt = m_gpsaltvalid ? alt : m_referencealt;
        m_altdiff = m_mapalt - m_referencealt;
        m_disttodest = m_mapcoord.spheric_distance_nmi(m_tocoord);
        WMM wmm;
        if (wmm.compute(get_mapalt() * (FPlanWaypoint::ft_to_m * 1e-3), get_mapcoord(), curtime))
                m_var = (uint16_t)(wmm.get_dec() * from_deg);
}

float FPlanLeg::get_verticalspeedtodest(void) const
{
        return (get_totruealtitude() - get_mapalt()) / std::max(1.0f / 60.0f, 60.0f * get_disttodest() / std::max(get_vgnd(), 1.0f));
}

float FPlanLeg::get_verticalspeed(void) const
{
        return (get_totruealtitude() - get_fromtruealtitude()) / std::max(1.0f / 60.0f, 60.0f * get_dist() / std::max(get_vgnd(), 1.0f));
}

const time_t FPlanWaypoint::unix_minus_palmos_epoch;
const uint16_t FPlanWaypoint::altflag_standard;
const uint16_t FPlanWaypoint::altflag_ifr;
const float FPlanWaypoint::ft_to_m;
const float FPlanWaypoint::m_to_ft;
const float FPlanWaypoint::to_rad;
const float FPlanWaypoint::to_deg;
const float FPlanWaypoint::from_rad;
const float FPlanWaypoint::from_deg;
const std::string FPlanWaypoint::reader_fields = "ICAO,NAME,LON,LAT,ALT,FREQ,TIME,SAVETIME,FLAGS,PATHNAME,PATHCODE,NOTE,WINDDIR,WINDSPEED,QFF,ISAOFFS";

FPlanWaypoint::FPlanWaypoint(void)
        : m_icao(), m_name(), m_pathname(), m_time(0), m_savetime(0), m_frequency(0), m_coord(), m_alt(0), m_flags(0), 
	  m_winddir(0), m_windspeed(0), m_qff(1013.25 * 32), m_isaoffset(0), m_pathcode(pathcode_none), m_dirty(false)
{
}

FPlanWaypoint::FPlanWaypoint(sqlite3x::sqlite3_cursor& cursor)
        : m_icao(cursor.getstring(0)), m_name(cursor.getstring(1)), m_pathname(cursor.getstring(9)), m_note(cursor.getstring(11)),
          m_time(cursor.getint64(6)), m_savetime(cursor.getint64(7)), m_frequency(cursor.getint(5)),
          m_coord(Point(cursor.getint(2), cursor.getint(3))), m_alt(cursor.getint(4)), m_flags(cursor.getint(8)),
	  m_winddir(cursor.getint(12)), m_windspeed(cursor.getint(13)), m_qff(cursor.getint(14)),
	  m_isaoffset(cursor.getint(15)), m_pathcode(static_cast<pathcode_t>(cursor.getint(10))), m_dirty(false)
{
}

Glib::ustring FPlanWaypoint::get_icao_name(void) const
{
        Glib::ustring r(m_icao);
        if (!m_icao.empty() && !m_name.empty())
                r += " ";
        r += m_name;
        return r;
}

const Glib::ustring& FPlanWaypoint::get_ident(void) const
{
	if (!m_icao.empty())
		return m_icao;
	return m_name;
}

const Glib::ustring& FPlanWaypoint::get_pathcode_name(pathcode_t pc)
{
	switch (pc) {
	case pathcode_none:
	{
		static const Glib::ustring r("None");
		return r;
	}

	case pathcode_sid:
	{
		static const Glib::ustring r("SID");
		return r;
	}

	case pathcode_star:
	{
		static const Glib::ustring r("STAR");
		return r;
	}

	case pathcode_approach:
	{
		static const Glib::ustring r("Approach");
		return r;
	}

	case pathcode_transition:
	{
		static const Glib::ustring r("Transition");
		return r;
	}

	case pathcode_vfrdeparture:
	{
		static const Glib::ustring r("VFR Departure");
		return r;
	}

	case pathcode_vfrarrival:
	{
		static const Glib::ustring r("VFR Arrival");
		return r;
	}

	case pathcode_vfrtransition:
	{
		static const Glib::ustring r("VFR Transition");
		return r;
	}

	case pathcode_airway:
	{
		static const Glib::ustring r("Airway");
		return r;
	}

	case pathcode_directto:
	{
		static const Glib::ustring r("Direct To");
		return r;
	}

	default:
	{
		static const Glib::ustring r("Invalid");
		return r;
	}

	}
}

bool FPlanWaypoint::enforce_pathcode_vfrifr(void)
{
	Glib::ustring pn(get_pathname());
	{
		Glib::ustring::size_type n(pn.size()), i(0);
		while (i < n && std::isspace(pn[i]))
			++i;
		while (n > i) {
			--n;
			if (std::isspace(pn[n]))
				continue;
			++n;
			break;
		}
		pn = pn.substr(i, n - i);
	}
	FPlanWaypoint::pathcode_t pc(get_pathcode());
	if (get_flags() & FPlanWaypoint::altflag_ifr) {
		pn = pn.uppercase();
		switch (pc) {
		case FPlanWaypoint::pathcode_none:
			pc = FPlanWaypoint::pathcode_airway;
			// fall through

		case FPlanWaypoint::pathcode_airway:
			if (pn.empty() || pn == "DCT")
				pc = FPlanWaypoint::pathcode_directto;
			break;

		default:
			pn.clear();
			pc = FPlanWaypoint::pathcode_directto;
			break;
		}
	} else {
		switch (pc) {
		case FPlanWaypoint::pathcode_vfrdeparture:
		case FPlanWaypoint::pathcode_vfrarrival:
		case FPlanWaypoint::pathcode_vfrtransition:
			break;

		default:
			pn.clear();
			pc = FPlanWaypoint::pathcode_none;
			break;
		}
	}
	bool work(false);
	if (get_pathcode() != pc) {
		set_pathcode(pc);
		work = true;
	}
	if (get_pathname() != pn) {
		set_pathname(pn);
		work = true;
	}
	return work;
}

char FPlanWaypoint::get_ruleschar(uint16_t flags)
{
	return (flags & altflag_ifr) ? 'I' : 'V';
}

char FPlanWaypoint::get_altchar(uint16_t flags)
{
	return (flags & altflag_standard) ? 'F' : 'A';
}

float FPlanWaypoint::get_sltemp_kelvin(void) const
{
	return IcaoAtmosphere::std_sealevel_temperature + get_isaoffset_kelvin();
}

void FPlanWaypoint::set_sltemp_kelvin(float temp)
{
	set_isaoffset_kelvin(temp - IcaoAtmosphere::std_sealevel_temperature);
}

float FPlanWaypoint::get_oat_kelvin(void) const
{
	IcaoAtmosphere atmo((get_flags() & altflag_standard) ? IcaoAtmosphere::std_sealevel_pressure : get_qff_hpa(), get_sltemp_kelvin());
	float oat(0);
	atmo.altitude_to_pressure(0, &oat, get_altitude() * FPlanWaypoint::ft_to_m);
	return oat;
}

void FPlanWaypoint::set_oat_kelvin(float temp)
{
	IcaoAtmosphere atmo((get_flags() & altflag_standard) ? IcaoAtmosphere::std_sealevel_pressure : get_qff_hpa());
	float oat(0);
	atmo.altitude_to_pressure(0, &oat, get_altitude() * FPlanWaypoint::ft_to_m);
	set_isaoffset_kelvin(temp - oat);
}

float FPlanWaypoint::get_pressure_altitude(void) const
{
	if (get_flags() & altflag_standard)
		return get_altitude();
	IcaoAtmosphere atmo(get_qff_hpa(), get_sltemp_kelvin());
        float p(0), temp(0), tempstd(0), alt(0);
        atmo.altitude_to_pressure(&p, &temp, get_altitude() * FPlanWaypoint::ft_to_m);
        atmo.std_pressure_to_altitude(&alt, &tempstd, p);
        return alt * FPlanWaypoint::m_to_ft;
}

float FPlanWaypoint::get_true_altitude(void) const
{
	if (!(get_flags() & altflag_standard))
		return get_altitude();
	IcaoAtmosphere atmo(get_qff_hpa(), get_sltemp_kelvin());
	float p(0), tempstd(0), temp(0), alt(0);
	atmo.std_altitude_to_pressure(&p, &tempstd, get_altitude() * FPlanWaypoint::ft_to_m);
        atmo.pressure_to_altitude(&alt, &temp, p);
        return alt * FPlanWaypoint::m_to_ft;
}

float FPlanWaypoint::get_density_altitude(void) const
{
        static const float c0(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::const_l0 * FPlanWaypoint::m_to_ft);
        static const float c1(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::std_sealevel_pressure);
        static const float c2((IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r) / (IcaoAtmosphere::const_g * IcaoAtmosphere::const_m / 1000.0 -
                                                                                      IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r));
	IcaoAtmosphere atmo((get_flags() & altflag_standard) ? IcaoAtmosphere::std_sealevel_pressure : get_qff_hpa());
	float p(0), temp(0);
	atmo.altitude_to_pressure(&p, &temp, get_altitude() * FPlanWaypoint::ft_to_m);
        return c0 * (1 - powf(p / temp * c1, c2));
}

FPlanRoute::FPlanRoute(FPlan & fpp)
        : m_fp(&fpp), m_id(0), m_note(), m_time_offblock(0), m_savetime_offblock(0), m_time_onblock(0), m_savetime_onblock(0), m_curwpt(0), m_wpts(), m_dirty(false)
{
}

FPlanWaypoint& FPlanRoute::operator[](uint32_t nr)
{
        if (nr >= m_wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        return m_wpts[nr];
}

const FPlanWaypoint& FPlanRoute::operator[](uint32_t nr) const
{
        if (nr >= m_wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        return m_wpts[nr];
}

void FPlanRoute::insert_wpt(uint32_t nr, const FPlanWaypoint & wpt)
{
	m_dirty = true;
        if (nr >= m_wpts.size()) {
                m_wpts.push_back(wpt);
                return;
        }
        m_wpts.insert(m_wpts.begin() + nr, wpt);
        if (m_id && m_fp) {
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR + 100000 WHERE PLANID = ? AND NR >= ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR - 99999 WHERE PLANID = ? AND NR >= ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "INSERT INTO waypoints (PLANID,NR) VALUES (?,?);");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                tran.commit();
        }
}

void FPlanRoute::delete_wpt(uint32_t nr)
{
        if (nr >= m_wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        if (m_id && m_fp) {
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ? AND NR = ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR + 99999 WHERE PLANID = ? AND NR > ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR - 100000 WHERE PLANID = ? AND NR > ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        m_wpts.erase(m_wpts.begin() + nr);
        m_dirty = true;
}

bool FPlanRoute::is_dirty(void) const
{
        if (m_dirty)
                return true;
        for (waypoints_t::const_iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; i++)
                if ((*i).m_dirty)
                        return true;
        return false;
}

void FPlanRoute::set_dirty(bool d)
{
        m_dirty = d;
        for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; i++)
                (*i).m_dirty = d;
}

void FPlanRoute::load_fp(id_t nid)
{
        m_id = nid;
	load_fp();
}

void FPlanRoute::load_fp(void)
{
        m_wpts.clear();
	if (!m_fp)
		throw std::runtime_error("load_fp: no database");
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT TIMEOFFBLOCK,SAVETIMEOFFBLOCK,TIMEONBLOCK,SAVETIMEONBLOCK,"
                                              "CURWPT,NOTE FROM fplan WHERE ID = ?;");
                cmd.bind(1, m_id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        throw std::runtime_error("load_fp: plan ID not found");
                m_time_offblock = cursor.getint64(0);
                m_savetime_offblock = cursor.getint64(1);
                m_time_onblock = cursor.getint64(2);
                m_savetime_onblock = cursor.getint64(3);
                m_curwpt = cursor.getint(4);
                m_note = cursor.getstring(5);

        }
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE PLANID = ? ORDER BY NR ASC;");
                cmd.bind(1, m_id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        m_wpts.push_back(FPlanWaypoint(cursor));
        }
        m_dirty = false;
}

void FPlanRoute::save_fp(void)
{
        if (!is_dirty())
                return;
	if (!m_fp)
		throw std::runtime_error("save_fp: no database");
        if (!m_id) {
                {
                        //sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT ID FROM fplan ORDER BY ID DESC LIMIT 1;");
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT MAX(ID) FROM fplan;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                        else
                                m_id = 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "INSERT INTO fplan (ID) VALUES(?);");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
        }
        sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ? AND (NR < 0 OR NR >= ?);");
                cmd.bind(1, m_id);
                cmd.bind(2, get_nrwpt());
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE fplan SET TIMEOFFBLOCK = ?,"
                                              "SAVETIMEOFFBLOCK = ?, TIMEONBLOCK = ?, SAVETIMEONBLOCK = ?,"
                                              "CURWPT = ?, NOTE = ? WHERE ID = ?;");
                cmd.bind(1, (long long int)m_time_offblock);
                cmd.bind(2, (long long int)m_savetime_offblock);
                cmd.bind(3, (long long int)m_time_onblock);
                cmd.bind(4, (long long int)m_savetime_onblock);
                cmd.bind(5, (int)m_curwpt);
                cmd.bind(6, m_note);
                cmd.bind(7, m_id);
                cmd.executenonquery();
        }
        for (unsigned int nr = 0; nr < get_nrwpt(); nr++) {
                FPlanWaypoint& w(m_wpts[nr]);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "INSERT OR IGNORE INTO waypoints (PLANID,NR) VALUES(?,?);");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET ICAO  = ?, NAME = ?, "
                                                      "LON = ?, LAT = ?, ALT = ?, FREQ = ?, TIME = ?, SAVETIME = ?, "
                                                      "FLAGS = ?, PATHNAME = ?, PATHCODE = ?, NOTE = ?, "
						      "WINDDIR = ?, WINDSPEED = ?, QFF = ?, ISAOFFS = ? "
						      "WHERE PLANID = ? AND NR = ?;");
                        cmd.bind(1, w.get_icao());
                        cmd.bind(2, w.get_name());
                        cmd.bind(3, w.get_lon());
                        cmd.bind(4, w.get_lat());
                        cmd.bind(5, w.get_altitude());
                        cmd.bind(6, (int)w.get_frequency());
                        cmd.bind(7, (long long int)w.get_time());
                        cmd.bind(8, (long long int)w.get_save_time());
                        cmd.bind(9, w.get_flags());
                        cmd.bind(10, w.get_pathname());
                        cmd.bind(11, w.get_pathcode());
                        cmd.bind(12, w.get_note());
			cmd.bind(13, w.get_winddir());
			cmd.bind(14, w.get_windspeed());
			cmd.bind(15, w.get_qff());
			cmd.bind(16, w.get_isaoffset());
                        cmd.bind(17, m_id);
                        cmd.bind(18, (int)nr);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        set_dirty(false);
}

void FPlanRoute::duplicate_fp(void)
{
        m_id = 0;
        m_dirty = true;
}

void FPlanRoute::new_fp(void)
{
        duplicate_fp();
        m_wpts.clear();
        m_note.clear();
        m_time_offblock = 0;
        m_savetime_offblock = 0;
        m_time_onblock = 0;
        m_savetime_onblock = 0;
        m_curwpt = 0;
        set_dirty(false);
}

void FPlanRoute::delete_fp(void)
{
        if (m_id) {
		if (!m_fp)
			throw std::runtime_error("delete_fp: no database");
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ?;");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM fplan WHERE ID = ?;");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        new_fp();
}

void FPlanRoute::reverse_fp(void)
{
        if (m_wpts.empty())
                return;
        reverse(m_wpts.begin(), m_wpts.end());
        set_dirty(true);
}

bool FPlanRoute::load_first_fp(void)
{
	if (!m_fp)
		throw std::runtime_error("load_first_fp: no database");
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT ID FROM fplan ORDER BY ID ASC LIMIT 1;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        return false;
		m_id = cursor.getint(0);
        }
        load_fp();
        return true;
}

bool FPlanRoute::load_next_fp(void)
{
	if (!m_fp)
		throw std::runtime_error("load_next_fp: no database");
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT ID FROM fplan WHERE ID > ? ORDER BY ID ASC LIMIT 1;");
                cmd.bind(1, m_id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        return false;
                m_id = cursor.getint(0);
        }
        load_fp();
        return true;
}

float FPlanRoute::total_distance_km(void) const
{
        float sum(0);
        for (unsigned int i = m_wpts.size() - 1; i > 0; --i)
               sum += m_wpts[i-1].get_coord().spheric_distance_km(m_wpts[i].get_coord());
        return sum;
}

double FPlanRoute::total_distance_km_dbl(void) const
{
        double sum(0);
        for (unsigned int i = m_wpts.size() - 1; i > 0; --i)
               sum += m_wpts[i-1].get_coord().spheric_distance_km_dbl(m_wpts[i].get_coord());
        return sum;
}

int16_t FPlanRoute::max_altitude(void) const
{
        int16_t a(std::numeric_limits<int16_t>::min());
        for (unsigned int i = m_wpts.size(); i > 0; )
                a = std::max(a, m_wpts[--i].get_altitude());
        return a;
}

int16_t FPlanRoute::max_altitude(uint16_t& flags) const
{
        int16_t a(std::numeric_limits<int16_t>::min());
	uint16_t f(0);
        for (unsigned int i = m_wpts.size(); i > 0; ) {
		const FPlanWaypoint& wpt(m_wpts[--i]);
		int16_t aw(wpt.get_altitude());
		if (aw < a)
			continue;
                a = aw;
		f = wpt.get_flags();
	}
	flags = f;
        return a;
}

Rect FPlanRoute::get_bbox(void) const
{
	if (!m_wpts.size())
                return Rect();
        Point pt(m_wpts[0].get_coord());
        Point sw(0, 0), ne(0, 0);
        for (unsigned int i = 1; i < m_wpts.size(); ++i) {
                Point p(m_wpts[i].get_coord() - pt);
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        return Rect(sw + pt, ne + pt);
}

char FPlanRoute::get_flightrules(void) const
{
	unsigned int i(m_wpts.size());
	if (!i)
		return ' ';
	uint16_t vfr(0), ifr(0);
	while (i > 0) {
		uint16_t f(m_wpts[--i].get_flags());
		vfr |= ~f;
		ifr |= f;
		if (vfr & ifr & FPlanWaypoint::altflag_ifr)
			return (m_wpts[0].get_flags() & FPlanWaypoint::altflag_ifr) ? 'Y' : 'Z';
	}
	return (ifr & FPlanWaypoint::altflag_ifr) ? 'I' : 'V';
}

void FPlanRoute::save_time(uint32_t nr)
{
        if (!nr) {
                save_time_offblock();
                return;
        }
        if (nr > m_wpts.size()) {
                save_time_onblock();
                return;
        }
        m_wpts[nr-1].save_time();
}

void FPlanRoute::restore_time(uint32_t nr)
{
        if (!nr) {
                restore_time_offblock();
                return;
        }
        if (nr > m_wpts.size()) {
                restore_time_onblock();
                return;
        }
        m_wpts[nr-1].restore_time();
}

void FPlanRoute::set_time(uint32_t nr, uint32_t t)
{
        if (!nr) {
                set_time_offblock(t);
                return;
        }
        if (nr > m_wpts.size()) {
                set_time_onblock(t);
                return;
        }
        m_wpts[nr-1].set_time(t);
}

uint32_t FPlanRoute::get_time(uint32_t nr) const
{
        if (!nr)
                return get_time_offblock();
        if (nr > m_wpts.size())
                return get_time_onblock();
        return m_wpts[nr-1].get_time();
}

FPlanLeg FPlanRoute::get_leg(uint32_t nr)
{
        bool future = nr > get_curwpt();
        const FPlanWaypoint *wpt0(0), *wpt1(0);
        if (nr > 0 && nr <= get_nrwpt())
                wpt0 = &(*this)[nr - 1];
        if (nr < get_nrwpt())
                wpt1 = &(*this)[nr];
        if (!wpt0 && !wpt1) {
                if (nr > 0 && get_nrwpt()) {
                        wpt0 = &(*this)[get_nrwpt()-1];
                        return FPlanLeg("", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                        "", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
				        wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_invalid, future);
                }
                return FPlanLeg("", _("Off Block"), Point(), 0, 0, 0, 0,
                                "", _("On Block"), Point(), 0, 0, 0, 0,
                                0, 0, 0, 0, FPlanLeg::legtype_invalid, future);
        }
        if (!wpt0)
                return FPlanLeg("", _("Off Block"), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_offblock(), wpt1->get_frequency(),
                                wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), wpt1->get_time(), wpt1->get_frequency(),
				wpt1->get_winddir_deg(), wpt1->get_windspeed_kts(), wpt1->get_qff_hpa(), wpt1->get_isaoffset_kelvin(), FPlanLeg::legtype_offblock_takeoff, future);
        if (!wpt1)
                return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                "", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_landing_onblock, future);
        return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                        wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_onblock(), wpt1->get_frequency(),
			wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_normal, future);
}

FPlanLeg FPlanRoute::get_directleg(void)
{
        uint32_t nr(get_nrwpt());
        bool future = false;
        if (!nr)
                return FPlanLeg("", _("Off Block"), Point(), 0, 0, 0, 0,
                                "", _("On Block"), Point(), 0, 0, 0, 0,
                                0, 0, 0, 0, FPlanLeg::legtype_invalid, future);
        const FPlanWaypoint *wpt0(&(*this)[0]), *wpt1(&(*this)[nr - 1]);
        return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                        wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_onblock(), wpt1->get_frequency(),
                        wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_normal, future);
}

bool FPlanRoute::has_pathcodes(void) const
{
        for (unsigned int i = 0; i < m_wpts.size(); ++i)
		if (m_wpts[i].get_pathcode() != FPlanWaypoint::pathcode_none)
			return true;
	return false;
}

bool FPlanRoute::enforce_pathcode_vfrdeparr(void)
{
	unsigned int nr(m_wpts.size());
	if (nr <= 2U)
		return false;
	bool work(false);
	unsigned int i(0);
	for (; i < nr - 1; ++i) {
		const FPlanWaypoint& wpt(m_wpts[i]);
		if (wpt.get_icao() != m_wpts[0].get_icao() ||
		    (wpt.get_flags() & FPlanWaypoint::altflag_ifr) ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrarrival ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrtransition)
			break;
	}
	if (i > 1) {
		work = true;
		while (i > 0) {
			--i;
			m_wpts[i].set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
		}
	}
	i = nr - 1;
	for (; i > 0; --i) {
		const FPlanWaypoint& wpt(m_wpts[i]);
		if (wpt.get_icao() != m_wpts[nr - 1].get_icao() ||
		    (wpt.get_flags() & FPlanWaypoint::altflag_ifr) ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrdeparture ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrtransition)
			break;
	}
	if (i < nr - 2) {
		work = true;
		while (i < nr - 1) {
			++i;
			m_wpts[i].set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
		}
	}
	return work;
}

bool FPlanRoute::enforce_pathcode_vfrifr(void)
{
	unsigned int nr(m_wpts.size());
	bool work(false);
	for (unsigned int i = 0; i < nr; ++i) {
		FPlanWaypoint& wpt(m_wpts[i]);
		work = wpt.enforce_pathcode_vfrifr() || work;
	}
	return work;
}

Glib::ustring FPlan::get_userdbpath(void)
{
#ifdef __MINGW32__
	Glib::ustring dbname(Glib::build_filename(Glib::get_user_config_dir(), "vfrnav"));
#else
        Glib::ustring dbname(Glib::build_filename(Glib::get_home_dir(), USERDBPATH));
#endif
#if defined(HAVE_GIOMM) && 0
        {
                Glib::RefPtr<Gio::File> dir(Gio::File::create_for_path(dbname));
                try {
#ifdef HAVE_GIOMM_MAKE_DIRECTORY_WITH_PARENTS
                        if (dir->make_directory_with_parents())
#else
                        if (dir->make_directory())
#endif
                                std::cerr << "vfrnav directory (" << dbname << ") created" << std::endl;
                        else
                                throw std::runtime_error("vfrnav directory creation failed");
                } catch (const Gio::Error& e) {
                        if (e.code() != Gio::Error::EXISTS)
                                throw;
                        std::cerr << "vfrnav directory (" << dbname << ") already exists" << std::endl;
                }
        }
#else
        if (g_mkdir_with_parents(dbname.c_str(), 0755)) {
                throw std::runtime_error("vfrnav directory (" + dbname +") creation failed");
        }
#endif
        return dbname;
}

FPlan::FPlan(void)
{
        Glib::ustring dbname(Glib::build_filename(get_userdbpath(), "fplan.db"));
        db.open(dbname);
        if (sqlite3_create_function(db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("FPlanDb: cannot add user function simpledist");
        init_database();
}

FPlan::FPlan(sqlite3 *dbh)
{
        db.take(dbh);
        if (sqlite3_create_function(db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("FPlanDb: cannot add user function simpledist");
        init_database();
}

FPlan::~FPlan() throw()
{
}

void FPlan::dbfunc_simpledist(sqlite3_context *ctxt, int, sqlite3_value **values)
{
        sqlite3_result_double(ctxt, Point(sqlite3_value_int(values[0]), sqlite3_value_int(values[1]))
                        .spheric_distance_km(Point(sqlite3_value_int(values[2]), sqlite3_value_int(values[3]))));
}

void FPlan::init_database(void)
{
        {
                sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS fplan (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "TIMEOFFBLOCK INTEGER,"
                                              "SAVETIMEOFFBLOCK INTEGER,"
                                              "TIMEONBLOCK INTEGER,"
                                              "SAVETIMEONBLOCK INTEGER,"
                                              "CURWPT INTEGER,"
                                              "NOTE TEXT);");
                cmd.executenonquery();
        }
	{
		static const struct field {
			const char *fldname;
			const char *fldtype;
		} fields[] = {
			{ "ICAO", "TEXT" },
			{ "NAME", "TEXT" },
			{ "LON", "INTEGER" },
			{ "LAT", "INTEGER" },
			{ "ALT", "INTEGER" },
			{ "FREQ", "INTEGER" },
			{ "TIME", "INTEGER" },
			{ "SAVETIME", "INTEGER" },
			{ "FLAGS", "INTEGER" },
			{ "PATHNAME", "TEXT" },
			{ "PATHCODE", "INTEGER" },
			{ "NOTE", "TEXT" },
			{ "WINDDIR", "INTEGER" },
			{ "WINDSPEED", "INTEGER" },
			{ "QFF", "INTEGER" },
			{ "ISAOFFS", "INTEGER" },
			{ 0, 0 }
		};
		{
			std::string flds;
			for (const struct field *f(fields); f->fldname && f->fldtype; ++f)
				flds += std::string(f->fldname) + " " + f->fldtype + ",";
			sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS waypoints (PLANID INTEGER NOT NULL "
						      "CONSTRAINT fk_planid REFERENCES fplan(ID) ON DELETE CASCADE,"
						      "NR INTEGER NOT NULL," + flds +
						      "PRIMARY KEY(PLANID,NR));");
			cmd.executenonquery();
		}
		// Auto Add new columns if they do not yet exist
		{
			sqlite3x::sqlite3_command cmd(db, "SELECT sql FROM sqlite_master WHERE type = \"table\" AND name = \"waypoints\";");
			sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
			if (!cursor.step())
				throw std::runtime_error("FPlan::init_database: unable to determine schema");
			std::string sql(cursor.getstring(0));
			unsigned int fldmask(0);
			for (std::string::const_iterator si(sql.begin()), se(sql.end()); si != se;) {
				if (!std::isalnum(*si)) {
					++si;
					continue;
				}
				std::string::const_iterator si2(si);
				++si2;
				while (si2 != se && std::isalnum(*si2))
					++si2;
				Glib::ustring tok(si, si2);
				si = si2;
				tok = tok.casefold();
				for (unsigned int i = 0; fields[i].fldname && fields[i].fldtype; ++i) {
					if (tok != Glib::ustring(fields[i].fldname).casefold())
						continue;
					fldmask |= 1U << i;
					break;
				}
			}
			for (unsigned int i = 0; fields[i].fldname && fields[i].fldtype; ++i) {
				if (fldmask & (1U << i))
					continue;
				sqlite3x::sqlite3_command cmd(db, std::string("ALTER TABLE waypoints ADD COLUMN ") + fields[i].fldname +
							      " " + fields[i].fldtype + ";");
				cmd.executenonquery();
			}
		}
	}
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptplanindex ON waypoints (PLANID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptplanidnr ON waypoints (PLANID, NR);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptlon ON waypoints (LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptlat ON waypoints (LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptname ON waypoints (NAME);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wpticao ON waypoints (ICAO);"); cmd.executenonquery(); }
}

void FPlan::delete_database(void)
{
        { sqlite3x::sqlite3_command cmd(db, "DROP TABLE IF EXISTS fplan;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP TABLE IF EXISTS waypoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptplanindex;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptplanidnr;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptlat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptname;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wpticao;"); cmd.executenonquery(); }
        init_database();
}

std::string FPlan::escape_string(const std::string& s)
{
        std::string r;
        r.reserve(s.size());
        for (std::string::const_iterator i1(s.begin()), ie(s.end()); i1 != ie; i1++) {
                std::string::value_type c(*i1);
                if (c == '_' || c == '%')
                        r.push_back('%');
                r.push_back(c);
        }
        return r;
}

std::vector<FPlanWaypoint> FPlan::find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, comp_t comp)
{
        std::vector<FPlanWaypoint> wpts;
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE NAME LIKE ? ORDER BY NAME ASC LIMIT ? OFFSET ?;");
                Glib::ustring pattern(escape_string(nm));
                switch (comp) {
                        case comp_contains:
                                pattern = '%' + pattern;
                                /* fall through */

                        case comp_startswith:
                                pattern += '%';
                                break;

                        case comp_exact:
                        default:
                                break;
                }
                cmd.bind(1, pattern);
                cmd.bind(2, (int)limit);
                cmd.bind(3, (int)skip);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}

std::vector<FPlanWaypoint> FPlan::find_by_icao(const Glib::ustring & ic, unsigned int limit, unsigned int skip, comp_t comp)
{
        std::vector<FPlanWaypoint> wpts;
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE ICAO LIKE ? ORDER BY ICAO ASC LIMIT ? OFFSET ?;");
                Glib::ustring pattern(escape_string(ic));
                switch (comp) {
                        case comp_contains:
                                pattern = '%' + pattern;
                                /* fall through */

                        case comp_startswith:
                                pattern += '%';
                                break;

                        case comp_exact:
                        default:
                                break;
                }
                cmd.bind(1, pattern);
                cmd.bind(2, (int)limit);
                cmd.bind(3, (int)skip);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}

std::vector<FPlanWaypoint> FPlan::find_nearest(const Point& pt, unsigned int limit, unsigned int skip, const Rect& rect)
{
        std::vector<FPlanWaypoint> wpts;
        {
                std::string rectexpr("LAT >= ?6 AND LAT <= ?8 AND LON >= ?5 AND LON <= ?7");
                if (rect.get_northeast().get_lon() < rect.get_southwest().get_lon())
                        std::string rectexpr("LAT >= ?6 AND LAT <= ?8 AND (LON <= ?5 OR LON >= ?7)");
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE " + rectexpr + " ORDER BY simpledist(LON,LAT,?1,?2) ASC LIMIT ?3 OFFSET ?4;");
                cmd.bind(1, pt.get_lon());
                cmd.bind(2, pt.get_lat());
                cmd.bind(3, (int)limit);
                cmd.bind(4, (int)skip);
                cmd.bind(5, rect.get_southwest().get_lon());
                cmd.bind(6, rect.get_southwest().get_lat());
                cmd.bind(7, rect.get_northeast().get_lon());
                cmd.bind(8, rect.get_northeast().get_lat());
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}
