/***************************************************************************
 *   Copyright (C) 2012, 2013 by Thomas Sailer                             *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"

#include <limits>
#include <cstdio>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <set>
#include <time.h>
#include "icaofpl.h"

class IcaoFlightPlan::FindCoord::SortAirportNamelen {
public:
	bool operator()(const AirportsDb::element_t& a, const AirportsDb::element_t& b) const { return a.get_name().size() < b.get_name().size(); }
};

class IcaoFlightPlan::FindCoord::SortNavaidNamelen {
public:
	bool operator()(const NavaidsDb::element_t& a, const NavaidsDb::element_t& b) const { return a.get_name().size() < b.get_name().size(); }
};

class IcaoFlightPlan::FindCoord::SortMapelementNamelen {
public:
	bool operator()(const MapelementsDb::element_t& a, const MapelementsDb::element_t& b) const { return a.get_name().size() < b.get_name().size(); }
};

class IcaoFlightPlan::FindCoord::SortAirportDist {
public:
	SortAirportDist(const Point& pt) : m_point(pt) {}
	bool operator()(const AirportsDb::element_t& a, const AirportsDb::element_t& b) const { return m_point.simple_distance_rel(a.get_coord()) < m_point.simple_distance_rel(b.get_coord()); }

protected:
	Point m_point;
};

class IcaoFlightPlan::FindCoord::SortNavaidDist {
public:
	SortNavaidDist(const Point& pt) : m_point(pt) {}
	bool operator()(const NavaidsDb::element_t& a, const NavaidsDb::element_t& b) const { return m_point.simple_distance_rel(a.get_coord()) < m_point.simple_distance_rel(b.get_coord()); }

protected:
	Point m_point;
};

class IcaoFlightPlan::FindCoord::SortWaypointDist {
public:
	SortWaypointDist(const Point& pt) : m_point(pt) {}
	bool operator()(const WaypointsDb::element_t& a, const WaypointsDb::element_t& b) const { return m_point.simple_distance_rel(a.get_coord()) < m_point.simple_distance_rel(b.get_coord()); }

protected:
	Point m_point;
};

class IcaoFlightPlan::FindCoord::SortMapelementDist {
public:
	SortMapelementDist(const Point& pt) : m_point(pt) {}
	bool operator()(const MapelementsDb::element_t& a, const MapelementsDb::element_t& b) const { return m_point.simple_distance_rel(a.get_coord()) < m_point.simple_distance_rel(b.get_coord()); }

protected:
	Point m_point;
};


const unsigned int IcaoFlightPlan::FindCoord::flag_airport;
const unsigned int IcaoFlightPlan::FindCoord::flag_navaid;
const unsigned int IcaoFlightPlan::FindCoord::flag_waypoint;
const unsigned int IcaoFlightPlan::FindCoord::flag_mapelement;
const unsigned int IcaoFlightPlan::FindCoord::flag_subtables;
const std::string IcaoFlightPlan::FindCoord::empty;

IcaoFlightPlan::FindCoord::FindCoord(Engine& engine)
	: m_engine(engine), m_coordresult(Point::invalid)
{
}

void IcaoFlightPlan::FindCoord::async_cancel(void)
{
        Engine::AirportResult::cancel(m_airportquery);
        Engine::AirportResult::cancel(m_airportquery1);
        Engine::NavaidResult::cancel(m_navaidquery);
        Engine::NavaidResult::cancel(m_navaidquery1);
        Engine::WaypointResult::cancel(m_waypointquery);
        Engine::MapelementResult::cancel(m_mapelementquery);
	Engine::AirwayGraphResult::cancel(m_airwayquery);
	Engine::AreaGraphResult::cancel(m_areaquery);
}

void IcaoFlightPlan::FindCoord::async_clear(void)
{
	m_airportquery = Glib::RefPtr<Engine::AirportResult>();
	m_airportquery1 = Glib::RefPtr<Engine::AirportResult>();
	m_navaidquery = Glib::RefPtr<Engine::NavaidResult>();
	m_navaidquery1 = Glib::RefPtr<Engine::NavaidResult>();
	m_waypointquery = Glib::RefPtr<Engine::WaypointResult>();
	m_mapelementquery = Glib::RefPtr<Engine::MapelementResult>();
	m_airwayquery = Glib::RefPtr<Engine::AirwayGraphResult>();
	m_areaquery = Glib::RefPtr<Engine::AreaGraphResult>();
	m_airports.clear();
	m_navaids.clear();
	m_waypoints.clear();
	m_mapelements.clear();
	m_airwaygraph.clear();
}

void IcaoFlightPlan::FindCoord::async_connectdone(void)
{
	if (m_airportquery)
		m_airportquery->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_airportquery1)
		m_airportquery1->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_navaidquery)
		m_navaidquery->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_navaidquery1)
		m_navaidquery1->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_waypointquery)
		m_waypointquery->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_mapelementquery)
		m_mapelementquery->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_airwayquery)
		m_airwayquery->connect(sigc::mem_fun(*this, &FindCoord::async_done));
	if (m_areaquery)
		m_areaquery->connect(sigc::mem_fun(*this, &FindCoord::async_done));
}		

void IcaoFlightPlan::FindCoord::async_done(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_cond.broadcast();
}

unsigned int IcaoFlightPlan::FindCoord::async_waitdone(void)
{
	unsigned int flags(0);
        Glib::Mutex::Lock lock(m_mutex);
        for (;;) {
		if ((m_airportquery || m_airportquery1) && (!m_airportquery || m_airportquery->is_done()) && (!m_airportquery1 || m_airportquery1->is_done())) {
			m_airports.clear();
			if (m_airportquery && !m_airportquery->is_error()) {
				m_airports = m_airportquery->get_result();
			} else if (m_airportquery1 && !m_airportquery1->is_error()) {
				m_airports = m_airportquery1->get_result();
				flags |= flag_airport;
			}
			m_airportquery = Glib::RefPtr<Engine::AirportResult>();
			m_airportquery1 = Glib::RefPtr<Engine::AirportResult>();
		}
		if ((m_navaidquery || m_navaidquery1) && (!m_navaidquery || m_navaidquery->is_done()) && (!m_navaidquery1 || m_navaidquery1->is_done())) {
			m_navaids.clear();
			if (m_navaidquery && !m_navaidquery->is_error()) {
				m_navaids = m_navaidquery->get_result();
			} else if (m_navaidquery1 && !m_navaidquery1->is_error()) {
				m_navaids = m_navaidquery1->get_result();
				flags |= flag_navaid;
			}
			m_navaidquery = Glib::RefPtr<Engine::NavaidResult>();
			m_navaidquery1 = Glib::RefPtr<Engine::NavaidResult>();
		}
		if (m_waypointquery && m_waypointquery->is_done()) {
			m_waypoints.clear();
			if (!m_waypointquery->is_error())
				m_waypoints = m_waypointquery->get_result();
			m_waypointquery = Glib::RefPtr<Engine::WaypointResult>();
		}
		if (m_mapelementquery && m_mapelementquery->is_done()) {
			m_mapelements.clear();
			if (!m_mapelementquery->is_error())
				m_mapelements = m_mapelementquery->get_result();
			m_mapelementquery = Glib::RefPtr<Engine::MapelementResult>();
		}
		if (m_airwayquery && m_airwayquery->is_done()) {
			m_airwaygraph.clear();
			if (!m_airwayquery->is_error())
				m_airwaygraph = m_airwayquery->get_result();
			m_airwayquery = Glib::RefPtr<Engine::AirwayGraphResult>();
		}
		if (m_areaquery && m_areaquery->is_done()) {
			m_airwaygraph.clear();
			if (!m_areaquery->is_error())
				m_airwaygraph = m_areaquery->get_result();
			m_areaquery = Glib::RefPtr<Engine::AreaGraphResult>();
		}
		if (!(m_airportquery || m_airportquery1 || m_navaidquery || m_navaidquery1 || m_waypointquery || m_mapelementquery || m_airwayquery || m_areaquery))
			break;
		m_cond.wait(m_mutex);
	}
	return flags;
}

void IcaoFlightPlan::FindCoord::sort_dist(const Point& pt)
{
	{
		SortAirportDist s(pt);
		sort(m_airports.begin(), m_airports.end(), s);
		if (!m_airports.empty())
			m_airports.resize(1);
	}
	{
		SortNavaidDist s(pt);
		sort(m_navaids.begin(), m_navaids.end(), s);
		if (!m_navaids.empty())
			m_navaids.resize(1);
	}
	{
		SortWaypointDist s(pt);
		sort(m_waypoints.begin(), m_waypoints.end(), s);
		if (!m_waypoints.empty())
			m_waypoints.resize(1);
	}
	{
		SortMapelementDist s(pt);
		sort(m_mapelements.begin(), m_mapelements.end(), s);
		if (!m_mapelements.empty())
			m_mapelements.resize(1);
	}
}

void IcaoFlightPlan::FindCoord::cut_dist(const Point& pt, float maxdist_km)
{
	if (!m_airports.empty() && pt.spheric_distance_km(m_airports.front().get_coord()) > maxdist_km)
		m_airports.clear();
	if (!m_navaids.empty() && pt.spheric_distance_km(m_navaids.front().get_coord()) > maxdist_km)
		m_navaids.clear();
	if (!m_waypoints.empty() && pt.spheric_distance_km(m_waypoints.front().get_coord()) > maxdist_km)
		m_waypoints.clear();
	if (!m_mapelements.empty() && pt.spheric_distance_km(m_mapelements.front().get_coord()) > maxdist_km)
		m_mapelements.clear();
	if (!m_coordresult.is_invalid() && pt.spheric_distance_km(m_coordresult) > maxdist_km)
		m_coordresult = Point::invalid;
}

void IcaoFlightPlan::FindCoord::retain_shortest_dist(const Point& pt)
{
	unsigned int what(0);
	uint64_t dist(std::numeric_limits<uint64_t>::max());
	if (!m_coordresult.is_invalid()) {
		uint64_t d(pt.simple_distance_rel(m_coordresult));
		if (d <= dist) {
			dist = d;
			what = 4;
		}
	}
	if (!m_mapelements.empty()) {
		uint64_t d(pt.simple_distance_rel(m_mapelements.front().get_coord()));
		if (d <= dist) {
			dist = d;
			what = 3;
		}
	}
	if (!m_waypoints.empty()) {
		uint64_t d(pt.simple_distance_rel(m_waypoints.front().get_coord()));
		if (d <= dist) {
			dist = d;
			what = 2;
		}
	}
	if (!m_navaids.empty()) {
		uint64_t d(pt.simple_distance_rel(m_navaids.front().get_coord()));
		if (d <= dist) {
			dist = d;
			what = 1;
		}
	}
	if (!m_airports.empty()) {
		uint64_t d(pt.simple_distance_rel(m_airports.front().get_coord()));
		if (d <= dist) {
			dist = d;
			what = 0;
		}
	}
	if (what != 0)
		m_airports.clear();
	if (what != 1)
		m_navaids.clear();
	if (what != 2)
		m_waypoints.clear();
	if (what != 3)
		m_mapelements.clear();
	if (what != 4)
		m_coordresult = Point::invalid;
}

bool IcaoFlightPlan::FindCoord::find(const std::string& icao, const std::string& name, unsigned int flags, const Point& pt)
{
	m_coordresult = Point::invalid;
	async_cancel();
	async_clear();
	if (!icao.empty()) {
		if (flags & flag_airport)
			m_airportquery = m_engine.async_airport_find_by_icao(icao, ~0, 0, AirportsDb::comp_exact,
									     (flags & flag_subtables) ? AirportsDb::element_t::subtables_all : AirportsDb::element_t::subtables_vfrroutes);
		if (flags & flag_navaid)
			m_navaidquery = m_engine.async_navaid_find_by_icao(icao, ~0, 0, NavaidsDb::comp_exact,
									   (flags & flag_subtables) ? NavaidsDb::element_t::subtables_all : NavaidsDb::element_t::subtables_none);
	}
	if (!name.empty()) {
		if (flags & flag_airport)
			m_airportquery1 = m_engine.async_airport_find_by_name(name, ~0, 0, AirportsDb::comp_contains,
									      (flags & flag_subtables) ? AirportsDb::element_t::subtables_all : AirportsDb::element_t::subtables_vfrroutes);
		if (flags & flag_navaid)
			m_navaidquery1 = m_engine.async_navaid_find_by_name(name, ~0, 0, NavaidsDb::comp_contains,
									    (flags & flag_subtables) ? NavaidsDb::element_t::subtables_all : NavaidsDb::element_t::subtables_none);
		if (flags & flag_waypoint)
			m_waypointquery = m_engine.async_waypoint_find_by_name(name, ~0, 0, WaypointsDb::comp_exact,
									       (flags & flag_subtables) ? WaypointsDb::element_t::subtables_all : WaypointsDb::element_t::subtables_none);
		if (flags & flag_mapelement)
			m_mapelementquery = m_engine.async_mapelement_find_by_name(name, ~0, 0, MapelementsDb::comp_contains,
										   (flags & flag_subtables) ? MapelementsDb::element_t::subtables_all : MapelementsDb::element_t::subtables_none);
		if (flags & flag_coord) {
			unsigned int r(m_coordresult.set_str(name));
			if ((Point::setstr_lat | Point::setstr_lon) & ~r)
				m_coordresult = Point::invalid;
		}
	}
	async_connectdone();
	{
		unsigned int flags(async_waitdone());
		// discard worse matching airports
		if (flags & flag_airport) {
			std::sort(m_airports.begin(), m_airports.end(), SortAirportNamelen());
			std::string::size_type namelen(0);
			if (!m_airports.empty())
				namelen = m_airports.front().get_name().size();
			for (Engine::AirportResult::elementvector_t::iterator i(m_airports.begin()), e(m_airports.end()); i != e; ++i) {
				if (i->get_name().size() <= namelen)
					continue;
				m_airports.erase(i, e);
				break;
			}
		}
		if (flags & flag_navaid) {
			std::sort(m_navaids.begin(), m_navaids.end(), SortNavaidNamelen());
			std::string::size_type namelen(0);
			if (!m_navaids.empty())
				namelen = m_navaids.front().get_name().size();
			for (Engine::NavaidResult::elementvector_t::iterator i(m_navaids.begin()), e(m_navaids.end()); i != e; ++i) {
				if (i->get_name().size() <= namelen)
					continue;
				m_navaids.erase(i, e);
				break;
			}
		}
	}
	{
		std::sort(m_mapelements.begin(), m_mapelements.end(), SortMapelementNamelen());
		std::string::size_type namelen(0);
		if (!m_mapelements.empty())
			namelen = m_mapelements.front().get_name().size();
		for (Engine::MapelementResult::elementvector_t::iterator i(m_mapelements.begin()), e(m_mapelements.end()); i != e; ++i) {
			if (i->get_name().size() <= namelen)
				continue;
			m_mapelements.erase(i, e);
			break;
		}
	}
	sort_dist(pt);
	retain_shortest_dist(pt);
       	return !(m_airports.empty() && m_navaids.empty() && m_waypoints.empty() && m_mapelements.empty());
}

bool IcaoFlightPlan::FindCoord::find(const Point& pt, unsigned int flags, float maxdist_km)
{
	m_coordresult = Point::invalid;
	async_cancel();
	async_clear();
	Rect bbox(pt.simple_box_km(maxdist_km));
	if (flags & flag_airport)
		m_airportquery = m_engine.async_airport_find_bbox(bbox, ~0,
								  (flags & flag_subtables) ? AirportsDb::element_t::subtables_all : AirportsDb::element_t::subtables_vfrroutes);
	if (flags & flag_navaid)
		m_navaidquery = m_engine.async_navaid_find_bbox(bbox, ~0,
								(flags & flag_subtables) ? NavaidsDb::element_t::subtables_all : NavaidsDb::element_t::subtables_none);
	if (flags & flag_waypoint)
		m_waypointquery = m_engine.async_waypoint_find_bbox(bbox, ~0,
								    (flags & flag_subtables) ? WaypointsDb::element_t::subtables_all : WaypointsDb::element_t::subtables_none);
	if (flags & flag_mapelement)
		m_mapelementquery = m_engine.async_mapelement_find_bbox(bbox, ~0,
									(flags & flag_subtables) ? MapelementsDb::element_t::subtables_all : MapelementsDb::element_t::subtables_none);
	if (flags & flag_coord) {
		m_coordresult.set_lon_deg_dbl(Point::round<int,double>(pt.get_lon_deg_dbl() * 60.0) * (1.0 / 60.0));
		m_coordresult.set_lat_deg_dbl(Point::round<int,double>(pt.get_lat_deg_dbl() * 60.0) * (1.0 / 60.0));
	}
	async_connectdone();
	async_waitdone();
	sort_dist(pt);
	cut_dist(pt, maxdist_km);
	retain_shortest_dist(pt);
	return !(m_airports.empty() && m_navaids.empty() && m_waypoints.empty() && m_mapelements.empty());
}

bool IcaoFlightPlan::FindCoord::find_airway(const std::string& awy)
{
	m_coordresult = Point::invalid;
	async_cancel();
	async_clear();
	m_airwayquery = m_engine.async_airway_graph_find_by_name(awy);
	async_connectdone();
	async_waitdone();
	return !!boost::num_vertices(m_airwaygraph);
}

bool IcaoFlightPlan::FindCoord::find_area(const Rect& rect, Engine::AirwayGraphResult::Vertex::typemask_t tmask)
{
	m_coordresult = Point::invalid;
	async_cancel();
	async_clear();
	m_areaquery = m_engine.async_area_graph_find_bbox(rect, tmask);
	async_connectdone();
	async_waitdone();
	return !!boost::num_vertices(m_airwaygraph);
}

std::string IcaoFlightPlan::FindCoord::get_name(void) const
{
	if (!m_airports.empty())
		return m_airports.front().get_name();
	if (!m_navaids.empty())
		return m_navaids.front().get_name();
	if (!m_waypoints.empty())
		return m_waypoints.front().get_name();
	if (!m_mapelements.empty())
		return m_mapelements.front().get_name();
	return empty;
}

std::string IcaoFlightPlan::FindCoord::get_icao(void) const
{
	if (!m_airports.empty())
		return m_airports.front().get_icao();
	if (!m_navaids.empty())
		return m_navaids.front().get_icao();
	if (!m_waypoints.empty())
		return m_waypoints.front().get_icao();
	return empty;
}

Point IcaoFlightPlan::FindCoord::get_coord(void) const
{
	if (!m_airports.empty())
		return m_airports.front().get_coord();
	if (!m_navaids.empty())
		return m_navaids.front().get_coord();
	if (!m_waypoints.empty())
		return m_waypoints.front().get_coord();
	if (!m_mapelements.empty())
		return m_mapelements.front().get_coord();
	if (!m_coordresult.is_invalid())
		return m_coordresult;
	return Point();
}

bool IcaoFlightPlan::FindCoord::get_airport(AirportsDb::element_t& x) const
{
	if (m_airports.empty())
		return false;
	x = m_airports.front();
	return true;
}

bool IcaoFlightPlan::FindCoord::get_navaid(NavaidsDb::element_t& x) const
{
	if (m_navaids.empty())
		return false;
	x = m_navaids.front();
	return true;
}

bool IcaoFlightPlan::FindCoord::get_waypoint(WaypointsDb::element_t& x) const
{
	if (m_waypoints.empty())
		return false;
	x = m_waypoints.front();
	return true;
}

bool IcaoFlightPlan::FindCoord::get_mapelement(MapelementsDb::element_t& x) const
{
	if (m_mapelements.empty())
		return false;
	x = m_mapelements.front();
	return true;
}

bool IcaoFlightPlan::FindCoord::get_coord(Point& x) const
{
	if (m_coordresult.is_invalid())
		return false;
	x = m_coordresult;
	return true;
}

bool IcaoFlightPlan::FindCoord::get_airwaygraph(Engine::AirwayGraphResult::Graph& x) const
{
	x = m_airwaygraph;
	return !!boost::num_vertices(m_airwaygraph);
}

IcaoFlightPlan::Waypoint::Waypoint(void)
	: m_time(0), m_alt(0), m_flags(0), m_frequency(0), m_pathcode(FPlanWaypoint::pathcode_none)
{
}

IcaoFlightPlan::Waypoint::Waypoint(const FPlanWaypoint& wpt)
	: m_icao(wpt.get_icao()), m_name(wpt.get_name()),
	  m_pathname(wpt.get_pathname()),
	  m_time(wpt.get_time_unix()), m_coord(wpt.get_coord()),
	  m_alt(wpt.get_altitude()), m_flags(wpt.get_flags()),
	  m_frequency(wpt.get_frequency()), m_pathcode(wpt.get_pathcode())
{
}

const std::string& IcaoFlightPlan::Waypoint::get_icao_or_name(void) const
{
	if (get_icao().empty())
		return get_name();
	return get_icao();
}

std::string IcaoFlightPlan::Waypoint::get_coordstr(void) const
{
	std::ostringstream oss;
	double c(get_lat() * Point::to_deg_dbl);
	char ch('N');
	if (c < 0) {
		c = -c;
		ch = 'S';
	}
	int ci(floor(c) + 0.01);
	c -= ci;
	c *= 60;
	oss << std::setw(2) << std::setfill('0') << ci;
	ci = floor(c + 0.5) + 0.01;
	oss << std::setw(2) << std::setfill('0') << ci << ch;
	c = get_lon() * Point::to_deg_dbl;
	ch = 'E';
	if (c < 0) {
		c = -c;
		ch = 'W';
	}
	ci = floor(c) + 0.01;
	c -= ci;
	c *= 60;
	oss << std::setw(3) << std::setfill('0') << ci;
	ci = floor(c + 0.5) + 0.01;
	oss << std::setw(2) << std::setfill('0') << ci << ch;
	return oss.str();	
}

bool IcaoFlightPlan::Waypoint::enforce_pathcode_vfrifr(void)
{
	Glib::ustring pn(get_pathname());
	{
		Glib::ustring::size_type n(pn.size()), i(0);
		while (i < n && std::isspace(pn[i]))
			++i;
		while (n > i) {
			--n;
			if (std::isspace(pn[n]))
				continue;
			++n;
			break;
		}
		pn = pn.substr(i, n - i);
	}
	FPlanWaypoint::pathcode_t pc(get_pathcode());
	if (get_flags() & FPlanWaypoint::altflag_ifr) {
		pn = pn.uppercase();
		switch (pc) {
		case FPlanWaypoint::pathcode_none:
			pc = FPlanWaypoint::pathcode_airway;
			// fall through

		case FPlanWaypoint::pathcode_airway:
			if (pn.empty() || pn == "DCT")
				pc = FPlanWaypoint::pathcode_directto;
			break;

		default:
			pn.clear();
			pc = FPlanWaypoint::pathcode_directto;
			break;
		}
	} else {
		switch (pc) {
		case FPlanWaypoint::pathcode_vfrdeparture:
		case FPlanWaypoint::pathcode_vfrarrival:
		case FPlanWaypoint::pathcode_vfrtransition:
			break;

		default:
			pn.clear();
			pc = FPlanWaypoint::pathcode_none;
			break;
		}
	}
	bool work(false);
	if (get_pathcode() != pc) {
		set_pathcode(pc);
		work = true;
	}
	if (get_pathname() != pn) {
		set_pathname(pn);
		work = true;
	}
	return work;
}

IcaoFlightPlan::IcaoFlightPlan(Engine& engine)
	: m_engine(engine), m_aircraftid("HBPBX"), m_aircrafttype("P28R"), m_equipment("SDFGLOR"), m_transponder("S"),
	  m_pbn("B2"), m_departuretime(0), m_totaleet(0), m_endurance(0), m_number(1), m_personsonboard(0),
	  m_departureflags(0), m_destinationflags(0), m_defaultalt(5000), m_flighttype('G'), m_wakecategory('L')
{
	time(&m_departuretime);
	m_departuretime += 60 * 60;
        Preferences& prefs(m_engine.get_prefs());
	{
		std::ostringstream oss;
		oss << 'N' << std::setw(4) << std::setfill('0') << (int)prefs.vcruise;
		m_cruisespeed = oss.str();
	}
}

void IcaoFlightPlan::trimleft(std::string::const_iterator& txti, std::string::const_iterator txte)
{
	std::string::const_iterator txti2(txti);
	for (; txti2 != txte && isspace(*txti2); ++txti2);
	txti = txti2;
}

void IcaoFlightPlan::trimright(std::string::const_iterator txti, std::string::const_iterator& txte)
{
	std::string::const_iterator txte2(txte);
	while (txti != txte2) {
		std::string::const_iterator txte3(txte2);
		--txte3;
		if (!isspace(*txte3))
			break;
		txte2 = txte3;
	}
	txte = txte2;
}

void IcaoFlightPlan::trim(std::string::const_iterator& txti, std::string::const_iterator& txte)
{
	trimleft(txti, txte);
	trimright(txti, txte);
}

std::string IcaoFlightPlan::upcase(const std::string& txt)
{
	Glib::ustring x(txt);
	return x.uppercase();
}

bool IcaoFlightPlan::parsetxt(std::string& txt, unsigned int len, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	trimleft(txti, txte);
	txt.clear();
	while (txti != txte) {
		if (isspace(*txti) || *txti == '/' || *txti == '-' || *txti == '(' || *txti == ')')
			break;
		txt.push_back(*txti);
		++txti;
		if (len && txt.size() >= len)
			break;
	}
	trimleft(txti, txte);
	return len ? txt.size() == len : !txt.empty();
}

bool IcaoFlightPlan::parsenum(unsigned int& num, unsigned int digits, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	trimleft(txti, txte);
	num = 0;
	unsigned int d = 0;
	while (txti != txte) {
		if (!isdigit(*txti))
			break;
		num *= 10;
		num += (*txti) - '0';
		++d;
		++txti;
		if (digits && d >= digits)
			break;
	}
	trimleft(txti, txte);
	return d && (!digits || d == digits);
}

bool IcaoFlightPlan::parsetime(time_t& t, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	t = 0;
	unsigned int n;
	if (!parsenum(n, 4, txti, txte))
		return false;
	t = ((n / 100) * 60 + (n % 100)) * 60;
	return true;
}

bool IcaoFlightPlan::parsespeed(double& spd, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	spd = 0;
	trimleft(txti, txte);
	if (txti == txte)
		return false;
	char ch(*txti);
	++txti;
	float factor(1);
	unsigned int len(4);
	switch (ch) {
	case 'N':
		break;

	case 'K':
		factor = Point::km_to_nmi;
		break;

	case 'M':
		factor = 330.0 * Point::mps_to_kts; // very approximate!
		len = 3;
		break;

	default:
		return false;
	}
	unsigned int num;
	if (!parsenum(num, len, txti, txte))
		return false;
	spd = num * factor;
	return true;
}

bool IcaoFlightPlan::parsespeed(std::string& spd, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	spd.clear();
	trimleft(txti, txte);
	if (txti == txte)
		return false;
	char ch(*txti);
	++txti;
	unsigned int len(4);
	switch (ch) {
	case 'N':
	case 'K':
		break;

	case 'M':
		len = 3;
		break;

	default:
		return false;
	}
	unsigned int num;
	if (!parsenum(num, len, txti, txte))
		return false;
	std::ostringstream oss;
	oss << ch << std::setw(len) << std::setfill('0') << num;
	spd = oss.str();
	return true;
}

bool IcaoFlightPlan::parsealt(int& alt, unsigned int& flags, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	alt = 0;
	flags &= ~FPlanWaypoint::altflag_standard;
	trimleft(txti, txte);
	if (txti == txte)
		return false;
	char ch(*txti);
	++txti;
	if (ch == 'V') {
		std::string s;
		if (!parsetxt(s, 2, txti, txte))
			return false;
		if (s != "FR")
			return false;
		flags |= 1 << 16;
		return true;
	}
	if (ch == 'F')
		flags |= FPlanWaypoint::altflag_standard;
	else if (ch != 'A')
		return false;
	unsigned int num;
	if (!parsenum(num, 3, txti, txte))
		return false;
	alt = num * 100;
	return true;
}

const unsigned int IcaoFlightPlan::parsewpt_coords;
const unsigned int IcaoFlightPlan::parsewpt_airway;
const unsigned int IcaoFlightPlan::parsewpt_frules;
const unsigned int IcaoFlightPlan::parsewpt_speedalt;

unsigned int IcaoFlightPlan::parsewpt(Waypoint& wpt, std::string::const_iterator& txti, std::string::const_iterator txte)
{
	trimleft(txti, txte);
	if (txti == txte)
		return 0;
	if (*txti == '/') {
		std::string::const_iterator txti1(txti);
		++txti1;
		std::string spd;
		if (!parsespeed(spd, txti1, txte))
			return 0;
		//set_cruisespeed(spd);
		int alt;
		unsigned int altflags;
		if (!parsealt(alt, altflags, txti1, txte))
			return 0;
		txti = txti1;
		altflags &= FPlanWaypoint::altflag_standard;
		altflags |= wpt.get_flags() & ~FPlanWaypoint::altflag_standard;
		wpt.set_altitude(alt);
		wpt.set_flags(altflags);
		return parsewpt_speedalt;
	}
	if (isdigit(*txti)) {
		unsigned int deg, min(0);
		std::string::const_iterator txti1(txti);
		if (!parsenum(deg, 2, txti1, txte))
			return 0;
		if (txti1 == txte)
			return 0;
		if (isdigit(*txti1)) {
			if (!parsenum(min, 2, txti1, txte))
				return 0;
			if (txti1 == txte)
				return 0;
		}
		if (*txti1 == 'S')
			wpt.set_lat_deg(-(deg * (1.0 / 60.0) * min));
		else if (*txti1 == 'N')
			wpt.set_lat_deg(deg * (1.0 / 60.0) * min);
		++txti1;
		min = 0;
		if (!parsenum(deg, 3, txti1, txte))
			return 0;
		if (txti1 == txte)
			return 0;
		if (isdigit(*txti1)) {
			if (!parsenum(min, 2, txti1, txte))
				return 0;
			if (txti1 == txte)
				return 0;
		}
		if (*txti1 == 'W')
			wpt.set_lat_deg(-(deg * (1.0 / 60.0) * min));
		else if (*txti1 == 'E')
			wpt.set_lat_deg(deg * (1.0 / 60.0) * min);
		++txti1;
		trimleft(txti1, txte);
		txti = txti1;
		wpt.set_icao("");
		wpt.set_name("");
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		wpt.set_pathname("");
		wpt.set_frequency(0);
		FindCoord f(m_engine);
		unsigned int acceptflags((wpt.get_flags() & FPlanWaypoint::altflag_ifr)
					 ? (FindCoord::flag_navaid | FindCoord::flag_waypoint)
					 : (FindCoord::flag_airport | FindCoord::flag_navaid | FindCoord::flag_waypoint));
		if (f.find(wpt.get_coord(), acceptflags, 0.1)) {
			wpt.set_icao(f.get_icao());
			wpt.set_name(f.get_name());
			wpt.set_coord(f.get_coord());
			{
				AirportsDb::Airport arpt;
				if (f.get_airport(arpt)) {
					if (arpt.get_nrcomm())
						wpt.set_frequency(arpt.get_comm(0)[0]);
				}
			}
			{
				NavaidsDb::Navaid nav;
				if (f.get_navaid(nav)) {
					wpt.set_frequency(nav.get_frequency());
				}
			}
		}
		return parsewpt_coords;
	}
	if (!isalpha(*txti))
		return 0;
	std::string::const_iterator txti1(txti);
	std::string name;
	if (!parsetxt(name, 0, txti, txte))
		return 0;
	if (name == "DCT") {
		wpt.set_pathcode(FPlanWaypoint::pathcode_directto);
		wpt.set_pathname("");
		return parsewpt_airway;
	}
	if (name == "VFR" || name == "IFR") {
		if (name == "VFR")
			wpt.set_flags(wpt.get_flags() & ~FPlanWaypoint::altflag_ifr);
		else
			wpt.set_flags(wpt.get_flags() | FPlanWaypoint::altflag_ifr);
		return parsewpt_frules;
	}
	FindCoord f(m_engine);
	unsigned int acceptflags((wpt.get_flags() & FPlanWaypoint::altflag_ifr)
				 ? (FindCoord::flag_navaid | FindCoord::flag_waypoint)
				 : (FindCoord::flag_airport | FindCoord::flag_navaid | FindCoord::flag_waypoint));
	if (!Engine::AirwayGraphResult::Vertex::is_ident_numeric(name) &&
	    f.find(name, name, acceptflags, wpt.get_coord())) {
		wpt.set_icao(f.get_icao());
		wpt.set_name(f.get_name());
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		wpt.set_pathname("");
		wpt.set_coord(f.get_coord());
		wpt.set_frequency(0);
		{
			AirportsDb::Airport arpt;
			if (f.get_airport(arpt)) {
				if (arpt.get_nrcomm())
					wpt.set_frequency(arpt.get_comm(0)[0]);
			}
		}
		{
			NavaidsDb::Navaid nav;
			if (f.get_navaid(nav)) {
				wpt.set_frequency(nav.get_frequency());
			}
		}
		return parsewpt_coords;
	}
	if (wpt.get_flags() & FPlanWaypoint::altflag_ifr) {
		if (f.find_airway(name)) {
			wpt.set_pathcode(FPlanWaypoint::pathcode_airway);
			wpt.set_pathname(name);
			return parsewpt_airway;
		}
	}
	txti = txti1;
	return 0;
}

void IcaoFlightPlan::otherinfo_clear(const std::string& category)
{
	otherinfo_t::iterator i(m_otherinfo.find(category));
	if (i != m_otherinfo.end())
		m_otherinfo.erase(i);
}

const std::string& IcaoFlightPlan::otherinfo_get(const std::string& category)
{
	static const std::string empty;
	otherinfo_t::iterator i(m_otherinfo.find(category));
	if (i == m_otherinfo.end())
		return empty;
	return i->second;
}

void IcaoFlightPlan::otherinfo_add(const std::string& category, std::string::const_iterator texti, std::string::const_iterator texte)
{
	if (category.empty())
		return;
	trim(texti, texte);
	std::pair<otherinfo_t::iterator,bool> ins(m_otherinfo.insert(std::make_pair(category, std::string(texti, texte))));
	if (ins.second)
		return;
	ins.first->second.push_back(' ');
	ins.first->second.insert(ins.first->second.end(), texti, texte);
}

void IcaoFlightPlan::set_departuretime(time_t x)
{
	m_departuretime = x;
	otherinfo_clear("DOF");
	if (x < 24*60*60)
		return;
	struct tm tm;
        if (!gmtime_r(&x, &tm))
                return;
	std::ostringstream oss;
	oss << std::setw(2) << std::setfill('0') << (tm.tm_year % 100)
	    << std::setw(2) << std::setfill('0') << (tm.tm_mon + 1)
	    << std::setw(2) << std::setfill('0') << tm.tm_mday;
	otherinfo_add("DOF", oss.str());
}

std::pair<std::string::const_iterator,std::string> IcaoFlightPlan::parse(std::string::const_iterator txti, std::string::const_iterator txte, bool do_expand_airways)
{
	m_defaultalt = 5000;
	{
		trimleft(txti, txte);
		if (txti != txte && *txti == '-') {
			++txti;
			trimleft(txti, txte);
		}
		if (txti == txte || *txti != '(')
			return std::make_pair(txti, "opening parenthesis not found");
		++txti;
		std::string s;
		if (!parsetxt(s, 3, txti, txte) || s != "FPL")
			return std::make_pair(txti, "flight plan does not start with FPL");
		if (txti == txte || *txti != '-')
			return std::make_pair(txti, "dash expected after FPL");
		++txti;
	}
	{
		std::string s;
		if (!parsetxt(s, 0, txti, txte) || s.size() < 1 || s.size() > 7)
			return std::make_pair(txti, "invalid callsign");
		set_aircraftid(s);
		if (txti == txte || *txti != '-')
			return std::make_pair(txti, "dash expected after callsign");
		++txti;
		trim(txti, txte);
	}
	if (txti == txte || (*txti != 'V' && *txti != 'I' && *txti != 'Z' && *txti != 'Y'))
		return std::make_pair(txti, "invalid flight rules");
	m_departureflags = 0;
	if (*txti == 'I' || *txti == 'Y')
		m_departureflags |= FPlanWaypoint::altflag_ifr;
	++txti;
	trim(txti, txte);
	if (txti == txte || !isalpha(*txti))
		return std::make_pair(txti, "invalid flight type");
	set_flighttype(*txti);
	++txti;
	trim(txti, txte);
	if (txti == txte || *txti != '-')
		return std::make_pair(txti, "dash expected aircraft type");
	++txti;
	{
		unsigned int num;
		set_number(1);
		if (parsenum(num, 0, txti, txte))
			set_number(num);
	}
	{
		std::string s;
		if (!parsetxt(s, 0, txti, txte) || s.size() < 2 || s.size() > 4)
			return std::make_pair(txti, "invalid aircraft type");
		set_aircrafttype(s);
	}
	if (txti == txte || *txti != '/')
		return std::make_pair(txti, "slash expected");
	++txti;
	trim(txti, txte);
	if (txti == txte || !isalpha(*txti))
		return std::make_pair(txti, "invalid wake turbulence category");
	set_wakecategory(*txti);
	++txti;
	trim(txti, txte);
	if (txti == txte || *txti != '-')
		return std::make_pair(txti, "dash expected before equipment");
	++txti;
	{
		std::string s;
		if (!parsetxt(s, 0, txti, txte))
			return std::make_pair(txti, "invalid equipment");
		set_equipment(s);
	}
	if (txti == txte || *txti != '/')
		return std::make_pair(txti, "slash expected");
	++txti;
	trim(txti, txte);
	{
		std::string s;
		if (txti == txte || !parsetxt(s, 0, txti, txte) || s.size() < 1)
			return std::make_pair(txti, "invalid transponder");
		set_transponder(s);
	}
	trim(txti, txte);
	if (txti == txte || *txti != '-')
		return std::make_pair(txti, "dash expected before departure");
	++txti;
	{
		std::string s;
		if (!parsetxt(s, 4, txti, txte))
			return std::make_pair(txti, "invalid departure");
		set_departure(s);
	}
	{
		time_t t;;
		if (!parsetime(t, txti, txte))
			return std::make_pair(txti, "invalid off-block");
		set_departuretime(t);
	}
	if (txti == txte || *txti != '-')
		return std::make_pair(txti, "dash expected before speed");
	++txti;
	{
		Waypoint wpt;
		// parse airspeed and altitude
		{
			std::string spd;
			if (!parsespeed(spd, txti, txte)) {
				std::string::const_iterator txti1(txti);
				std::string txt;
				if (parsetxt(txt, 0, txti1, txte))
					txt = " " + txt;;
				return std::make_pair(txti, "invalid speed" + txt);
			}
			set_cruisespeed(spd);
			int alt;
			unsigned int altflags;
			if (!parsealt(alt, altflags, txti, txte)) {
				std::string::const_iterator txti1(txti);
				std::string txt;
				if (parsetxt(txt, 0, txti1, txte))
					txt = " " + txt;;
				return std::make_pair(txti, "invalid altitude" + txt);
			}
			altflags &= ~FPlanWaypoint::altflag_ifr;
			altflags |= m_departureflags & FPlanWaypoint::altflag_ifr;
			wpt.set_altitude(alt);
			wpt.set_flags(altflags);
		}
		// try to find coordinates of departure aerodrome
		{
			FindCoord f(m_engine);
			if (!f.find(get_departure() == "ZZZZ" ? "" : get_departure(), otherinfo_get("DEP"), FindCoord::flag_airport))
				return std::make_pair(txti, "invalid departure aerodrome");
			wpt.set_coord(f.get_coord());
		}
		// parse route
		m_sid.clear();
		m_star.clear();
		while (txti != txte && *txti != '-' && *txti != ')') {
			unsigned int pw(parsewpt(wpt, txti, txte));
			if (!pw) {
				std::string::const_iterator txti1(txti);
				std::string txt;
				bool pok(parsetxt(txt, 0, txti1, txte));
				if (pok && ((m_route.empty() && (m_departureflags & FPlanWaypoint::altflag_ifr)) ||
					    (*txti1 == '-' && (wpt.get_flags() & FPlanWaypoint::altflag_ifr)))) {
					bool ok1(txt.size() >= 4);
					for (unsigned int i = 0; i < txt.size() && ok1; ++i) {
						if (i + 2 == txt.size()) {
							ok1 = ok1 && std::isdigit(txt[i]);
							continue;
						}
						ok1 = ok1 && std::isalpha(txt[i]);
					}
					if (ok1) {
						txti = txti1;
						if (m_route.empty() && (m_departureflags & FPlanWaypoint::altflag_ifr)) {
							m_sid = txt;
							continue;
						}
						m_star = txt;
						continue;
					}
				}
				if (pok)
					txt = " " + txt;;
				return std::make_pair(txti, "invalid waypoint" + txt);
			}
			if (pw == parsewpt_coords) {
				m_route.push_back(wpt);
				continue;
			}
			if (m_route.empty()) {
				if (pw == parsewpt_airway) {
					if (wpt.get_pathcode() == FPlanWaypoint::pathcode_directto)
						m_sid = "DCT";
					else if (wpt.get_pathcode() == FPlanWaypoint::pathcode_airway)
						m_sid = wpt.get_pathname();
				}
			} else {
				m_route.back() = wpt;
			}
		}
		m_destinationflags = wpt.get_flags() & FPlanWaypoint::altflag_ifr;
	}
	if (txti == txte || *txti != '-')
		return std::make_pair(txti, "dash expected before destination");
	++txti;
	{
		std::string s;
		if (!parsetxt(s, 4, txti, txte))
			return std::make_pair(txti, "invalid destination");
		set_destination(s);
	}
	{
		time_t t;;
		if (!parsetime(t, txti, txte))
			return std::make_pair(txti, "invalid total eet");
		set_totaleet(t);
	}
	if (txti != txte && std::isalpha(*txti)) {
		std::string s;
		if (!parsetxt(s, 4, txti, txte))
			return std::make_pair(txti, "invalid first alternate");
		set_alternate1(s);
	}
	if (txti != txte && std::isalpha(*txti)) {
		std::string s;
		if (!parsetxt(s, 4, txti, txte))
			return std::make_pair(txti, "invalid second alternate");
		set_alternate2(s);
	}
	if (txti != txte && *txti == '-') {
		++txti;
		std::string cat;
		while (txti != txte && *txti != '-' && *txti != ')') {
			std::string s;
			if (!parsetxt(s, 0, txti, txte))
				return std::make_pair(txti, "invalid other information");
			if (txti != txte && *txti == '/') {
				cat.swap(s);
				++txti;
				continue;
			}
			if (cat.empty())
				return std::make_pair(txti, "invalid other information");
			if (cat == "PBN")
				set_pbn(s);
			else
				otherinfo_add(cat, s);
		}
		if (txti != txte && *txti == '-') {
			++txti;
			std::string cat;
			while (txti != txte && *txti != '-' && *txti != ')') {
				std::string s;
				if (!parsetxt(s, 0, txti, txte))
					return std::make_pair(txti, "invalid other2 information");
				if (txti != txte && *txti == '/') {
					cat.swap(s);
					++txti;
					if (cat == "E") {
						time_t t;;
						if (!parsetime(t, txti, txte))
							return std::make_pair(txti, "invalid endurance");
						set_endurance(t);
						continue;
					}
					if (cat == "P") {
						unsigned int num;
						set_personsonboard_tbn();
						if (parsenum(num, 3, txti, txte)) {
							set_personsonboard(num);
							continue;
						}
						if (!parsetxt(s, 3, txti, txte) || s != "TBN")
							return std::make_pair(txti, "invalid number of persons on board");
						continue;
					}
					continue;
				}
				if (cat.empty())
					return std::make_pair(txti, "invalid other2 information");
				if (cat == "A") {
					if (get_colourmarkings().empty())
						set_colourmarkings(s);
					else
						set_colourmarkings(get_colourmarkings() + " " + s);
					continue;
				}
				if (cat == "N") {
					if (get_remarks().empty())
						set_remarks(s);
					else
						set_remarks(get_remarks() + " " + s);
					continue;
				}
				if (cat == "C") {
					if (get_picname().empty())
						set_picname(s);
					else
						set_picname(get_picname() + " " + s);
					continue;
				}
				if (cat == "R") {
					if (get_emergencyradio().empty())
						set_emergencyradio(s);
					else
						set_emergencyradio(get_emergencyradio() + " " + s);
					continue;
				}
				if (cat == "S") {
					if (get_survival().empty())
						set_survival(s);
					else
						set_survival(get_survival() + " " + s);
					continue;
				}
				if (cat == "J") {
					if (get_lifejackets().empty())
						set_lifejackets(s);
					else
						set_lifejackets(get_lifejackets() + " " + s);
					continue;
				}
				if (cat == "D") {
					if (get_dinghies().empty())
						set_dinghies(s);
					else
						set_dinghies(get_dinghies() + " " + s);
					continue;
				}
			}
		}
	}
	normalize_pogo();
	std::string err(expand_airways(do_expand_airways));
	return std::make_pair(txti, err);
}

Rect IcaoFlightPlan::get_bbox(void) const
{
	route_t::const_iterator ri(m_route.begin()), re(m_route.end());
	if (ri == re)
                return Rect();
        Point pt(ri->get_coord());
	++ri;
        Point sw(0, 0), ne(0, 0);
        for (; ri != re; ++ri) {
                Point p(ri->get_coord() - pt);
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        return Rect(sw + pt, ne + pt);
}

char IcaoFlightPlan::get_flightrules(void) const
{
	unsigned int ftype(!!(m_departureflags & FPlanWaypoint::altflag_ifr));
	{
		unsigned int ft(!!(m_destinationflags & FPlanWaypoint::altflag_ifr));
		ft ^= ftype;
		ft &= 1;
		ftype |= ft << 1;
	}
	static const char frules[4] = { 'V', 'I', 'Z', 'Y' };
	if (ftype & 2)
		return frules[ftype];
	for (route_t::const_iterator ri(m_route.begin()), re(m_route.end()); ri != re; ++ri) {
		unsigned int ft(!!(ri->get_flags() & FPlanWaypoint::altflag_ifr));
		ft ^= ftype;
		ft &= 1;
		ftype |= ft << 1;
	}
	return frules[ftype];
}

IcaoFlightPlan::nameset_t IcaoFlightPlan::intersect(const nameset_t& s1, const nameset_t& s2)
{
	nameset_t r;
	std::set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(r, r.begin()));
	return r;
}

std::string IcaoFlightPlan::expand_airways(bool really_expand)
{
	std::string err;
	// add all intermediate airway segments
	static const float awy_tolerance_corr(50);
	static const float awy_tolerance(0.1);
	Engine::AirwayGraphResult::Graph areagraph;
	{
		FindCoord f(m_engine);
		f.find_area(get_bbox().oversize_nmi(200.f),
			    Engine::AirwayGraphResult::Vertex::typemask_navaid |
			    Engine::AirwayGraphResult::Vertex::typemask_intersection |
			    Engine::AirwayGraphResult::Vertex::typemask_airway);
		f.get_airwaygraph(areagraph);
		areagraph.delete_disconnected_vertices();
	}
	if (!m_route.empty()) {
		Waypoint& wpt(m_route.back());
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		wpt.set_pathname("");
	}
	for (route_t::iterator ri(m_route.end()), rb(m_route.begin()); ri != rb; ) {
		--ri;
		Waypoint& wpt2(*ri);
		if (ri == rb)
			continue;
		--ri;
		Waypoint& wpt1(*ri);
		++ri;
		if (!(wpt1.get_flags() & FPlanWaypoint::altflag_ifr)) {
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt1.set_pathname("");
			continue;
		}
		if (wpt1.get_pathname().empty()) {
			if (wpt1.get_pathcode() == FPlanWaypoint::pathcode_airway ||
			    wpt1.get_pathcode() == FPlanWaypoint::pathcode_sid ||
			    wpt1.get_pathcode() == FPlanWaypoint::pathcode_star)
				wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			continue;
		}
		if (wpt1.get_pathcode() != FPlanWaypoint::pathcode_airway)
			continue;
		Engine::AirwayGraphResult::Graph::vertex_descriptor u1;
		if (!areagraph.find_nearest(u1, wpt1.get_coord(), wpt1.get_pathname(),
					    Engine::AirwayGraphResult::Vertex::typemask_navaid |
					    Engine::AirwayGraphResult::Vertex::typemask_intersection |
					    Engine::AirwayGraphResult::Vertex::typemask_undefined,
					    true, false)) {
			err += "; Airway " + wpt1.get_pathname() + " not found";
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt1.set_pathname("");
			continue;
		}
		const Engine::AirwayGraphResult::Vertex& v1(areagraph[u1]);
		if ((v1.get_ident() != wpt1.get_icao_or_name() && !wpt1.get_icao_or_name().empty()) ||
		    wpt1.get_coord().spheric_distance_nmi(v1.get_coord()) > awy_tolerance_corr) {
			err += "; Airway " + wpt1.get_pathname() + " not found or does not contain point " + wpt1.get_icao_or_name();
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt1.set_pathname("");
			continue;
		}
		wpt1.set_coord(v1.get_coord());
		Engine::AirwayGraphResult::Graph::vertex_descriptor u2;
		if (!areagraph.find_nearest(u2, wpt2.get_coord(), wpt1.get_pathname(),
					    Engine::AirwayGraphResult::Vertex::typemask_navaid |
					    Engine::AirwayGraphResult::Vertex::typemask_intersection |
					    Engine::AirwayGraphResult::Vertex::typemask_undefined,
					    false, true)) {
			err += "; Airway " + wpt1.get_pathname() + " not found";
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt1.set_pathname("");
			continue;
		}
		if (u1 == u2) {
			err += "; Airway " + wpt1.get_pathname() + " points " + wpt1.get_icao_or_name() + " and " + wpt2.get_icao_or_name() + " are the same";
			--ri;
			ri = m_route.erase(ri);
			rb = m_route.begin();
			continue;
		}
		const Engine::AirwayGraphResult::Vertex& v2(areagraph[u2]);
		if ((v2.get_ident() != wpt2.get_icao_or_name() && !wpt2.get_icao_or_name().empty()) ||
		    wpt2.get_coord().spheric_distance_nmi(v2.get_coord()) > awy_tolerance) {
			err += "; Airway " + wpt1.get_pathname() + " not found or does not contain point " + wpt2.get_icao_or_name();
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt1.set_pathname("");
			continue;
		}
		std::vector<Engine::AirwayGraphResult::Graph::vertex_descriptor> predecessors;
		areagraph.shortest_paths(u1, wpt1.get_pathname(), predecessors);
		Engine::AirwayGraphResult::Graph::vertex_descriptor uu(predecessors[u2]);
		if (uu == u2) {
			err += "; Airway " + wpt1.get_pathname() + " does not connect points " + wpt1.get_icao_or_name() + " and " + wpt2.get_icao_or_name();
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
			wpt1.set_pathname("");
			continue;
		}
		if (!really_expand)
			continue;
		Waypoint wpt(wpt1);
		for (; uu != u1; uu = predecessors[uu]) {
			const Engine::AirwayGraphResult::Vertex& vv(areagraph[uu]);
			wpt.set_name(vv.get_ident());
			wpt.set_coord(vv.get_coord());
			ri = m_route.insert(ri, wpt);
		}
		rb = m_route.begin();
	}
	if (err.size() < 2)
		return err;
	return err.substr(2);
}

void IcaoFlightPlan::populate(const FPlanRoute& route, awymode_t awymode, double dct_limit)
{
	m_departure.clear();
	m_destination.clear();
	m_sid.clear();
	m_star.clear();
	m_totaleet = 0;
	m_route.clear();
	m_departureflags = m_destinationflags = 0;
	m_defaultalt = 5000;
	uint16_t nrwpt = route.get_nrwpt();
	if (!nrwpt)
		return;
	set_departuretime(route.get_time_offblock_unix());
	m_totaleet = route[nrwpt - 1].get_time_unix() - route[0].get_time_unix();
	m_departure = upcase(route[0].get_icao());
	otherinfo_clear("DEP");
	if (AirportsDb::Airport::is_fpl_zzzz(m_departure)) {
		m_departure = "ZZZZ";
		otherinfo_add("DEP", upcase(route[0].get_name()) + " " + route[0].get_coord().get_fpl_str());
		if (false)
			std::cerr << "DEP: " << otherinfo_get("DEP") << std::endl;
	}
	m_departureflags = route[0].get_flags();
	m_destination = upcase(route[nrwpt - 1].get_icao());
	otherinfo_clear("DEST");
	if (AirportsDb::Airport::is_fpl_zzzz(m_destination)) {
		m_destination = "ZZZZ";
		otherinfo_add("DEST", upcase(route[nrwpt - 1].get_name()) + " " + route[nrwpt - 1].get_coord().get_fpl_str());
		if (false)
			std::cerr << "DEST: " << otherinfo_get("DEST") << std::endl;
	}
	m_destinationflags = route[nrwpt - 1].get_flags();
	m_defaultalt = std::max(0, (int)std::max(route[0].get_altitude(), route[nrwpt - 1].get_altitude()));
	if (m_defaultalt >= 5000)
		m_defaultalt += 1000;
	m_defaultalt += 1999;
	m_defaultalt /= 1000;
	m_defaultalt *= 1000;
	{
		bool have_ifr(!!((m_departureflags | m_destinationflags) & FPlanWaypoint::altflag_ifr));
		for (uint16_t nr = 1; nr + 1 < nrwpt; ++nr) {
			if (route[nr].get_flags() & FPlanWaypoint::altflag_ifr &&
			    route[nr].get_pathcode() == FPlanWaypoint::pathcode_sid)
				continue;
			if ((Engine::AirwayGraphResult::Vertex::is_ident_numeric(route[nr].get_ident()) ||
			     route[nr].get_ident().size() < 2) &&
			    route[nr].get_pathcode() == FPlanWaypoint::pathcode_airway &&
			    route[nr-1].get_pathcode() == FPlanWaypoint::pathcode_airway &&
			    route[nr].get_pathname() == route[nr-1].get_pathname() &&
			    (route[nr].get_flags() & route[nr-1].get_flags() & FPlanWaypoint::altflag_ifr))
				continue;
			m_route.push_back(route[nr]);
			m_route.back().set_time(route[nr].get_time_unix() - route[0].get_time_unix());
			have_ifr = have_ifr || (m_route.back().get_flags() & FPlanWaypoint::altflag_ifr);
			if (route[nr].get_flags() & FPlanWaypoint::altflag_ifr &&
			    route[nr].get_pathcode() == FPlanWaypoint::pathcode_star)
				break;
		}
		if (have_ifr)
			otherinfo_add("RMK", /*"IFPS RTE AMDT ACPT"*/ "IFPSRA");
	}
	bool has_pc(has_pathcodes());
	if (m_departureflags & FPlanWaypoint::altflag_ifr) {
		if (route[0].get_pathcode() == FPlanWaypoint::pathcode_sid) {
			m_sid = route[0].get_pathname();
			{
				std::string::size_type n(m_sid.find('-'));
				if (n != std::string::npos)
					m_sid.erase(n);
			}
			if (m_sid.size() < 3) {
				m_sid.clear();
			} else {
				if (std::isdigit(m_sid[m_sid.size() - 2]) &&
				    std::isalpha(m_sid[m_sid.size() - 1]) &&
				    m_route.size() >= 2) {
					const std::string *name(&m_route.front().get_icao());
					if (name->empty() || *name == m_departure)
						name = &m_route.front().get_name();
					if (m_sid.substr(0, m_sid.size() - 2) == name->substr(0, m_sid.size() - 2))
						m_sid = *name + m_sid.substr(m_sid.size() - 2);
				}
			}
		}
	} else {
		bool vd(route[0].get_pathcode() == FPlanWaypoint::pathcode_vfrdeparture);
		if (vd) {
			while (!m_route.empty()) {
				const Waypoint& wpt(m_route[0]);
				if (wpt.get_flags() & FPlanWaypoint::altflag_ifr)
					break;
				if (wpt.get_pathcode() != FPlanWaypoint::pathcode_vfrdeparture) {
					if (wpt.get_pathcode() != FPlanWaypoint::pathcode_none)
						break;
					m_route.erase(m_route.begin(), m_route.begin() + 1);
					break;
				}
				m_route.erase(m_route.begin(), m_route.begin() + 1);
			}
		} else if (!has_pc) {
			erase_vfrroute_begin();
		}
	}
	if (!m_route.empty()) {
		if (m_destinationflags & FPlanWaypoint::altflag_ifr) {
			if (m_route[m_route.size() - 1U].get_pathcode() == FPlanWaypoint::pathcode_star) {
				m_star = m_route[m_route.size() - 1U].get_pathname();
				{
					std::string::size_type n(m_star.find('-'));
					if (n != std::string::npos)
						m_star.erase(n);
				}
				if (m_star.size() < 3) {
					m_star.clear();
				} else {
					if (std::isdigit(m_star[m_star.size() - 2]) &&
					    std::isalpha(m_star[m_star.size() - 1]) &&
					    m_route.size() >= 2) {
						const std::string *name(&m_route.back().get_icao());
						if (name->empty() || *name == m_destination)
							name = &m_route.back().get_name();
						if (m_star.substr(0, m_star.size() - 2) == name->substr(0, m_star.size() - 2))
							m_star = *name + m_star.substr(m_star.size() - 2);
					}
				}
			}
		} else {
			bool va(m_route[m_route.size() - 1U].get_pathcode() == FPlanWaypoint::pathcode_vfrarrival);
			if (va) {
				while (!m_route.empty()) {
					const Waypoint& wpt(m_route[m_route.size() - 1U]);
					if (wpt.get_flags() & FPlanWaypoint::altflag_ifr ||
					    wpt.get_pathcode() != FPlanWaypoint::pathcode_vfrarrival)
						break;
					m_route.resize(m_route.size() - 1U);
				}
			} else if (!has_pc) {
				erase_vfrroute_end();
			}
		}
	}
	redo_route_names();
       	find_airways();
	enforce_pathcode_vfrifr();
	fix_max_dct_distance(dct_limit);
	if (awymode == awymode_collapse || awymode == awymode_collapse_all)
		erase_unnecessary_airway(awymode != awymode_collapse_all);
	add_eet();
	normalize_pogo();
	if (false) {
		std::cerr << "Other Info:";
                for (otherinfo_t::const_iterator oi(m_otherinfo.begin()), oe(m_otherinfo.end()); oi != oe; ++oi)
			std::cerr << ' ' << oi->first << '/' << oi->second;
		std::cerr << std::endl;
	}
}

bool IcaoFlightPlan::has_pathcodes(void) const
{
        for (unsigned int i = 0; i < m_route.size(); ++i)
		if (m_route[i].get_pathcode() != FPlanWaypoint::pathcode_none)
			return true;
	return false;
}

bool IcaoFlightPlan::enforce_pathcode_vfrdeparr(void)
{
	unsigned int nr(m_route.size());
	if (nr < 1U)
		return false;
	bool work(false);
	if (!(m_departureflags & FPlanWaypoint::altflag_ifr)) {
		unsigned int i(0);
		for (; i < nr; ++i) {
			const Waypoint& wpt(m_route[i]);
			if (wpt.get_icao() != m_departure ||
			    (wpt.get_flags() & FPlanWaypoint::altflag_ifr) ||
			    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrarrival ||
			    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrtransition)
				break;
		}
		if (i > 0) {
			work = true;
			while (i > 0) {
				--i;
				m_route[i].set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
			}
		}
	}
	if (!(m_destinationflags & FPlanWaypoint::altflag_ifr)) {
		unsigned int i(nr);
		while (i > 0) {
			--i;
			const Waypoint& wpt(m_route[i]);
			if (wpt.get_icao() != m_destination ||
			    (wpt.get_flags() & FPlanWaypoint::altflag_ifr) ||
			    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrdeparture ||
			    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrtransition)
				break;
		}
		if (i < nr - 1) {
			work = true;
			while (i < nr - 1) {
				++i;
				m_route[i].set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
			}
		}
	}
	return work;
}

bool IcaoFlightPlan::enforce_pathcode_vfrifr(void)
{
	unsigned int nr(m_route.size());
	bool work(false);
	for (unsigned int i = 0; i < nr; ++i) {
		Waypoint& wpt(m_route[i]);
		work = wpt.enforce_pathcode_vfrifr() || work;
	}
	return work;
}

bool IcaoFlightPlan::mark_vfrroute_begin(FPlanRoute& route)
{
	if (route.get_nrwpt() < 2U)
		return false;
	AirportsDb::element_t apt;
	{
		FindCoord f(m_engine);
		if (!f.find(route[0U].get_icao(), route[0U].get_name(), FindCoord::flag_airport))
			return false;
		if (!f.get_airport(apt))
			return false;
	}
	unsigned int bestroute(0);
	unsigned int bestcount(0);
	for (unsigned int rtenr(0); rtenr < apt.get_nrvfrrte(); ++rtenr) {
		const AirportsDb::Airport::VFRRoute& rte(apt.get_vfrrte(rtenr));
		if (!rte.size())
			continue;
		unsigned int ptnr(0);
		for (; ptnr < rte.size() && ptnr <= route.get_nrwpt(); ++ptnr)
			if (rte[ptnr].get_coord() != route[ptnr].get_coord())
				break;
		if (ptnr > bestcount) {
			bestcount = ptnr;
			bestroute = rtenr;
		}
	}
	if (bestcount < 2U)
		return false;
	--bestcount;
	const AirportsDb::Airport::VFRRoute& rte(apt.get_vfrrte(bestroute));
	for (unsigned int ptnr(0); ptnr < bestcount; ++ptnr) {
		FPlanWaypoint& wpt(route[ptnr]);
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(rte[ptnr]);
		wpt.set_pathname(rte.get_name());
		switch (vfrpt.get_pathcode()) {
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
			break;

		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
			break;

		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_vfrtransition:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
			break;

		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_holding:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_trafficpattern:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_other:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_invalid:
		default:
			wpt.set_pathcode(FPlanWaypoint::pathcode_none);
			break;
		}
	}
	return true;
}

bool IcaoFlightPlan::mark_vfrroute_end(FPlanRoute& route)
{
	if (route.get_nrwpt() < 2U)
		return false;
	AirportsDb::element_t apt;
	{
		FindCoord f(m_engine);
		if (!f.find(route[route.get_nrwpt() - 1U].get_icao(), route[route.get_nrwpt() - 1U].get_name(), FindCoord::flag_airport))
			return false;
		if (!f.get_airport(apt))
			return false;
	}
	unsigned int bestroute(0);
	unsigned int bestcount(0);
	for (unsigned int rtenr(0); rtenr < apt.get_nrvfrrte(); ++rtenr) {
		const AirportsDb::Airport::VFRRoute& rte(apt.get_vfrrte(rtenr));
		if (!rte.size())
			continue;
		unsigned int ptnr(1);
		for (; ptnr < rte.size() && ptnr <= route.get_nrwpt(); ++ptnr)
			if (rte[rte.size() - ptnr].get_coord() != route[route.get_nrwpt() - ptnr].get_coord())
				break;
		--ptnr;
		if (ptnr > bestcount) {
			bestcount = ptnr;
			bestroute = rtenr;
		}
	}
	if (bestcount < 2U)
		return false;
	const AirportsDb::Airport::VFRRoute& rte(apt.get_vfrrte(bestroute));
	for (unsigned int ptnr(2); ptnr <= bestcount; ++ptnr) {
		FPlanWaypoint& wpt(route[route.get_nrwpt() - ptnr]);
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(rte[rte.size() - ptnr]);
		wpt.set_pathname(rte.get_name());
		switch (vfrpt.get_pathcode()) {
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
			break;

		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
			break;

		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_vfrtransition:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
			break;

		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_holding:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_trafficpattern:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_other:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_invalid:
		default:
			wpt.set_pathcode(FPlanWaypoint::pathcode_none);
			break;
		}
	}
	return true;
}

void IcaoFlightPlan::erase_vfrroute_begin(void)
{
	if (m_departureflags & FPlanWaypoint::altflag_ifr)
		return;
	AirportsDb::element_t apt;
	{
		FindCoord f(m_engine);
		if (!f.find(m_departure == "ZZZZ" ? "" : m_departure, otherinfo_get("DEP"), FindCoord::flag_airport))
			return;
		if (!f.get_airport(apt))
			return;
	}
	unsigned int bestroute(0);
	unsigned int bestcount(0);
	for (unsigned int rtenr(0); rtenr < apt.get_nrvfrrte(); ++rtenr) {
		const AirportsDb::Airport::VFRRoute& rte(apt.get_vfrrte(rtenr));
		if (!rte.size())
			continue;
		if (!rte[0].is_at_airport())
			continue;
		unsigned int ptnr(1);
		for (; ptnr < rte.size() && ptnr <= m_route.size(); ++ptnr)
			if (rte[ptnr].get_coord() != m_route[ptnr-1].get_coord())
				break;
		--ptnr;
		if (ptnr > bestcount) {
			bestcount = ptnr;
			bestroute = rtenr;
		}
	}
	if (!bestcount)
		return;
	m_route.erase(m_route.begin(), m_route.begin() + bestcount);
}

void IcaoFlightPlan::erase_vfrroute_end(void)
{
	if (m_destinationflags & FPlanWaypoint::altflag_ifr)
		return;
	AirportsDb::element_t apt;
	{
		FindCoord f(m_engine);
		if (!f.find(m_destination == "ZZZZ" ? "" : m_destination, otherinfo_get("DEST"), FindCoord::flag_airport))
			return;
		if (!f.get_airport(apt))
			return;
	}
	unsigned int bestroute(0);
	unsigned int bestcount(0);
	for (unsigned int rtenr(0); rtenr < apt.get_nrvfrrte(); ++rtenr) {
		const AirportsDb::Airport::VFRRoute& rte(apt.get_vfrrte(rtenr));
		if (!rte.size())
			continue;
		if (!rte[rte.size()-1].is_at_airport())
			continue;
		unsigned int ptnr(1);
		for (; ptnr < rte.size() && ptnr <= m_route.size(); ++ptnr)
			if (rte[rte.size()-1-ptnr].get_coord() != m_route[m_route.size() - ptnr].get_coord())
				break;
		--ptnr;
		if (ptnr > bestcount) {
			bestcount = ptnr;
			bestroute = rtenr;
		}
	}
	if (!bestcount)
		return;
	m_route.erase(m_route.end() - bestcount, m_route.end());
}

void IcaoFlightPlan::redo_route_names(void)
{
	for (route_t::iterator ri(m_route.begin()), re(m_route.end()); ri != re; ++ri) {
		Waypoint& wpt(*ri);
		FindCoord f(m_engine);
		unsigned int acceptflags((wpt.get_flags() & FPlanWaypoint::altflag_ifr)
					 ? (FindCoord::flag_navaid | FindCoord::flag_waypoint)
					 : (FindCoord::flag_airport | FindCoord::flag_navaid | FindCoord::flag_waypoint));
		if (true)
			std::cerr << "IcaoFlightPlan::redo_route_names: try " << wpt.get_icao() << ' ' << wpt.get_name() << std::endl;
		bool ok(f.find(wpt.get_icao(), wpt.get_name(), acceptflags));
		ok = ok && (wpt.get_coord().spheric_distance_km(f.get_coord()) < 0.1);
		if (!ok) {
			if (true)
				std::cerr << "IcaoFlightPlan::redo_route_names: try " << wpt.get_coord().get_lat_str2()
					  << ' ' << wpt.get_coord().get_lon_str2() << std::endl;
			ok = f.find(wpt.get_coord(), acceptflags, 0.1);
		}
		if (ok) {
			if (true)
				std::cerr << "IcaoFlightPlan::redo_route_names: set " << wpt.get_icao() << ' ' << wpt.get_name() << std::endl;
			wpt.set_icao(f.get_icao());
			wpt.set_name(f.get_name());
			continue;
		}
		wpt.set_icao("");
		wpt.set_name("");
	}
}

void IcaoFlightPlan::find_airways(void)
{
	if (m_route.size() < 2)
		return;
	// add airways to IFR segments
	static const float awy_tolerance(0.1);
	Engine::AirwayGraphResult::Graph areagraph;
	{
		FindCoord f(m_engine);
		f.find_area(get_bbox().oversize_nmi(200));
		f.get_airwaygraph(areagraph);
	}
	for (route_t::iterator ri(m_route.begin()), re(m_route.end()); ri != re; ) {
		Waypoint& wpt1(*ri);
		++ri;
		if (ri == re)
			break;
		if (!(wpt1.get_flags() & FPlanWaypoint::altflag_ifr))
			continue;
		switch (wpt1.get_pathcode()) {
		case FPlanWaypoint::pathcode_directto:
			wpt1.set_pathname("");
			continue;

		case FPlanWaypoint::pathcode_airway:
		case FPlanWaypoint::pathcode_sid:
		case FPlanWaypoint::pathcode_star:
			if (!wpt1.get_pathname().empty())
				continue;
			/* fall through */

		default:
			wpt1.set_pathcode(FPlanWaypoint::pathcode_none);
		case FPlanWaypoint::pathcode_none:
			wpt1.set_pathname("");
			break;
		}
		Waypoint& wpt2(*ri);
		Engine::AirwayGraphResult::Graph::vertex_descriptor u1;
		if (!areagraph.find_nearest(u1, wpt1.get_coord(),
					    Engine::AirwayGraphResult::Vertex::typemask_navaid |
					    Engine::AirwayGraphResult::Vertex::typemask_intersection |
					    Engine::AirwayGraphResult::Vertex::typemask_undefined,
					    true, false)) {
			if (true)
				std::cerr << "Point not found: " << wpt1.get_icao_or_name() << ' '
					  << wpt1.get_coord().get_lat_str2() << ' ' << wpt1.get_coord().get_lon_str2() << std::endl;
			continue;
		}
		const Engine::AirwayGraphResult::Vertex& v1(areagraph[u1]);
		if ((v1.get_ident() != wpt1.get_icao_or_name() && !wpt1.get_icao_or_name().empty()) ||
		    wpt1.get_coord().spheric_distance_nmi(v1.get_coord()) > awy_tolerance) {
			if (true)
				std::cerr << "Point found for: " << wpt1.get_icao_or_name() << ' '
					  << wpt1.get_coord().get_lat_str2() << ' ' << wpt1.get_coord().get_lon_str2()
					  << " has different ID or is too far away: " << v1.get_ident() << ' '
					  << v1.get_coord().get_lat_str2() << ' ' << v1.get_coord().get_lon_str2() << std::endl;
			continue;
		}
		Engine::AirwayGraphResult::Graph::vertex_descriptor u2;
		if (!areagraph.find_nearest(u2, wpt2.get_coord(),
					    Engine::AirwayGraphResult::Vertex::typemask_navaid |
					    Engine::AirwayGraphResult::Vertex::typemask_intersection |
					    Engine::AirwayGraphResult::Vertex::typemask_undefined,
					    false, true)) {
			if (true)
				std::cerr << "Point not found: " << wpt2.get_icao_or_name() << ' '
					  << wpt2.get_coord().get_lat_str2() << ' ' << wpt2.get_coord().get_lon_str2() << std::endl;
			continue;
		}
		const Engine::AirwayGraphResult::Vertex& v2(areagraph[u2]);
		if (u1 == u2) {
			if (true)
				std::cerr << "Points: " << v1.get_ident() << ' '
					  << v1.get_coord().get_lat_str2() << ' ' << v1.get_coord().get_lon_str2()
					  << " and: " << v2.get_ident() << ' '
					  << v2.get_coord().get_lat_str2() << ' ' << v2.get_coord().get_lon_str2()
					  << " are the same" << std::endl;
			continue;
		}
		if ((v2.get_ident() != wpt2.get_icao_or_name() && !wpt2.get_icao_or_name().empty()) ||
		    wpt2.get_coord().spheric_distance_nmi(v2.get_coord()) > awy_tolerance) {
			if (true)
				std::cerr << "Point found for: " << wpt2.get_icao_or_name() << ' '
					  << wpt2.get_coord().get_lat_str2() << ' ' << wpt2.get_coord().get_lon_str2()
					  << " has different ID or is too far away: " << v2.get_ident() << ' '
					  << v2.get_coord().get_lat_str2() << ' ' << v2.get_coord().get_lon_str2() << std::endl;
			continue;
		}
		std::vector<Engine::AirwayGraphResult::Graph::vertex_descriptor> predecessors;
		areagraph.shortest_paths(u1, predecessors);
		Engine::AirwayGraphResult::Graph::vertex_descriptor uu2(u2);
		Engine::AirwayGraphResult::Graph::vertex_descriptor uu1(predecessors[uu2]);
		if (uu1 == uu2) {
			if (true)
				std::cerr << "No Path from point: " << v1.get_ident() << ' '
					  << v1.get_coord().get_lat_str2() << ' ' << v1.get_coord().get_lon_str2()
					  << " to point: " << v2.get_ident() << ' '
					  << v2.get_coord().get_lat_str2() << ' ' << v2.get_coord().get_lon_str2() << std::endl;
       			continue;
		}
		nameset_t awynames;
		for (bool first = true; uu2 != u1; uu2 = uu1, uu1 = predecessors[uu2]) {
			nameset_t an;
			int a((wpt1.get_altitude() + 50) / 100);
			Engine::AirwayGraphResult::Graph::out_edge_iterator ei0, ei1;
#if 0
			tie(ei0, ei1) = boost::edge_range(uu1, uu2, areagraph);
#else
			tie(ei0, ei1) = boost::out_edges(uu1, areagraph);
#endif			
			for (; ei0 != ei1; ++ei0) {
#if 1
				if (boost::target(*ei0, areagraph) != uu2)
					continue;
#endif
				const Engine::AirwayGraphResult::Edge& e(areagraph[*ei0]);
				if (!e.is_level_ok(a)) {
					if (true)
						std::cerr << "Skipping: " << v1.get_ident() << "->" << v2.get_ident() << ": " << e.get_ident()
							  << ": level " << a << " out of range " << e.get_base_level() << "..." << e.get_top_level() << std::endl;
					continue;
				}
				if (true)
					std::cerr << v1.get_ident() << "->" << v2.get_ident() << ": " << e.get_ident() << std::endl;
				an.insert(e.get_ident());
			}

			if (first) {
				awynames.swap(an);
				first = false;
				continue;
			}
			awynames = intersect(awynames, an);
			if (awynames.empty())
				break;
		}
		if (awynames.empty()) {
			if (true)
				std::cerr << "No single airway found from " << v1.get_ident() << " to " << v2.get_ident() << std::endl;
			continue;
		}
		wpt1.set_pathcode(FPlanWaypoint::pathcode_airway);
		wpt1.set_pathname(*awynames.begin());
		if (wpt1.get_name().empty())
			wpt1.set_name(v1.get_ident());
		if (wpt2.get_name().empty())
			wpt2.set_name(v2.get_ident());
	}
}

void IcaoFlightPlan::fix_max_dct_distance(double dct_limit)
{
	// check DCT limit
	for (route_t::iterator ri(m_route.end()), rb(m_route.begin()); ri != rb; ) {
		--ri;
		if (ri == rb)
			break;
		Waypoint& wpt2(*ri);
		route_t::iterator ri1(ri);
		--ri1;
		Waypoint& wpt1(*ri1);
		// VFR or DCT - check for too long segments (limit at 50nmi)
		if (!(wpt1.get_flags() & FPlanWaypoint::altflag_ifr) || wpt1.get_pathname().empty()) {
			double dist(wpt1.get_coord().spheric_distance_nmi_dbl(wpt2.get_coord()));
			if (dist <= dct_limit)
				continue;
	                float crs(wpt1.get_coord().spheric_true_course(wpt2.get_coord()));
			unsigned int nseg(ceil(dist / dct_limit) + 0.1);
			if (nseg > 1)
				dist /= nseg;
			while (nseg > 1) {
				--nseg;
				Waypoint w(wpt1);
				w.set_coord(wpt1.get_coord().spheric_course_distance_nmi(crs, dist * nseg));
				w.set_name("");
				w.set_icao("");
				ri = m_route.insert(ri, w);
				rb = m_route.begin();
			}
			continue;
		}
	}
}

void IcaoFlightPlan::erase_unnecessary_airway(bool keep_turnpoints)
{
	route_t::iterator ri(m_route.end()), rb(m_route.begin());
	if (ri == rb)
		return;
	--ri;
	if (ri == rb)
		return;
	--ri;
	while (ri != rb) {
		route_t::iterator ri1(ri);
		route_t::iterator ri2(ri);
		++ri2;
		Waypoint& wpt2(*ri2);
		Waypoint& wpt1(*ri1);
		--ri;
		Waypoint& wpt0(*ri);
		if (!(wpt0.get_flags() & wpt1.get_flags() & FPlanWaypoint::altflag_ifr) ||
		    ((wpt0.get_flags() ^ wpt1.get_flags()) & FPlanWaypoint::altflag_standard) ||
		    (wpt0.get_altitude() != wpt1.get_altitude()))
			continue;
		if (wpt0.get_pathcode() != FPlanWaypoint::pathcode_airway || wpt1.get_pathcode() != FPlanWaypoint::pathcode_airway)
			continue;
		if (wpt0.get_pathname().empty() || wpt0.get_pathname() != wpt1.get_pathname())
			continue;
		if (keep_turnpoints) {
			static const float argscale((1 << 14) / 90.0);
			static const int16_t argtolerance(5.0 * argscale);
			uint16_t crs0(argscale * wpt0.get_coord().spheric_true_course(wpt1.get_coord()));
			uint16_t crs1(argscale * wpt1.get_coord().spheric_true_course(wpt2.get_coord()));
			if (abs((int16_t)(uint16_t)(crs0 - crs1)) > argtolerance)
				continue;
		}
		ri = m_route.erase(ri1);
		--ri;
		rb = m_route.begin();
	}
}

void IcaoFlightPlan::add_eet(void)
{
	for (route_t::iterator ri(m_route.begin()), re(m_route.end()); ri != re; ++ri) {
		time_t t(ri->get_time());
		if (!t)
			continue;
		struct tm tm;
		if (!gmtime_r(&t, &tm))
			continue;
		const std::string& ident(ri->get_icao_or_name());
		if (Engine::AirwayGraphResult::Vertex::is_ident_numeric(ident) ||
		    ident.size() < 2)
			continue;
		std::ostringstream oss;
		if (!ident.empty())
			oss << upcase(ident);
		else
			oss << ri->get_coordstr();
		oss << std::setw(2) << std::setfill('0') << tm.tm_hour
		    << std::setw(2) << std::setfill('0') << tm.tm_min;
		otherinfo_add("EET", oss.str());
	}
}

std::vector<std::string> IcaoFlightPlan::tokenize(const std::string& s)
{
	std::vector<std::string> r;
	for (std::string::const_iterator si(s.begin()), se(s.end()); si != se; ) {
		if (std::isspace(*si)) {
			++si;
			continue;
		}
		std::string::const_iterator si2(si);
		++si2;
		while (si2 != se && !std::isspace(*si2))
			++si2;
		r.push_back(std::string(si, si2));
		si = si2;
	}
	return r;
}

std::string IcaoFlightPlan::untokenize(const std::vector<std::string>& s)
{
	bool space(false);
	std::string r;
	for (std::vector<std::string>::const_iterator si(s.begin()), se(s.end()); si != se; ++si) {
		if (space)
			r.push_back(' ');
		else
			space = true;
		r += *si;		
	}
	return r;
}

bool IcaoFlightPlan::is_aiport_paris_tma(const std::string& icao) // Paris group airport
{
	if (icao.size() != 4)
		return false;
	if ((icao[0] != 'L' && icao[0] != 'l') ||
	    (icao[1] != 'F' && icao[1] != 'f') ||
	    (icao[2] != 'P' && icao[2] != 'p'))
		return false;
	switch (icao[3]) {
	default:
		return false;

	case 'B': // Le Bourget
	case 'b':
	case 'G': // Charles De Gaulle
	case 'g':
	case 'N': // Toussus Le Noble
	case 'n':
	case 'O': // Paris Orly
	case 'o':
	case 'T': // Pontoise Cormeilles En Vexin
	case 't':
	case 'V': // Villacoublay Velizy
	case 'v':
		return true;

	case 'C': // Creil
	case 'c':
	case 'M': // Melun Villaroche
	case 'm':
		return true;
	}
}

bool IcaoFlightPlan::is_aiport_lfpnv(const std::string& icao)
{
	if (icao.size() != 4)
		return false;
	if ((icao[0] != 'L' && icao[0] != 'l') ||
	    (icao[1] != 'F' && icao[1] != 'f') ||
	    (icao[2] != 'P' && icao[2] != 'p'))
		return false;
	switch (icao[3]) {
	default:
		return false;

	case 'N': // Toussus Le Noble
	case 'n':
	case 'V': // Villacoublay Velizy
	case 'v':
		return true;
	}
}

bool IcaoFlightPlan::is_aiport_lfob(const std::string& icao)
{
	if (icao.size() != 4)
		return false;
	if ((icao[0] != 'L' && icao[0] != 'l') ||
	    (icao[1] != 'F' && icao[1] != 'f') ||
	    (icao[2] != 'O' && icao[2] != 'o') ||
	    (icao[3] != 'B' && icao[3] != 'b'))
		return false;
	return true;
}

bool IcaoFlightPlan::is_route_pogo(const std::string& dep, const std::string& dest)
{
	if (is_aiport_lfob(dep))
		return is_aiport_lfpnv(dest);
	if (is_aiport_lfob(dest))
		return is_aiport_lfpnv(dep);
	return is_aiport_paris_tma(dep) && is_aiport_paris_tma(dest);
}

void IcaoFlightPlan::normalize_pogo(void)
{
	static const std::string remark("RMK");
	static const std::string pogo("POGO");
	std::vector<std::string> tok(tokenize(otherinfo_get(remark)));
	if ((m_departureflags & m_destinationflags & FPlanWaypoint::altflag_ifr) &&
	    is_route_pogo()) {
		for (std::vector<std::string>::const_iterator i(tok.begin()), e(tok.end()); i != e; ++i)
			if (Glib::ustring(*i).uppercase() == pogo)
				return;
		otherinfo_add(remark, pogo);
		return;
	}
	bool chg(false);
	for (std::vector<std::string>::iterator i(tok.begin()), e(tok.end()); i != e; ++i) {
		if (Glib::ustring(*i).uppercase() != pogo)
			continue;
		i = tok.erase(i);
		e = tok.end();
		chg = true;
	}
	if (!chg)
		return;
	m_otherinfo[remark] = untokenize(tok);
}

std::string IcaoFlightPlan::get_fpl(void)
{
	// defaults
	static const std::string fpl_time = "1000";

	if (get_aircraftid().size() < 1 || get_aircraftid().size() > 7)
		return "Error: invalid aircraft ID: " + get_aircraftid();
	if (get_aircrafttype().size() < 2 || get_aircrafttype().size() > 4)
		return "Error: invalid aircraft type: " + get_aircrafttype();
	if (get_departure().size() != 4)
		return "Error: invalid departure: " + get_departure();
	if (get_destination().size() != 4)
		return "Error: invalid destination: " + get_destination();
	std::ostringstream fpl;
	fpl << "-(FPL-" << get_aircraftid() << '-' << get_flightrules() << get_flighttype()
	    << " -" << std::max(get_number(), 1U) << get_aircrafttype() << '/' << get_wakecategory()
	    << " -" << get_equipment() << '/' << get_transponder() << " -" << get_departure();
	{
		time_t t(get_departuretime());
		struct tm tm;
		if (gmtime_r(&t, &tm))
			fpl << std::setw(2) << std::setfill('0') << tm.tm_hour
			    << std::setw(2) << std::setfill('0') << tm.tm_min;
		else
			fpl << fpl_time;
	}
	uint16_t prevflags(m_departureflags);
	int16_t prevalt(m_defaultalt);
	if (!m_route.empty()) {
		prevflags ^= (prevflags ^ m_route[0].get_flags()) & FPlanWaypoint::altflag_standard;
		prevalt = m_route[0].get_altitude();
	}
	fpl << " -";
	if (get_cruisespeed().empty()) {
		int alt(prevalt);
		int cs(Point::round<int,float>(get_cruisespeed(alt)));
		fpl << 'N' << std::setw(4) << std::setfill('0') << cs;
	} else {
		fpl << get_cruisespeed();
	}
	if (!(prevflags & FPlanWaypoint::altflag_ifr)) {
		fpl << "VFR";
	} else {
		fpl << ((prevflags & FPlanWaypoint::altflag_standard) ? 'F' : 'A')
		    << std::setw(3) << std::setfill('0') << ((prevalt + 50) / 100);
	}
	if ((m_departureflags & FPlanWaypoint::altflag_ifr) && !m_sid.empty())
		fpl << ' ' << m_sid;
	{
		for (route_t::const_iterator ri(m_route.begin()), re(m_route.end()); ri != re;) {
			const Waypoint& wpt(*ri);
			const Glib::ustring& ident(wpt.get_icao_or_name());
			++ri;
			fpl << ' ';
			if (!ident.empty())
				fpl << upcase(ident);
			else
				fpl << wpt.get_coordstr();
			{
				// change of level or speed and/or change of rules
				bool lvlchg(prevalt != wpt.get_altitude() ||
					    ((prevflags ^ wpt.get_flags()) & FPlanWaypoint::altflag_standard) ||
					    ((wpt.get_flags() & ~prevflags) & FPlanWaypoint::altflag_ifr));
				bool rulechg((prevflags ^ wpt.get_flags()) & FPlanWaypoint::altflag_ifr);
				prevalt = wpt.get_altitude();
				prevflags = wpt.get_flags();
				if (lvlchg) {
					fpl << '/';
					if (get_cruisespeed().empty()) {
						int alt(prevalt);
						int cs(Point::round<int,float>(get_cruisespeed(alt)));
						fpl << 'N' << std::setw(4) << std::setfill('0') << cs;
					} else {
						fpl << get_cruisespeed();
					}
					fpl << ((prevflags & FPlanWaypoint::altflag_standard) ? 'F' : 'A')
					    << std::setw(3) << std::setfill('0') << ((prevalt + 50) / 100);
				}
				if (rulechg)
					fpl << ' ' << ((prevflags & FPlanWaypoint::altflag_ifr) ? 'I' : 'V') << "FR";
			}
			if (ri == re)
				break;
			if (true) {
				switch (wpt.get_pathcode()) {
				case FPlanWaypoint::pathcode_airway:
					fpl << ' ' << wpt.get_pathname();
					break;

				case FPlanWaypoint::pathcode_directto:
					fpl << " DCT";
					break;

				default:
					break;
				}
			} else if ((wpt.get_flags() & FPlanWaypoint::altflag_ifr)) {
				if (wpt.get_pathname().empty())
					fpl << " DCT";
				else
					fpl << ' ' << wpt.get_pathname();
			}
		}
	}
	if (m_route.empty() && (m_departureflags & m_destinationflags & FPlanWaypoint::altflag_ifr))
		fpl << " DCT";
	if ((prevflags ^ m_destinationflags) & FPlanWaypoint::altflag_ifr)
		fpl << ' ' << ((m_destinationflags & FPlanWaypoint::altflag_ifr) ? 'I' : 'V') << "FR";
	if ((m_destinationflags & FPlanWaypoint::altflag_ifr) && !m_star.empty())
		fpl << ' ' << m_star;
	fpl << " -" << get_destination();
	{
		time_t t(get_totaleet());
		struct tm tm;
		if (gmtime_r(&t, &tm))
			fpl << std::setw(2) << std::setfill('0') << tm.tm_hour
			    << std::setw(2) << std::setfill('0') << tm.tm_min;
		else
			fpl << "0000";
	}
	if (!get_alternate1().empty())
		fpl << ' ' << get_alternate1();
	if (!get_alternate2().empty())
		fpl << ' ' << get_alternate2();
	fpl << " -";
	{
		bool pbndone(get_pbn().empty());
		for (otherinfo_t::const_iterator oi(m_otherinfo.begin()), oe(m_otherinfo.end()); oi != oe; ++oi) {
			fpl << oi->first << '/' << oi->second << ' ';
			if (!pbndone && oi->first == "PBN") {
				fpl << get_pbn() << ' ';
				pbndone = true;
			}
		}
		if (!pbndone)
			fpl << "PBN/" << get_pbn() << ' ';
	}
	fpl << '-';
	{
		bool subseq(false);
		if (get_endurance()) {
			time_t t(get_endurance());
			struct tm tm;
			if (gmtime_r(&t, &tm)) {
				fpl << "E/" << std::setw(2) << std::setfill('0') << tm.tm_hour
				    << std::setw(2) << std::setfill('0') << tm.tm_min;
				subseq = true;
			}
		}
		if (get_personsonboard()) {
			if (subseq)
				fpl << ' ';
			fpl << "P/";
			if (get_personsonboard() == ~0U)
				fpl << "TBN";
			else
				fpl << std::setw(3) << std::setfill('0') << get_personsonboard();
			subseq = true;
		}
		if (!get_emergencyradio().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "R/" << get_emergencyradio();
			subseq = true;
		}
		if (!get_survival().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "S/" << get_survival();
			subseq = true;
		}
		if (!get_lifejackets().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "J/" << get_lifejackets();
			subseq = true;
		}
		if (!get_dinghies().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "D/" << get_dinghies();
			subseq = true;
		}
		if (!get_colourmarkings().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "A/" << get_colourmarkings();
			subseq = true;
		}
		if (!get_remarks().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "N/" << get_remarks();
			subseq = true;
		}
		if (!get_picname().empty()) {
			if (subseq)
				fpl << ' ';
			fpl << "C/" << get_picname();
			subseq = true;
		}
	}
	fpl << ')';
	return fpl.str();
}

void IcaoFlightPlan::set_route(FPlanRoute& route)
{
	route.set_time_offblock_unix(get_departuretime());
	route.set_time_onblock_unix(get_departuretime() + get_totaleet());
	{
		FPlanWaypoint wpt;
		wpt.set_icao(get_departure());
		wpt.set_name(otherinfo_get("DEP"));
		if (get_departure() == "ZZZZ")
			wpt.set_icao("");
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		wpt.set_pathname("");
		wpt.set_time_unix(get_departuretime());
		wpt.set_flags(m_departureflags);
		FindCoord f(m_engine);
		if (f.find(wpt.get_icao(), wpt.get_name(), FindCoord::flag_airport)) {
			wpt.set_icao(f.get_icao());
			wpt.set_name(f.get_name());
			wpt.set_coord(f.get_coord());
			AirportsDb::Airport arpt;
			if (f.get_airport(arpt)) {
				wpt.set_altitude(arpt.get_elevation());
				if (arpt.get_nrcomm())
					wpt.set_frequency(arpt.get_comm(0)[0]);
			}
		}
		route.insert_wpt(~0, wpt);
	}
	for (route_t::iterator ri(m_route.begin()), re(m_route.end()); ri != re; ++ri) {
		Waypoint& wpt(*ri);
		FPlanWaypoint fwpt;
		fwpt.set_icao(wpt.get_icao());
		fwpt.set_name(wpt.get_name());
		fwpt.set_pathcode(wpt.get_pathcode());
		fwpt.set_pathname(wpt.get_pathname());
		fwpt.set_coord(wpt.get_coord());
                fwpt.set_frequency(wpt.get_frequency());
		fwpt.set_altitude(wpt.get_altitude());
		fwpt.set_flags(wpt.get_flags() & (FPlanWaypoint::altflag_standard | FPlanWaypoint::altflag_ifr));
		fwpt.set_note(wpt.get_pathname());
		fwpt.set_time_unix(get_departuretime());
		route.insert_wpt(~0, fwpt);
	}
	{
		FPlanWaypoint wpt;
		wpt.set_icao(get_destination());
		wpt.set_name(otherinfo_get("DEST"));
		if (get_departure() == "ZZZZ")
			wpt.set_icao("");
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		wpt.set_pathname("");
		wpt.set_time_unix(get_departuretime() + get_totaleet());
		wpt.set_flags(m_destinationflags);
		FindCoord f(m_engine);
		if (f.find(wpt.get_icao(), wpt.get_name(), FindCoord::flag_airport)) {
			wpt.set_icao(f.get_icao());
			wpt.set_name(f.get_name());
			wpt.set_coord(f.get_coord());
			AirportsDb::Airport arpt;
			if (f.get_airport(arpt)) {
				wpt.set_altitude(arpt.get_elevation());
				if (arpt.get_nrcomm())
					wpt.set_frequency(arpt.get_comm(0)[0]);
			}
		}
		route.insert_wpt(~0, wpt);
	}
	if (route.get_nrwpt() >= 2U && (m_departureflags & FPlanWaypoint::altflag_ifr) && !m_sid.empty()) {
		if (m_sid == "DCT") {
			route[0].set_pathcode(FPlanWaypoint::pathcode_directto);
			route[0].set_pathname("");
		} else {
			route[0].set_pathcode(FPlanWaypoint::pathcode_sid);
			route[0].set_pathname(m_sid);
		}
	}
	if (route.get_nrwpt() >= 3U && (m_destinationflags & FPlanWaypoint::altflag_ifr) && !m_star.empty()) {
		route[route.get_nrwpt() - 2U].set_pathcode(FPlanWaypoint::pathcode_star);
		route[route.get_nrwpt() - 2U].set_pathname(m_star);
	}
}

void IcaoFlightPlan::recompute_eet(void)
{
	if (m_cruisespeeds.empty())
		return;
	time_t eet(0);
	for (route_t::iterator ri(m_route.begin()), re(m_route.end()); ri != re;) {
		route_t::iterator ri0(ri);
		++ri;
		if (ri == re)
			break;
		int alt(ri0->get_altitude() + ri->get_altitude());
		alt >>= 1;
		float tas(get_cruisespeed(alt));
		if (std::isnan(tas) || tas <= 0)
			return;
		float d(ri0->get_coord().spheric_distance_nmi(ri->get_coord()));
		float t(d / tas * 3600.0);
		eet += t;
	}
	set_totaleet(eet);
}

float IcaoFlightPlan::get_cruisespeed(int& alt) const
{
	if (m_cruisespeeds.empty())
		return std::numeric_limits<float>::quiet_NaN();
	cruisespeeds_t::const_iterator i1(m_cruisespeeds.lower_bound(alt));
	if (i1 == m_cruisespeeds.end()) {
		--i1;
		alt = i1->first;
		return i1->second;
	}
	if (i1->first == alt)
		return i1->second;
	if (i1 == m_cruisespeeds.begin()) {
		alt = i1->first;
		return i1->second;
	}
	cruisespeeds_t::const_iterator i0(i1);
	--i0;
	return (i0->second * (i1->first - alt) + i1->second * (alt - i0->first)) / static_cast<float>(i1->first - i0->first);
}

void IcaoFlightPlan::set_aircraft(const Aircraft& acft, double rpm, double mp, double bhp)
{
	set_aircraftid(acft.get_callsign());
	set_aircrafttype(acft.get_icaotype());
	set_wakecategory(acft.get_wakecategory());
	set_equipment(acft.get_equipment());
	set_transponder(acft.get_transponder());
	set_pbn(acft.get_pbn());
	set_emergencyradio(acft.get_emergencyradio());
	set_survival(acft.get_survival());
	set_lifejackets(acft.get_lifejackets());
	set_dinghies(acft.get_dinghies());
	for (Aircraft::otherinfo_const_iterator_t oii(acft.otherinfo_begin()), oie(acft.otherinfo_end()); oii != oie; ++oii)
		otherinfo_add(*oii);
	set_cruisespeed("");
	{
		m_cruisespeeds.clear();
		Aircraft::Cruise& cruise(const_cast<Aircraft::Cruise&>(acft.get_cruise()));
		double avgff(0), avgtas(0);
		unsigned int avgnr(0);
		for (double da = 0; da < 60000; da += 500) {
			double da1(da), rpm1(rpm), mp1(mp), tas(0), fuelflow(0), bhp1(bhp);
			cruise.calculate(tas, fuelflow, bhp1, rpm1, mp1, da1);
			if (da1 < da - 100 || tas <= 0)
				break;
			m_cruisespeeds.insert(std::make_pair(static_cast<int>(da1), static_cast<float>(tas)));
			if (da < 5000)
				continue;
			avgff += fuelflow;
			avgtas += tas;
			++avgnr;
		}
		if (avgnr) {
			avgff /= avgnr;
			avgtas /= avgnr;
			// compute total fuel content
			double fuel(0);
			const Aircraft::WeightBalance& wb(acft.get_wb());
			for (unsigned int i = 0; i < wb.get_nrelements(); ++i) {
				const Aircraft::WeightBalance::Element& e(wb.get_element(i));
				if (!(e.get_flags() & Aircraft::WeightBalance::Element::flag_avgas))
					continue;
				fuel += e.get_massmax();
			}
			if (true)
				std::cerr << "Total fuel mass: " << fuel << std::endl;
			{
				double fm(acft.get_fuelmass());
				if (fm > 0)
					fuel /= fm;
			}
			if (true)
				std::cerr << "Total fuel volume: " << fuel << " average cruise fuel flow: " << avgff << std::endl;
			if (avgff > 0) {
				fuel /= avgff;
				m_endurance = fuel * 3600.0;
			}
		}
	}
	set_colourmarkings(acft.get_colormarking());
}

#if 0

(FPL-OEABC-VG
-DV20/L-S/C
-LOWI0800
-N0090VFR WHISKEY
-LOWI0001
-STS/EET0200 , RMK/THIS IS A REMARK DOF/051102
-E/0500 P/003 R/E J/F
A/RED WHITE
C/HUBER)

(FPL-OEABC-VG
-C150/L-OV/C
-LOGG1220
-N0080VFR LH MK VM
-LOWK0130 LOWG
-DOF/050719)

(FPL-OEABC-IN
-A321/M-SEHRWY/S
-LOWW0915
-N0458F340 OSPEN UM984 BZO/N0458F350 UM984 NEDED/N0456F370 UM984
DIVKO UN852 VERSO UL129 LUNIK
-LEPA0157 LEIB
-EET/LIMM0030 LFFF0106 LECB0135 SEL/APHL OPR/TESTAIR DOF/050719
RVR/75 ORGN/LOWWZPZX)

(FPL-OEABC-IS
-B772/H-SEGHIJRWXY/S
-LOWW0913
-N0480F350 ABLOM UL725 ALAMU UY33 NARKA UB499 REBLA UL601 CND UM747
SOBLO/N0490F350 B143 VEKAT UM747 DUKAN B449 DURDY/N0490F370 B449 RJ
A240 AFGAN W615A BUROT B143D ABDAN N644 DI A466 ASARI/N0490F370
A466E DPN L759S AGG L759 KKJ/N0490F370 L759 IBUDA/N0490F370 L759
LIBDI/N0490F370 L759 NISUN/N0490F370 L759 MIPAK/N0490F370 L759 PUT
R325 VIH A464 VBA
-WMKK1046 WSSS
-EET/UKFV0122 URRV0153 UGGG0212 UBBA0244 UTAK0311 UTAA0332 OAKX0426
OPLR0504 VIDF0540 VABF0639 VECF0700 VYYF0838 VOMF0842 VYYF0903
VTBB0925 WMFC1003 SEL/AEBQ OPR/ABC AIR PER/C DAT/SV RALT/VECC
VTSP RMK/AGCS DOF/050719 ORGN/LOWWZPZX)

(FPL-OEABC-IS
-CRJ2/M-SERWY/S
-LOWS0550
-N0450F340 TRAUN L856 MANAL UL856 DERAK UM982 TINIL
-LFPG0120 LFPO
-DOF/050719 RVR/200 ORGN/RPL)

#endif
